{*************************************************************************
                Copyright (c) PilotLogic Software House

  Package pl_MagicScript
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: LGPL-2.1 with link exception (Modified LGPL)
 *
 * The contents of this file are subject to the
 * GNU LIBRARY GENERAL PUBLIC LICENSE Version 2.1 (the "License")
 * with the following modification:
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent modules,
 * and to copy and distribute the resulting executable under terms of your choice,
 * provided that you also meet, for each linked independent module, the terms
 * and conditions of the license of that module. An independent module is a
 * module which is not derived from or based on this library. If you modify this
 * library, you may extend this exception to your version of the library, but
 * you are not obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * ***** END LICENSE BLOCK *****
 
***************************************************************************}

Unit msScriptHighlighter;

{$H+}

Interface

Uses
   SysUtils, LCLProc,
   Classes, Registry, Graphics, SynEditHighlighterFoldBase, SynEditMiscProcs,
   SynEditTypes, SynEditHighlighter, SynEditTextBase, SynEditStrConst, SynEditMiscClasses;

Type
   TmscrPasStringMode = (spsmDefault, spsmStringOnly, spsmNone);

   TmscrTokenKind = (tkComment, tkIdentifier, tkKey, tkNull, tkNumber,
      tkSpace, tkString, tkSymbol, tkDirective, tkIDEDirective,
      tkUnknown);

   TmscrRangeState = (
      rsAnsi,
      rsBor,
      rsSlash,
      rsIDEDirective,
      rsDirective,
      rsAfterEqualOrColon,
      rsAfterEqual,
      rsAfterIdentifierOrValue,
      rsAfterIdentifierOrValueAdd,
      rsAtClosingBracket,
      rsAtCaseLabel,
      rsInProcHeader,
      rsVarTypeInSpecification,
      rsInTypeBlock,
      rsAfterEqualThenType,
      rsSkipAllPasBlocks
      );
   TmscrRangeStates = Set Of TmscrRangeState;

Type
   TmscrPascalCodeFoldBlockType = (// Do *not* change the order
      cfbtBeginEnd,
      cfbtTopBeginEnd,
      cfbtNestedComment,
      cfbtProcedure,
      cfbtUses,
      cfbtVarType,
      cfbtLocalVarType,
      cfbtUnitSection,
      cfbtProgram,
      cfbtUnit,
      cfbtTry,
      cfbtExcept,
      cfbtRepeat,
      cfbtCase,
      cfbtIfDef,
      cfbtRegion,
      cfbtAnsiComment,
      cfbtBorCommand,
      cfbtSlashComment,
      cfbtIfThen,
      cfbtForDo,
      cfbtWhileDo,
      cfbtWithDo,
      cfbtIfElse,
      cfbtCaseElse,
      cfbtNone
      );
   TmscrPascalCodeFoldBlockTypes = Set Of TmscrPascalCodeFoldBlockType;

Const
   cfbtLastPublic = cfbtIfElse;
   cfbtFirstPrivate = cfbtCaseElse;

   CountPascalCodeFoldBlockOffset =
      Pointer(PtrInt(integer(high(TmscrPascalCodeFoldBlockType)) + 1));

   cfbtAll = TmscrPascalCodeFoldBlockTypes([low(TmscrPascalCodeFoldBlockType)..high(TmscrPascalCodeFoldBlockType)]);

   PascalWordTripletRanges = TmscrPascalCodeFoldBlockTypes([cfbtBeginEnd, cfbtTopBeginEnd, cfbtProcedure, cfbtProgram, cfbtTry, cfbtExcept, cfbtRepeat, cfbtCase, cfbtCaseElse,
      cfbtIfDef, cfbtRegion, cfbtIfThen, cfbtForDo, cfbtWhileDo, cfbtWithDo]);
   PascalNoOutlineRanges = TmscrPascalCodeFoldBlockTypes([cfbtProgram, cfbtUnit, cfbtUnitSection, cfbtRegion, cfbtVarType, cfbtCaseElse, cfbtIfDef, cfbtAnsiComment, cfbtBorCommand, cfbtSlashComment, cfbtNestedComment]);

   ProcModifierAllowed = TmscrPascalCodeFoldBlockTypes([cfbtNone, cfbtProcedure, cfbtProgram, cfbtUnitSection, cfbtVarType, cfbtLocalVarType]);

   PascalStatementBlocks = TmscrPascalCodeFoldBlockTypes([cfbtBeginEnd, cfbtTopBeginEnd, cfbtCase, cfbtTry, cfbtExcept, cfbtRepeat, cfbtCaseElse, cfbtIfThen, cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfElse]);

   cfbtEssential = TmscrPascalCodeFoldBlockTypes([cfbtUnitSection, cfbtProcedure, cfbtProgram, cfbtVarType, cfbtLocalVarType, cfbtAnsiComment, cfbtBorCommand, cfbtSlashComment,
      cfbtNestedComment, cfbtCase, cfbtTry] + ProcModifierAllowed + PascalStatementBlocks - [cfbtBeginEnd, cfbtCase, cfbtExcept, cfbtRepeat, cfbtCaseElse, cfbtIfThen, cfbtForDo, cfbtWhileDo, cfbtWithDo]);

   PascalFoldTypeCompatibility: Array[TmscrPascalCodeFoldBlockType] Of TmscrPascalCodeFoldBlockType =
   (cfbtBeginEnd,
      cfbtBeginEnd,
      cfbtNestedComment,
      cfbtProcedure,
      cfbtUses,
      cfbtVarType,
      cfbtVarType,
      cfbtUnitSection,
      cfbtProgram,
      cfbtUnit,
      cfbtTry,
      cfbtExcept,
      cfbtRepeat,
      cfbtCase,
      cfbtIfDef,
      cfbtRegion,
      cfbtNestedComment,
      cfbtNestedComment,
      cfbtSlashComment,
      cfbtIfThen,
      cfbtForDo,
      cfbtWhileDo,
      cfbtWithDo,
      cfbtIfElse,
      cfbtCaseElse,
      cfbtNone
      );

   FOLDGROUP_PASCAL = 1;
   FOLDGROUP_REGION = 2;
   FOLDGROUP_IFDEF = 3;

Type

   TPascalCompilerMode = (
      pcmObjFPC,
      pcmDelphi,
      pcmFPC,
      pcmTP,
      pcmGPC,
      pcmMacPas
      );

   TSynPasDividerDrawLocation = (
      pddlUnitSection,
      pddlUses,
      pddlVarGlobal, // Var, Type, Const in global scope
      pddlVarLocal, // Var, Type, Const in local (procedure) scope
      pddlStructGlobal, // Class, Object, Record in global type block
      pddlStructLocal, // Class, Object, Record in local (procedure) type block
      pddlProcedure,
      pddlBeginEnd, // Includes Repeat
      pddlTry
      );

Const

   PasDividerDrawLocationDefaults: Array[TSynPasDividerDrawLocation] Of
   integer = (1, 0, // unit, uses
      1, 0, // var
      1, 0, // struct
      2, 0, // proc, begin
      0);

Type

   TSynPasRangeInfo = Record
      EndLevelIfDef: smallint;
      MinLevelIfDef: smallint;
      EndLevelRegion: smallint;
      MinLevelRegion: smallint;
   End;

   { TSynHighlighterPasRangeList }

   TSynHighlighterPasRangeList = Class(TSynHighlighterRangeList)
   Private
      FItemOffset: integer;
      Function GetTSynPasRangeInfo(Index: integer): TSynPasRangeInfo;
      Procedure SetTSynPasRangeInfo(Index: integer; Const AValue: TSynPasRangeInfo);
   Public
      Constructor Create;
      Property PasRangeInfo[Index: integer]: TSynPasRangeInfo Read GetTSynPasRangeInfo Write SetTSynPasRangeInfo;
   End;

   { TmscrScriptHighlighterRange }

   TmscrScriptHighlighterRange = Class(TSynCustomHighlighterRange)
   Private
      FMode: TPascalCompilerMode;
      FNestedComments: boolean;
      FBracketNestLevel: integer;
      FLastLineCodeFoldLevelFix: integer;
      FPasFoldFixLevel: smallint;
      FTypeHelpers: boolean;
   Public
      Procedure Clear; Override;
      Function Compare(Range: TSynCustomHighlighterRange): integer; Override;
      Procedure Assign(Src: TSynCustomHighlighterRange); Override;
      Function MaxFoldLevel: integer; Override;
      Procedure IncBracketNestLevel;
      Procedure DecBracketNestLevel;
      Procedure DecLastLineCodeFoldLevelFix;
      Procedure DecLastLinePasFoldFix;
      Property Mode: TPascalCompilerMode Read FMode Write FMode;
      Property NestedComments: boolean Read FNestedComments Write FNestedComments;
      Property TypeHelpers: boolean Read FTypeHelpers Write FTypeHelpers;
      Property BracketNestLevel: integer Read FBracketNestLevel Write FBracketNestLevel;
      Property LastLineCodeFoldLevelFix: integer Read FLastLineCodeFoldLevelFix Write FLastLineCodeFoldLevelFix;
      Property PasFoldFixLevel: smallint Read FPasFoldFixLevel Write FPasFoldFixLevel;
   End;

   TProcTableProc = Procedure Of Object;

   PIdentFuncTableFunc = ^TIdentFuncTableFunc;
   TIdentFuncTableFunc = Function: TmscrTokenKind Of Object;

   { TmscrScriptHighlighter }

   TmscrScriptHighlighter = Class(TSynCustomFoldHighlighter)
   Private
      fAsmStart: boolean;
      FExtendedKeywordsMode: boolean;
      FNestedComments: boolean;
      FProcedureHeaderNameAttr: TSynHighlighterAttributesModifier;
      FCurProcedureHeaderNameAttr: TSynSelectedColorMergeResult;
      FStartCodeFoldBlockLevel: integer; // TODO: rename FStartNestedFoldBlockLevel
      FPasStartLevel: smallint;
      fRange: TmscrRangeStates;
      FOldRange: TmscrRangeStates;
      FStringKeywordMode: TmscrPasStringMode;
      FSynPasRangeInfo: TSynPasRangeInfo;
      FAtLineStart: boolean; // Line had only spaces or comments sofar
      fLineStr: String;
      fLine: PChar;
      fLineLen: integer;
      fLineNumber: integer;
      fProcTable: Array[#0..#255] Of TProcTableProc;
      FTypeHelpers: boolean;
      Run: longint; // current parser postion in fLine
      fStringLen: integer; // current length of hash
      fToIdent: integer; // start of current identifier in fLine
      fIdentFuncTable: Array[0..191] Of TIdentFuncTableFunc;
      fTokenPos: integer; // start of current token in fLine
      FTokenID: TmscrTokenKind;
      FTokenIsCaseLabel: boolean;
      fStringAttri: TSynHighlighterAttributes;
      fNumberAttri: TSynHighlighterAttributes;
      fKeyAttri: TSynHighlighterAttributes;
      fSymbolAttri: TSynHighlighterAttributes;
      fCommentAttri: TSynHighlighterAttributes;
      FIDEDirectiveAttri: TSynHighlighterAttributesModifier;
      FCurIDEDirectiveAttri: TSynSelectedColorMergeResult;
      fIdentifierAttri: TSynHighlighterAttributes;
      fSpaceAttri: TSynHighlighterAttributes;
      FCaseLabelAttri: TSynHighlighterAttributesModifier;
      FCurCaseLabelAttri: TSynSelectedColorMergeResult;
      fDirectiveAttri: TSynHighlighterAttributes;
      FCompilerMode: TPascalCompilerMode;
      fD4syntax: boolean;
      // Divider
      FDividerDrawConfig: Array[TSynPasDividerDrawLocation] Of TSynDividerDrawConfig;

      Function GetPasCodeFoldRange: TmscrScriptHighlighterRange;
      Procedure SetCompilerMode(Const AValue: TPascalCompilerMode);
      Procedure SetExtendedKeywordsMode(Const AValue: boolean);
      Procedure SetNestedComments(Const ANestedComments: boolean);
      Procedure SetStringKeywordMode(Const AValue: TmscrPasStringMode);
      Function TextComp(aText: PChar): boolean;
      Function KeyHash: integer;
      Function Func15: TmscrTokenKind;
      Function Func19: TmscrTokenKind;
      Function Func20: TmscrTokenKind;
      Function Func21: TmscrTokenKind;
      Function Func23: TmscrTokenKind;
      Function Func25: TmscrTokenKind;
      Function Func28: TmscrTokenKind;
      Function Func29: TmscrTokenKind; // "on"
      Function Func32: TmscrTokenKind;
      Function Func33: TmscrTokenKind;
      Function Func35: TmscrTokenKind;
      Function Func37: TmscrTokenKind;
      Function Func38: TmscrTokenKind;
      Function Func39: TmscrTokenKind;
      Function Func41: TmscrTokenKind;
      Function Func44: TmscrTokenKind;
      Function Func45: TmscrTokenKind;
      Function Func47: TmscrTokenKind;
      Function Func49: TmscrTokenKind;
      Function Func52: TmscrTokenKind;
      Function Func57: TmscrTokenKind;
      Function Func58: TmscrTokenKind;
      Function Func59: TmscrTokenKind;
      Function Func60: TmscrTokenKind;
      Function Func63: TmscrTokenKind;
      Function Func64: TmscrTokenKind;
      Function Func65: TmscrTokenKind;
      Function Func71: TmscrTokenKind;
      Function Func73: TmscrTokenKind;
      Function Func76: TmscrTokenKind;
      Function Func79: TmscrTokenKind;
      Function Func85: TmscrTokenKind;
      Function Func86: TmscrTokenKind;
      Function Func87: TmscrTokenKind;
      Function Func88: TmscrTokenKind;
      Function Func91: TmscrTokenKind;
      Function Func97: TmscrTokenKind;
      Function Func99: TmscrTokenKind;
      Function Func100: TmscrTokenKind;
      Function Func101: TmscrTokenKind;
      Function Func102: TmscrTokenKind;
      Function Func105: TmscrTokenKind;
      Function Func122: TmscrTokenKind;
      Function Func125: TmscrTokenKind;
      Function Func128: TmscrTokenKind;
      Function Func130: TmscrTokenKind;
      Function Func139: TmscrTokenKind;
      Function Func158: TmscrTokenKind;
      Function Func167: TmscrTokenKind;
      Function Func170: TmscrTokenKind;
      Function Func181: TmscrTokenKind;
      Function AltFunc: TmscrTokenKind;
      Procedure InitIdent;
      Function IdentKind(p: integer): TmscrTokenKind;
      Procedure MakeMethodTables;
      Procedure AddressOpProc;
      Procedure AsciiCharProc;
      Procedure AnsiProc;
      Procedure BorProc;
      Procedure BraceOpenProc;
      Procedure ColonProc;
      Procedure GreaterProc;
      Procedure CRProc;
      Procedure DirectiveProc;
      Procedure IdentProc;
      Procedure HexProc;
      Procedure BinaryProc;
      Procedure OctalProc;
      Procedure LFProc;
      Procedure LowerProc;
      Procedure CaretProc;
      Procedure NullProc;
      Procedure NumberProc;
      Procedure PointProc;
      Procedure RoundOpenProc;
      Procedure RoundCloseProc;
      Procedure SquareOpenProc;
      Procedure SquareCloseProc;
      Procedure EqualSignProc;
      Procedure SemicolonProc; //mh 2000-10-08
      Procedure SlashProc;
      Procedure SlashContinueProc;
      Procedure SpaceProc;
      Procedure StringProc;
      Procedure SymbolProc;
      Procedure UnknownProc;
      Procedure SetD4syntax(Const Value: boolean);
      // Divider
      Procedure CreateDividerDrawConfig;
      Procedure DestroyDividerDrawConfig;
   Protected
      Function KeyComp(Const aKey: String): boolean;
      Function KeyCompEx(AText1, AText2: PChar; ALen: integer): boolean;
      Function GetIdentChars: TSynIdentChars; Override;
      Function IsFilterStored: boolean; Override; //mh 2000-10-08
   Protected
      // "Range"
      Function GetRangeClass: TSynCustomHighlighterRangeClass; Override;
      Procedure CreateRootCodeFoldBlock; Override;
      Function CreateRangeList(ALines: TSynEditStringsBase): TSynHighlighterRangeList; Override;
      Function UpdateRangeInfoAtLine(Index: integer): boolean; Override; // Returns true if range changed

      Property PasCodeFoldRange: TmscrScriptHighlighterRange Read GetPasCodeFoldRange;
      Function TopPascalCodeFoldBlockType
         (DownIndex: integer = 0): TmscrPascalCodeFoldBlockType;

      // Open/Close Folds
      Procedure GetTokenBounds(Out LogX1, LogX2: integer); Override;
      Function StartPascalCodeFoldBlock
         (ABlockType: TmscrPascalCodeFoldBlockType; ForceDisabled: boolean = False): TSynCustomCodeFoldBlock;
      Procedure EndPascalCodeFoldBlock(NoMarkup: boolean = False; UndoInvalidOpen: boolean = False);
      Procedure CloseBeginEndBlocksBeforeProc;
      Procedure SmartCloseBeginEndBlocks(SearchFor: TmscrPascalCodeFoldBlockType);
      Procedure EndPascalCodeFoldBlockLastLine;
      Procedure StartCustomCodeFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType);
      Procedure EndCustomCodeFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType);
      Procedure CollectNodeInfo(FinishingABlock: boolean; ABlockType: Pointer; LevelChanged: boolean); Override;

      // Info about Folds
      Function CreateFoldNodeInfoList: TLazSynFoldNodeInfoList; Override;
      Procedure ScanFoldNodeInfo(); Override;
      Procedure DoInitNode(Var Node: TSynFoldNodeInfo;
         //EndOffs: Integer;
         FinishingABlock: boolean; ABlockType: Pointer; aActions: TSynFoldActions; AIsFold: boolean); Override;

   Protected
      Function LastLinePasFoldLevelFix(Index: integer; AType: integer = 1; AIncludeDisabled: boolean = False): integer; // TODO deprecated; // foldable nodes

      // Divider
      Function GetDrawDivider(Index: integer): TSynDividerDrawConfigSetting; Override;
      Function GetDividerDrawConfig(Index: integer): TSynDividerDrawConfig; Override;
      Function GetDividerDrawConfigCount: integer; Override;

      // Fold Config
      Function CreateFoldConfigInstance(Index: integer): TSynCustomFoldConfig; Override;
      Function GetFoldConfigInstance(Index: integer): TSynCustomFoldConfig; Override;
      Function GetFoldConfigCount: integer; Override;
      Function GetFoldConfigInternalCount: integer; Override;
      Procedure DoFoldConfigChanged(Sender: TObject); Override;
   Public
      Class Function GetCapabilities: TSynHighlighterCapabilities; Override;
      Class Function GetLanguageName: String; Override;
   Public
      Constructor Create(AOwner: TComponent); Override;
      Destructor Destroy; Override;
      Function GetDefaultAttribute(Index: integer): TSynHighlighterAttributes; Override;
      Function GetEol: boolean; Override;
      Function GetRange: Pointer; Override;
      Function GetToken: String; Override;
      Procedure GetTokenEx(Out TokenStart: PChar; Out TokenLength: integer); Override;
      Function GetTokenAttribute: TSynHighlighterAttributes; Override;
      Function GetTokenID: TmscrTokenKind;
      Function GetTokenKind: integer; Override;
      Function GetTokenPos: integer; Override;
      Function IsKeyword(Const AKeyword: String): boolean; Override;
      Procedure Next; Override;

      Procedure ResetRange; Override;
      Procedure SetLine(Const NewValue: String; LineNumber: integer); Override;
      Procedure SetRange(Value: Pointer); Override;
      Procedure StartAtLineIndex(LineNumber: integer); Override; // 0 based
      Function GetEndOfLineAttribute: TSynHighlighterAttributes; Override;

      Function UseUserSettings(settingIndex: integer): boolean; Override;
      Procedure EnumUserSettings(settings: TStrings); Override;

      Function IsLineStartingInDirective(ALineIndex: TLineIdx): boolean;
      // Info about Folds
      //function FoldBlockOpeningCount(ALineIndex: TLineIdx; const AFilter: TSynFoldBlockFilter): integer; override; overload;
      //function FoldBlockClosingCount(ALineIndex: TLineIdx; const AFilter: TSynFoldBlockFilter): integer; override; overload;
      Function FoldBlockEndLevel(ALineIndex: TLineIdx; Const AFilter: TSynFoldBlockFilter): integer; Override; Overload;
      Function FoldBlockMinLevel(ALineIndex: TLineIdx; Const AFilter: TSynFoldBlockFilter): integer; Override; Overload;
      Function FoldBlockNestedTypes(ALineIndex: TLineIdx; ANestIndex: integer; Out AType: Pointer; Const AFilter: TSynFoldBlockFilter): boolean; Override; Overload;

      Function FoldTypeCount: integer; Override;
      Function FoldTypeAtNodeIndex(ALineIndex, FoldIndex: integer; // accesses FoldNodeInfo
         UseCloseNodes: boolean = False): integer; Override;
      Function FoldLineLength(ALineIndex, FoldIndex: integer): integer; Override; // accesses FoldNodeInfo
      Function FoldEndLine(ALineIndex, FoldIndex: integer): integer; Override; // accesses FoldNodeInfo

   Published
      Property CommentAttri: TSynHighlighterAttributes Read fCommentAttri Write fCommentAttri;
      Property IDEDirectiveAttri: TSynHighlighterAttributesModifier Read FIDEDirectiveAttri Write FIDEDirectiveAttri;
      Property IdentifierAttri: TSynHighlighterAttributes Read fIdentifierAttri Write fIdentifierAttri;
      Property KeyAttri: TSynHighlighterAttributes Read fKeyAttri Write fKeyAttri;
      Property NumberAttri: TSynHighlighterAttributes Read fNumberAttri Write fNumberAttri;
      Property SpaceAttri: TSynHighlighterAttributes Read fSpaceAttri Write fSpaceAttri;
      Property StringAttri: TSynHighlighterAttributes Read fStringAttri Write fStringAttri;
      Property SymbolAttri: TSynHighlighterAttributes Read fSymbolAttri Write fSymbolAttri;
      Property ProcedureHeaderName: TSynHighlighterAttributesModifier Read FProcedureHeaderNameAttr Write FProcedureHeaderNameAttr;
      Property CaseLabelAttri: TSynHighlighterAttributesModifier Read FCaseLabelAttri Write FCaseLabelAttri;
      Property DirectiveAttri: TSynHighlighterAttributes Read fDirectiveAttri Write fDirectiveAttri;
      Property CompilerMode: TPascalCompilerMode Read FCompilerMode Write SetCompilerMode;
      Property NestedComments: boolean Read FNestedComments Write SetNestedComments;
      Property TypeHelpers: boolean Read FTypeHelpers Write FTypeHelpers;
      Property D4syntax: boolean Read FD4syntax Write SetD4syntax Default True;
      Property ExtendedKeywordsMode: boolean Read FExtendedKeywordsMode Write SetExtendedKeywordsMode Default False;
      Property StringKeywordMode: TmscrPasStringMode Read FStringKeywordMode Write SetStringKeywordMode Default spsmDefault;
   End;

   { TSynFreePascalSyn }

   TSynFreePascalSyn = Class(TmscrScriptHighlighter)
   Public
      Constructor Create(AOwner: TComponent); Override;
      Procedure ResetRange; Override;
   End;

Implementation

Const
   RESERVED_WORDS_TP: Array[1..54] Of String = (
      'absolute', 'and', 'array', 'asm',
      'begin',
      'case', 'const', 'constructor',
      'destructor', 'div', 'do', 'downto',
      'else', 'end',
      'file', 'for', 'function',
      'goto',
      'if', 'implementation', 'in', 'inherited', 'inline', 'interface',
      'label',
      'mod',
      'nil', 'not',
      'object', 'of', 'on', 'operator', 'or',
      'packed', 'procedure', 'program',
      'record', 'reintroduce', 'repeat',
      'self', 'set', 'shl', 'shr', 'string',
      'then', 'to', 'type',
      'unit', 'until', 'uses',
      'var',
      'while', 'with',
      'xor'
      );

   RESERVED_WORDS_DELPHI: Array[1..15] Of String = (
      'as',
      'class',
      'except', 'exports',
      'finalization', 'finally',
      'initialization', 'is',
      'library',
      'on', 'out',
      'property',
      'raise',
      'threadvar',
      'try'
      );

   RESERVED_WORDS_FPC: Array[1..5] Of String = (
      'dispose', 'exit', 'false', 'new', 'true'
      );

Var
   Identifiers: Array[#0..#255] Of bytebool;
   mHashTable: Array[#0..#255] Of integer;
   KeywordsList: TStringList;
   IsIntegerChar: Array[char] Of boolean;
   IsNumberChar: Array[char] Of boolean;
   IsSpaceChar: Array[char] Of boolean;
   IsUnderScoreOrNumberChar: Array[char] Of boolean;
   IsLetterChar: Array[char] Of boolean;

Function dbgs(FoldType: TmscrPascalCodeFoldBlockType): String; Overload;
Begin
   WriteStr(Result, FoldType);
End;

Procedure MakeIdentTable;
Var
   I, J: char;
Begin
   For I := #0 To #255 Do
   Begin
      Case I Of
         '_', '0'..'9', 'a'..'z', 'A'..'Z': Identifiers[I] := True;
         Else
            Identifiers[I] := False;
      End;
      J := UpCase(I);
      Case I Of
         'a'..'z', 'A'..'Z': mHashTable[I] := Ord(J) - 64;
         '_': mHashTable[I] := 27; // after Z
         '0'..'9': mHashTable[I] := Ord(J) - 48 + 28; // after "_"
         Else
            mHashTable[char(I)] := 0;
      End;
      IsIntegerChar[I] := (I In ['0'..'9', 'A'..'F', 'a'..'f']);
      IsNumberChar[I] := (I In ['0'..'9']);
      IsSpaceChar[I] := (I In [#1..#9, #11, #12, #14..#32]);
      IsUnderScoreOrNumberChar[I] := (I In ['_', '0'..'9']);
      IsLetterChar[I] := (I In ['a'..'z', 'A'..'Z']);
   End;
End;

Procedure TmscrScriptHighlighter.InitIdent;
Var
   I: integer;
   pF: PIdentFuncTableFunc;
Begin
   pF := PIdentFuncTableFunc(@fIdentFuncTable);
   For I := Low(fIdentFuncTable) To High(fIdentFuncTable) Do
   Begin
      pF^ := @AltFunc;
      Inc(pF);
   End;
   fIdentFuncTable[15] := @Func15;
   fIdentFuncTable[19] := @Func19;
   fIdentFuncTable[20] := @Func20;
   fIdentFuncTable[21] := @Func21;
   fIdentFuncTable[23] := @Func23;
   fIdentFuncTable[25] := @Func25;
   fIdentFuncTable[28] := @Func28;
   fIdentFuncTable[29] := @Func29;
   fIdentFuncTable[32] := @Func32;
   fIdentFuncTable[33] := @Func33;
   fIdentFuncTable[35] := @Func35;
   fIdentFuncTable[37] := @Func37;
   fIdentFuncTable[38] := @Func38;
   fIdentFuncTable[39] := @Func39;
   fIdentFuncTable[41] := @Func41;
   fIdentFuncTable[44] := @Func44;
   fIdentFuncTable[45] := @Func45;
   fIdentFuncTable[47] := @Func47;
   fIdentFuncTable[49] := @Func49;
   fIdentFuncTable[52] := @Func52;
   fIdentFuncTable[57] := @Func57;
   fIdentFuncTable[58] := @Func58;
   fIdentFuncTable[59] := @Func59;
   fIdentFuncTable[60] := @Func60;
   fIdentFuncTable[63] := @Func63;
   fIdentFuncTable[64] := @Func64;
   fIdentFuncTable[65] := @Func65;
   fIdentFuncTable[71] := @Func71;
   fIdentFuncTable[73] := @Func73;
   fIdentFuncTable[76] := @Func76;
   fIdentFuncTable[79] := @Func79;
   fIdentFuncTable[85] := @Func85;
   fIdentFuncTable[86] := @Func86;
   fIdentFuncTable[87] := @Func87;
   fIdentFuncTable[88] := @Func88;
   fIdentFuncTable[91] := @Func91;
   fIdentFuncTable[97] := @Func97;
   fIdentFuncTable[99] := @Func99;
   fIdentFuncTable[100] := @Func100;
   fIdentFuncTable[101] := @Func101;
   fIdentFuncTable[102] := @Func102;
   fIdentFuncTable[105] := @Func105;
   fIdentFuncTable[122] := @Func122;
   fIdentFuncTable[125] := @Func125;
   fIdentFuncTable[128] := @Func128;
   fIdentFuncTable[130] := @Func130;
   fIdentFuncTable[139] := @Func139;
   fIdentFuncTable[158] := @Func158;
   fIdentFuncTable[167] := @Func167;
   fIdentFuncTable[170] := @Func170;
   fIdentFuncTable[181] := @Func181;
End;

Function TmscrScriptHighlighter.KeyHash: integer;
Var
   Start, ToHash: PChar;
Begin
   Result := 0;
   If (fToIdent < fLineLen) Then
   Begin
      Start := fLine + fToIdent;
      ToHash := Start;
      If IsLetterChar[ToHash^] Then
      Begin
         Inc(Result, mHashTable[ToHash^]);
         Inc(ToHash);
         While (IsLetterChar[ToHash^] Or IsUnderScoreOrNumberChar[ToHash^]) Do
         Begin
            Inc(Result, mHashTable[ToHash^]);
            Inc(ToHash);
         End;
      End;
      If IsUnderScoreOrNumberChar[ToHash^] Then
         Inc(ToHash);
      fStringLen := PtrUInt(ToHash) - PtrUInt(Start);
      //if CompareText(copy(fLineStr,fToIdent+1,fStringLen),'varargs')=0 then debugln('TmscrScriptHighlighter.KeyHash '+copy(fLineStr,fToIdent+1,fStringLen)+'='+dbgs(Result));
   End
   Else
   Begin
      fStringLen := 0;
   End;
End; { KeyHash }

Function TmscrScriptHighlighter.KeyComp(Const aKey: String): boolean;
Var
   I: integer;
   Temp: PChar;
Begin
   If Length(aKey) = fStringLen Then
   Begin
      Temp := fLine + fToIdent;
      Result := True;
      For i := 1 To fStringLen Do
      Begin
         If mHashTable[Temp^] <> mHashTable[aKey[i]] Then
         Begin
            Result := False;
            break;
         End;
         Inc(Temp);
      End;
   End
   Else
      Result := False;
End; { KeyComp }

Function TmscrScriptHighlighter.KeyCompEx(AText1, AText2: PChar; ALen: integer): boolean;
Begin
   Result := False;
   While ALen > 0 Do
   Begin
      If mHashTable[AText1^] <> mHashTable[AText2^] Then
         exit;
      Dec(ALen);
      Inc(AText1);
      Inc(AText2);
   End;
   Result := True;
End;

Function TmscrScriptHighlighter.TextComp(aText: PChar): boolean;
Var
   CurPos: PChar;
Begin
   CurPos := @fLine[Run];
   While (aText^ <> #0) Do
   Begin
      If mHashTable[aText^] <> mHashTable[CurPos^] Then exit(False);
      Inc(aText);
      Inc(CurPos);
   End;
   Result := True;
End;

Procedure TmscrScriptHighlighter.SetCompilerMode(Const AValue: TPascalCompilerMode);
Begin
   NestedComments := AValue In [pcmFPC, pcmObjFPC];
   TypeHelpers := AValue In [pcmDelphi];
   If FCompilerMode = AValue Then exit;
   FCompilerMode := AValue;
   PasCodeFoldRange.Mode := FCompilerMode;

End;

Procedure TmscrScriptHighlighter.SetExtendedKeywordsMode(Const AValue: boolean);
Begin
   If FExtendedKeywordsMode = AValue Then exit;
   FExtendedKeywordsMode := AValue;
   FAttributeChangeNeedScan := True;
   DefHighlightChange(self);
End;

Procedure TmscrScriptHighlighter.SetStringKeywordMode(Const AValue: TmscrPasStringMode);
Begin
   If FStringKeywordMode = AValue Then exit;
   FStringKeywordMode := AValue;
   FAttributeChangeNeedScan := True;
   DefHighlightChange(self);
End;

Function TmscrScriptHighlighter.GetPasCodeFoldRange: TmscrScriptHighlighterRange;
Begin
   Result := TmscrScriptHighlighterRange(CodeFoldRange);
End;

Function TmscrScriptHighlighter.Func15: TmscrTokenKind;
Begin
   If KeyComp('If') Then
   Begin
      // Anything that may be nested in a "case", and does not have an end (like "end", "until",...)
      StartPascalCodeFoldBlock(cfbtIfThen,
         TopPascalCodeFoldBlockType In [cfbtCase, cfbtIfThen, cfbtIfElse, cfbtForDo, cfbtWhileDo, cfbtWithDo]);
      Result := tkKey;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func19: TmscrTokenKind;
Var
   pas: TmscrPascalCodeFoldBlockType;
Begin
   If KeyComp('Do') Then
   Begin
      Result := tkKey;
      pas := TopPascalCodeFoldBlockType;
      If pas In [cfbtForDo, cfbtWhileDo, cfbtWithDo] Then
      Begin
         EndPascalCodeFoldBlock();
         StartPascalCodeFoldBlock(pas);
      End;
   End
   Else If KeyComp('And') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func20: TmscrTokenKind;
Begin
   If KeyComp('As') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func21: TmscrTokenKind;
Begin
   If KeyComp('Of') Then
   Begin
      Result := tkKey;
      If (PasCodeFoldRange.BracketNestLevel = 0) Then
      Begin
         // Accidental start of block // End at next semicolon (usually same line)
         fRange := fRange + [rsSkipAllPasBlocks];
         //CodeFoldRange.Pop(false); // avoid minlevel // does not work, still minlevel for disabled
         //CodeFoldRange.Add(Pointer(PtrInt(cfbtUses)), false);
      End
      Else If (TopPascalCodeFoldBlockType = cfbtCase) Then
      Begin
         EndPascalCodeFoldBlock();
         StartPascalCodeFoldBlock(cfbtCase, True);
         fRange := fRange + [rsAtCaseLabel];
      End;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func23: TmscrTokenKind;
Var
   tfb: TmscrPascalCodeFoldBlockType;
   sl: integer;
Begin
   If KeyComp('End') Then
   Begin
      If ((fToIdent < 2) Or (fLine[fToIdent - 1] <> '@')) Then
      Begin
         Result := tkKey;

         PasCodeFoldRange.BracketNestLevel := 0; // Reset in case of partial code
         sl := fStringLen;
         // there may be more than on block ending here
         tfb := TopPascalCodeFoldBlockType;
         fStringLen := 0;
         While (tfb In [cfbtIfThen, cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfElse]) Do
         Begin // no semicolon before end
            EndPascalCodeFoldBlock(True);
            tfb := TopPascalCodeFoldBlockType;
         End;
         fStringLen := sl;
         If tfb = cfbtUnit Then
         Begin
            EndPascalCodeFoldBlock;
         End
         Else If tfb = cfbtExcept Then
         Begin
            EndPascalCodeFoldBlock;
            If TopPascalCodeFoldBlockType = cfbtTry Then
               EndPascalCodeFoldBlock;
         End
         Else If tfb = cfbtTry Then
         Begin
            EndPascalCodeFoldBlock;

            If TopPascalCodeFoldBlockType In [cfbtProcedure] Then
               EndPascalCodeFoldBlock;
         End
         Else If tfb In [cfbtCaseElse] Then
         Begin
            EndPascalCodeFoldBlock;
            EndPascalCodeFoldBlock; // must be cfbtCase
            fRange := fRange - [rsAtCaseLabel];
         End
         Else If tfb In [cfbtCase] Then
         Begin
            EndPascalCodeFoldBlock;
            fRange := fRange - [rsAtCaseLabel];
         End
         Else If tfb In [cfbtBeginEnd] Then
         Begin
            EndPascalCodeFoldBlock;
            If TopPascalCodeFoldBlockType = cfbtProgram Then
               EndPascalCodeFoldBlock;
            fStringLen := 0;
            While (TopPascalCodeFoldBlockType In [{cfbtIfThen,} cfbtIfElse, cfbtForDo, cfbtWhileDo, cfbtWithDo]) Do
            Begin // no semicolon after end
               EndPascalCodeFoldBlock(True);
            End;
            fStringLen := sl;
         End
         Else If tfb = cfbtProcedure Then
         Begin
            //        EndPascalCodeFoldBlock; // wrong source: procedure end, without begin
         End
         Else If tfb = cfbtUnitSection Then
         Begin
            EndPascalCodeFoldBlockLastLine;
            If TopPascalCodeFoldBlockType = cfbtUnit Then // "Unit".."end."
               EndPascalCodeFoldBlock;
         End
         Else
         Begin

            // After type declaration, allow "deprecated"?
            If TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType] Then
               fRange := fRange + [rsVarTypeInSpecification];
         End;
      End
      Else
      Begin
         Result := tkKey; // @@end or @end label
      End;
   End
   Else If KeyComp('In') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func25: TmscrTokenKind;
Begin
   If KeyComp('Far') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func28: TmscrTokenKind;
Begin
   If KeyComp('Is') Then
      Result := tkKey
   Else If KeyComp('Case') Then
   Begin
      If TopPascalCodeFoldBlockType In PascalStatementBlocks + [cfbtUnitSection] Then
         StartPascalCodeFoldBlock(cfbtCase, True);
      Result := tkKey;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func29: TmscrTokenKind;
Begin
   If KeyComp('On') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func32: TmscrTokenKind;
Begin
   If KeyComp('Label') Then
   Begin
      If (TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType, cfbtNone, cfbtProcedure, cfbtProgram, cfbtUnit, cfbtUnitSection]) Then
      Begin
         If TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType] Then
            EndPascalCodeFoldBlockLastLine;
         If TopPascalCodeFoldBlockType In [cfbtProcedure] Then
            StartPascalCodeFoldBlock(cfbtLocalVarType)
         Else
            StartPascalCodeFoldBlock(cfbtVarType);
      End;
      Result := tkKey;
   End
   Else If KeyComp('Mod') Then
      Result := tkKey
   Else If KeyComp('File') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func33: TmscrTokenKind;
Begin
   If KeyComp('Or') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func35: TmscrTokenKind;
Begin
   If KeyComp('Nil') Then
      Result := tkKey
   Else If KeyComp('To') Then
      Result := tkKey
   Else If KeyComp('Div') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func37: TmscrTokenKind;
Begin
   If KeyComp('Begin') Then
   Begin

      PasCodeFoldRange.BracketNestLevel := 0; // Reset in case of partial code
      If TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType] Then
         EndPascalCodeFoldBlockLastLine;
      Result := tkKey;
      If TopPascalCodeFoldBlockType In [cfbtProcedure] Then
         StartPascalCodeFoldBlock(cfbtTopBeginEnd)
      Else
         StartPascalCodeFoldBlock(cfbtBeginEnd);
      //debugln('TmscrScriptHighlighter.Func37 BEGIN ',dbgs(ord(TopPascalCodeFoldBlockType)),' LineNumber=',dbgs(fLineNumber),' ',dbgs(MinimumNestFoldBlockLevel),' ',dbgs(CurrentCodeFoldBlockLevel));
   End
   Else If FExtendedKeywordsMode And KeyComp('Break') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func38: TmscrTokenKind;
Begin
   If KeyComp('Near') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func39: TmscrTokenKind;
Begin
   If KeyComp('For') Then
   Begin
      Result := tkKey;
      If TopPascalCodeFoldBlockType In PascalStatementBlocks Then
         StartPascalCodeFoldBlock(cfbtForDo);
   End
   Else If KeyComp('Shl') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func41: TmscrTokenKind;
Begin
   If KeyComp('Else') Then
   Begin
      Result := tkKey;
      // close all parent "else" and "do" // there can only be one else
      While (TopPascalCodeFoldBlockType In [cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfElse]) Do
      Begin
         //DebugLn('    Ending: %s', [dbgs(TopPascalCodeFoldBlockType)]);
         EndPascalCodeFoldBlockLastLine;
      End;
      If (TopPascalCodeFoldBlockType In [cfbtIfThen]) Then
      Begin
         EndPascalCodeFoldBlock;
         StartPascalCodeFoldBlock(cfbtIfElse);
      End
      Else If TopPascalCodeFoldBlockType = cfbtCase Then
      Begin
         FTokenIsCaseLabel := True;
         StartPascalCodeFoldBlock(cfbtCaseElse, True);
      End;
   End
   Else If KeyComp('Var') Then
   Begin
      If (PasCodeFoldRange.BracketNestLevel = 0) And (TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType, cfbtNone, cfbtProcedure, cfbtProgram, cfbtUnit, cfbtUnitSection]) Then
      Begin
         If TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType] Then
            EndPascalCodeFoldBlockLastLine;
         If TopPascalCodeFoldBlockType In [cfbtProcedure] Then
            StartPascalCodeFoldBlock(cfbtLocalVarType)
         Else
            StartPascalCodeFoldBlock(cfbtVarType);
      End;
      Result := tkKey;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func44: TmscrTokenKind;
Begin
   If KeyComp('Set') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func45: TmscrTokenKind;
Begin
   If KeyComp('Shr') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func47: TmscrTokenKind;
Begin
   If KeyComp('Then') Then
   Begin
      Result := tkKey;

      If (TopPascalCodeFoldBlockType = cfbtIfThen) Then
         EndPascalCodeFoldBlock;
      StartPascalCodeFoldBlock(cfbtIfThen,
         TopPascalCodeFoldBlockType In [cfbtCase, cfbtIfThen, cfbtIfElse, cfbtForDo, cfbtWhileDo, cfbtWithDo]);
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func49: TmscrTokenKind;
Begin
   If KeyComp('Not') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func52: TmscrTokenKind;
Begin
   If KeyComp('Pascal') And (TopPascalCodeFoldBlockType In ProcModifierAllowed) Then
      Result := tkKey
   Else If KeyComp('Raise') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func57: TmscrTokenKind;
Begin
   If KeyComp('Goto') Then
      Result := tkKey
   Else If KeyComp('While') Then
   Begin
      Result := tkKey;
      StartPascalCodeFoldBlock(cfbtWhileDo);
   End
   Else If KeyComp('Xor') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func58: TmscrTokenKind;
Begin
   If FExtendedKeywordsMode And KeyComp('Exit') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func59: TmscrTokenKind;
Begin
   If KeyComp('Safecall') And (TopPascalCodeFoldBlockType In ProcModifierAllowed) Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func60: TmscrTokenKind;
Begin
   If KeyComp('With') Then
   Begin
      Result := tkKey;
      StartPascalCodeFoldBlock(cfbtWithDo);
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func63: TmscrTokenKind;
Begin
   If KeyComp('Array') Then
      Result := tkKey
   Else If KeyComp('Try') Then
   Begin
      If TopPascalCodeFoldBlockType In PascalStatementBlocks + [cfbtUnitSection] Then
         StartPascalCodeFoldBlock(cfbtTry);
      Result := tkKey;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func64: TmscrTokenKind;
Begin
   If KeyComp('Unit') Then
   Begin
      If TopPascalCodeFoldBlockType = cfbtNone Then StartPascalCodeFoldBlock(cfbtUnit);
      Result := tkKey;
   End
   Else If KeyComp('Uses') Then
   Begin
      If (TopPascalCodeFoldBlockType In [cfbtNone, cfbtProgram, cfbtUnit, cfbtUnitSection]) Then
      Begin
         StartPascalCodeFoldBlock(cfbtUses);
      End;
      Result := tkKey;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func65: TmscrTokenKind;
Begin
   If KeyComp('Repeat') Then
   Begin
      Result := tkKey;
      StartPascalCodeFoldBlock(cfbtRepeat);
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func71: TmscrTokenKind;
Begin
   If KeyComp('Stdcall') And (TopPascalCodeFoldBlockType In ProcModifierAllowed) Then
      Result := tkKey
   Else If KeyComp('Const') Then
   Begin
      If (PasCodeFoldRange.BracketNestLevel = 0) And (TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType, cfbtNone, cfbtProcedure, cfbtProgram, cfbtUnit, cfbtUnitSection]) Then
      Begin
         If TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType] Then
            EndPascalCodeFoldBlockLastLine;
         If TopPascalCodeFoldBlockType In [cfbtProcedure] Then
            StartPascalCodeFoldBlock(cfbtLocalVarType)
         Else
            StartPascalCodeFoldBlock(cfbtVarType);
      End;
      Result := tkKey;
   End
   Else If KeyComp('Bitpacked') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func73: TmscrTokenKind;
Begin
   If KeyComp('Except') Then
   Begin
      Result := tkKey;
      While (TopPascalCodeFoldBlockType In [cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfThen, cfbtIfElse]) Do // no semicolon before except
         EndPascalCodeFoldBlock(True);
      SmartCloseBeginEndBlocks(cfbtTry);
      If TopPascalCodeFoldBlockType = cfbtTry Then
         StartPascalCodeFoldBlock(cfbtExcept);
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func76: TmscrTokenKind;
Begin
   If KeyComp('Until') Then
   Begin
      Result := tkKey;
      While (TopPascalCodeFoldBlockType In [cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfThen, cfbtIfElse]) Do // no semicolon before until
         EndPascalCodeFoldBlock(True);
      SmartCloseBeginEndBlocks(cfbtRepeat);
      If TopPascalCodeFoldBlockType = cfbtRepeat Then EndPascalCodeFoldBlock;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func79: TmscrTokenKind;
Begin
   If KeyComp('Finally') Then
   Begin
      Result := tkKey;
      While (TopPascalCodeFoldBlockType In [cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfThen, cfbtIfElse]) Do // no semicolon before finally
         EndPascalCodeFoldBlock(True);
      SmartCloseBeginEndBlocks(cfbtTry);
      If TopPascalCodeFoldBlockType = cfbtTry Then
         StartPascalCodeFoldBlock(cfbtExcept);
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func85: TmscrTokenKind;
Begin
   If KeyComp('Forward') Then
   Begin
      Result := tkKey;
      If TopPascalCodeFoldBlockType = cfbtProcedure Then
      Begin
         EndPascalCodeFoldBlock(True);
      End;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func86: TmscrTokenKind;
Begin
   If KeyComp('VarArgs') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func87: TmscrTokenKind;
Begin
   If (FStringKeywordMode In [spsmDefault, spsmStringOnly]) And KeyComp('String') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func88: TmscrTokenKind;
Begin
   If KeyComp('Program') Then
   Begin

      If TopPascalCodeFoldBlockType = cfbtNone Then
         StartPascalCodeFoldBlock(cfbtProgram);
      Result := tkKey;
   End
   Else If KeyComp('Mwpascal') And (TopPascalCodeFoldBlockType In ProcModifierAllowed) Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func91: TmscrTokenKind;
Begin
   If KeyComp('Downto') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func97: TmscrTokenKind;
Begin
   If KeyComp('Threadvar') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func99: TmscrTokenKind;
Begin
   If KeyComp('External') Then
   Begin
      Result := tkKey;
      If TopPascalCodeFoldBlockType = cfbtProcedure Then
      Begin
         EndPascalCodeFoldBlock(True);
      End;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func100: TmscrTokenKind;
Begin
   If KeyComp('Automated') Then
      Result := tkKey
   Else If (rsInProcHeader In fRange) And KeyComp('constref') And (PasCodeFoldRange.BracketNestLevel = 1) Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func101: TmscrTokenKind;
Var
   tbf: TmscrPascalCodeFoldBlockType;
Begin
   If KeyComp('Register') And (TopPascalCodeFoldBlockType In ProcModifierAllowed) Then
      Result := tkKey
   Else If KeyComp('Platform') Then
   Begin
      tbf := TopPascalCodeFoldBlockType;
   End
   Else If FExtendedKeywordsMode And KeyComp('Continue') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func102: TmscrTokenKind;
Var
   InClass: boolean;
Begin
   If KeyComp('Function') Then
   Begin
      If Not (rsAfterEqualOrColon In fRange) Then
      Begin
         PasCodeFoldRange.BracketNestLevel := 0; // Reset in case of partial code
         CloseBeginEndBlocksBeforeProc;

         If TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType] Then
            EndPascalCodeFoldBlockLastLine;
      End;
      fRange := fRange + [rsInProcHeader];
      Result := tkKey;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func105: TmscrTokenKind;
Var
   InClass: boolean;
Begin
   If KeyComp('Procedure') Then
   Begin
      If Not (rsAfterEqualOrColon In fRange) Then
      Begin
         PasCodeFoldRange.BracketNestLevel := 0; // Reset in case of partial code
         CloseBeginEndBlocksBeforeProc;

         If TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType] Then
            EndPascalCodeFoldBlockLastLine;
      End;
      fRange := fRange + [rsInProcHeader];
      Result := tkKey;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func122: TmscrTokenKind;
Begin
   If KeyComp('Otherwise') Then
   Begin
      Result := tkKey;

      While (TopPascalCodeFoldBlockType In [cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfThen, cfbtIfElse]) Do
      Begin

         EndPascalCodeFoldBlockLastLine;
      End;
      If TopPascalCodeFoldBlockType = cfbtCase Then
      Begin
         StartPascalCodeFoldBlock(cfbtCaseElse, True);
         FTokenIsCaseLabel := True;
      End;
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func125: TmscrTokenKind;
Begin
   If KeyComp('NoReturn') And (TopPascalCodeFoldBlockType In ProcModifierAllowed) Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func128: TmscrTokenKind;
Begin
   If (FStringKeywordMode In [spsmDefault]) And KeyComp('Widestring') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func130: TmscrTokenKind;
Begin
   If (FStringKeywordMode In [spsmDefault]) And KeyComp('Ansistring') Then
      Result := tkKey
   Else If KeyComp('Enumerator') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func139: TmscrTokenKind;
Begin
   If KeyComp('WeakExternal') Then
   Begin
      Result := tkKey;
      If TopPascalCodeFoldBlockType = cfbtProcedure Then
         EndPascalCodeFoldBlock(True);
   End
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func158: TmscrTokenKind;
Begin
   If (FStringKeywordMode In [spsmDefault]) And KeyComp('UnicodeString') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func167: TmscrTokenKind;
Begin
   If (FStringKeywordMode In [spsmDefault]) And KeyComp('Shortstring') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func170: TmscrTokenKind;
Begin
   If (FStringKeywordMode In [spsmDefault]) And KeyComp('UTF8String') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.Func181: TmscrTokenKind;
Begin
   If (FStringKeywordMode In [spsmDefault]) And KeyComp('RawByteString') Then
      Result := tkKey
   Else
      Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.AltFunc: TmscrTokenKind;
Begin
   Result := tkIdentifier;
End;

Function TmscrScriptHighlighter.IdentKind(p: integer): TmscrTokenKind;
Var
   HashKey: integer;
Begin
   fToIdent := p;
   HashKey := KeyHash;
   If HashKey <= High(fIdentFuncTable) Then
      Result := fIdentFuncTable[HashKey]()
   Else
      Result := tkIdentifier;
End;

Procedure TmscrScriptHighlighter.MakeMethodTables;
Var
   I: char;
Begin
   For I := #0 To #255 Do
      Case I Of
         #0: fProcTable[I] := @NullProc;
         #10: fProcTable[I] := @LFProc;
         #13: fProcTable[I] := @CRProc;
         #1..#9, #11, #12, #14..#32:
            fProcTable[I] := @SpaceProc;
         '#': fProcTable[I] := @AsciiCharProc;
         '$': fProcTable[I] := @HexProc;
         '%': fProcTable[I] := @BinaryProc;
         '&': fProcTable[I] := @OctalProc;
         #39: fProcTable[I] := @StringProc;
         '0'..'9': fProcTable[I] := @NumberProc;
         'A'..'Z', 'a'..'z', '_':
            fProcTable[I] := @IdentProc;
         '^': fProcTable[I] := @CaretProc;
         '{': fProcTable[I] := @BraceOpenProc;
         '}', '!', '"', '('..'/', ':'..'@', '[', ']', '\', '`', '~':
            Begin
               Case I Of
                  '(': fProcTable[I] := @RoundOpenProc;
                  ')': fProcTable[I] := @RoundCloseProc;
                  '[': fProcTable[I] := @SquareOpenProc;
                  ']': fProcTable[I] := @SquareCloseProc;
                  '=': fProcTable[I] := @EqualSignProc;
                  '.': fProcTable[I] := @PointProc;
                  ';': fProcTable[I] := @SemicolonProc; //mh 2000-10-08
                  '/': fProcTable[I] := @SlashProc;
                  ':': fProcTable[I] := @ColonProc;
                  '>': fProcTable[I] := @GreaterProc;
                  '<': fProcTable[I] := @LowerProc;
                  '@': fProcTable[I] := @AddressOpProc;
                  Else
                     fProcTable[I] := @SymbolProc;
               End;
            End;
         Else
            fProcTable[I] := @UnknownProc;
      End;
End;

Constructor TmscrScriptHighlighter.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   FStringKeywordMode := spsmDefault;
   FExtendedKeywordsMode := False;
   CreateDividerDrawConfig;
   fD4syntax := True;
   fCommentAttri := TSynHighlighterAttributes.Create(@SYNS_AttrComment, SYNS_XML_AttrComment);
   fCommentAttri.Style := [fsItalic];
   AddAttribute(fCommentAttri);
   FIDEDirectiveAttri := TSynHighlighterAttributesModifier.Create(@SYNS_AttrIDEDirective, SYNS_XML_AttrIDEDirective);
   AddAttribute(FIDEDirectiveAttri);
   // FCurIDEDirectiveAttri, FCurCaseLabelAttri, FCurProcedureHeaderNameAttr
   // They are not available through the "Attribute" property (not added via AddAttribute
   // But they are returned via GetTokenAttribute, so they should have a name.
   FCurIDEDirectiveAttri := TSynSelectedColorMergeResult.Create(@SYNS_AttrIDEDirective, SYNS_XML_AttrIDEDirective);
   fIdentifierAttri := TSynHighlighterAttributes.Create(@SYNS_AttrIdentifier, SYNS_XML_AttrIdentifier);
   AddAttribute(fIdentifierAttri);
   fKeyAttri := TSynHighlighterAttributes.Create(@SYNS_AttrReservedWord, SYNS_XML_AttrReservedWord);
   fKeyAttri.Style := [fsBold];
   AddAttribute(fKeyAttri);
   fNumberAttri := TSynHighlighterAttributes.Create(@SYNS_AttrNumber, SYNS_XML_AttrNumber);
   AddAttribute(fNumberAttri);
   fSpaceAttri := TSynHighlighterAttributes.Create(@SYNS_AttrSpace, SYNS_XML_AttrSpace);
   AddAttribute(fSpaceAttri);
   fStringAttri := TSynHighlighterAttributes.Create(@SYNS_AttrString, SYNS_XML_AttrString);
   AddAttribute(fStringAttri);
   fSymbolAttri := TSynHighlighterAttributes.Create(@SYNS_AttrSymbol, SYNS_XML_AttrSymbol);
   AddAttribute(fSymbolAttri);
   FProcedureHeaderNameAttr := TSynHighlighterAttributesModifier.Create(@SYNS_AttrProcedureHeaderName, SYNS_XML_AttrProcedureHeaderName);
   AddAttribute(FProcedureHeaderNameAttr);
   FCaseLabelAttri := TSynHighlighterAttributesModifier.Create(@SYNS_AttrCaseLabel, SYNS_XML_AttrCaseLabel);
   AddAttribute(FCaseLabelAttri);
   FCurCaseLabelAttri := TSynSelectedColorMergeResult.Create(@SYNS_AttrCaseLabel, SYNS_XML_AttrCaseLabel);
   FCurProcedureHeaderNameAttr := TSynSelectedColorMergeResult.Create(@SYNS_AttrProcedureHeaderName, SYNS_XML_AttrProcedureHeaderName);
   fDirectiveAttri := TSynHighlighterAttributes.Create(@SYNS_AttrDirective, SYNS_XML_AttrDirective);
   fDirectiveAttri.Style := [fsItalic];
   AddAttribute(fDirectiveAttri);
   CompilerMode := pcmDelphi;
   SetAttributesOnChange(@DefHighlightChange);

   InitIdent;
   MakeMethodTables;
   fRange := [];
   fAsmStart := False;
   fDefaultFilter := SYNS_FilterPascal;
End; { Create }

Destructor TmscrScriptHighlighter.Destroy;
Begin
   DestroyDividerDrawConfig;
   FreeAndNil(FCurCaseLabelAttri);
   FreeAndNil(FCurIDEDirectiveAttri);
   FreeAndNil(FCurProcedureHeaderNameAttr);
   Inherited Destroy;
End;

Procedure TmscrScriptHighlighter.SetLine(Const NewValue: String; LineNumber: integer);
Begin
   //DebugLn(['TmscrScriptHighlighter.SetLine START LineNumber=',LineNumber,' Line="',NewValue,'"']);
   fLineStr := NewValue;
   fLineLen := length(fLineStr);
   fLine := PChar(Pointer(fLineStr));
   Run := 0;
   Inherited SetLine(NewValue, LineNumber);
   PasCodeFoldRange.LastLineCodeFoldLevelFix := 0;
   PasCodeFoldRange.PasFoldFixLevel := 0;
   FStartCodeFoldBlockLevel := PasCodeFoldRange.MinimumNestFoldBlockLevel;
   FPasStartLevel := PasCodeFoldRange.MinimumCodeFoldBlockLevel;
   FSynPasRangeInfo.MinLevelIfDef := FSynPasRangeInfo.EndLevelIfDef;
   FSynPasRangeInfo.MinLevelRegion := FSynPasRangeInfo.EndLevelRegion;
   fLineNumber := LineNumber;
   FAtLineStart := True;
   If Not IsCollectingNodeInfo Then
      Next;
End; { SetLine }

Procedure TmscrScriptHighlighter.SetNestedComments(Const ANestedComments: boolean);
Begin
   If FNestedComments = ANestedComments Then Exit;
   FNestedComments := ANestedComments;
   PasCodeFoldRange.NestedComments := FNestedComments;
End;

Procedure TmscrScriptHighlighter.AddressOpProc;
Begin
   fTokenID := tkSymbol;
   Inc(Run);
   If fLine[Run] = '@' Then Inc(Run);
End;

Procedure TmscrScriptHighlighter.AsciiCharProc;
Begin
   fTokenID := tkString;
   Inc(Run);
   Case FLine[Run] Of
      '%':
         Begin
            Inc(Run);
            If (FLine[Run] In ['0'..'1']) Then
               While (FLine[Run] In ['0'..'1']) Do
                  Inc(Run)
            Else
               fTokenID := tkSymbol;
         End;
      '&':
         Begin
            Inc(Run);
            If (FLine[Run] In ['0'..'7']) Then
               While (FLine[Run] In ['0'..'7']) Do
                  Inc(Run)
            Else
               fTokenID := tkSymbol;
         End;
      '$':
         Begin
            Inc(Run);
            If (IsIntegerChar[FLine[Run]]) Then
               While (IsIntegerChar[FLine[Run]]) Do
                  Inc(Run)
            Else
               fTokenID := tkSymbol;
         End;
      '0'..'9': While (FLine[Run] In ['0'..'9']) Do
            Inc(Run);
      Else
         fTokenID := tkSymbol;
   End;
End;

Procedure TmscrScriptHighlighter.BorProc;
Var
   p: longint;
Begin
   p := Run;
   fTokenID := tkComment;
   If rsIDEDirective In fRange Then
      fTokenID := tkIDEDirective;
   Repeat
      Case fLine[p] Of
         #0, #10, #13: break;
         '}':
            If TopPascalCodeFoldBlockType = cfbtNestedComment Then
            Begin
               Run := p;
               EndPascalCodeFoldBlock;
               p := Run;
            End
            Else
            Begin
               fRange := fRange - [rsBor, rsIDEDirective];
               Inc(p);
               If TopPascalCodeFoldBlockType = cfbtBorCommand Then
                  EndPascalCodeFoldBlock;
               break;
            End;
         '{':
            If NestedComments Then
            Begin
               fStringLen := 1;
               Run := p;
               StartPascalCodeFoldBlock(cfbtNestedComment);
               p := Run;
            End;
      End;
      Inc(p);
   Until (p >= fLineLen);
   Run := p;
End;

Procedure TmscrScriptHighlighter.DirectiveProc;
Begin
   fTokenID := tkDirective;
   If TextComp('modeswitch') Then
   Begin
      // modeswitch directive
      Inc(Run, 10);
      // skip space
      While (fLine[Run] In [' ', #9, #10, #13]) Do
         Inc(Run);
      If TextComp('nestedcomments') Then
      Begin
         Inc(Run, 14);
         // skip space
         While (fLine[Run] In [' ', #9, #10, #13]) Do
            Inc(Run);
         If fLine[Run] In ['+', '}'] Then
            NestedComments := True
         Else If fLine[Run] = '-' Then
            NestedComments := False;
      End;
      If TextComp('typehelpers') Then
      Begin
         Inc(Run, 11);
         // skip space
         While (fLine[Run] In [' ', #9, #10, #13]) Do
            Inc(Run);
         If fLine[Run] In ['+', '}'] Then
            TypeHelpers := True
         Else If fLine[Run] = '-' Then
            TypeHelpers := False;
      End;
   End;
   If TextComp('mode') Then
   Begin
      // $mode directive
      Inc(Run, 4);
      // skip space
      While (fLine[Run] In [' ', #9, #10, #13]) Do
         Inc(Run);
      If TextComp('objfpc') Then
         CompilerMode := pcmObjFPC
      Else If TextComp('delphi') Then
         CompilerMode := pcmDelphi
      Else If TextComp('fpc') Then
         CompilerMode := pcmFPC
      Else If TextComp('gpc') Then
         CompilerMode := pcmGPC
      Else If TextComp('tp') Then
         CompilerMode := pcmTP
      Else If TextComp('macpas') Then
         CompilerMode := pcmMacPas;
   End;
   Repeat
      Case fLine[Run] Of
         #0, #10, #13: break;
         '}':
            If TopPascalCodeFoldBlockType = cfbtNestedComment Then
               EndPascalCodeFoldBlock
            Else
            Begin
               fRange := fRange - [rsDirective];
               Inc(Run);
               break;
            End;
         '{':
            If NestedComments Then
            Begin
               fStringLen := 1;
               StartPascalCodeFoldBlock(cfbtNestedComment);
            End;
      End;
      Inc(Run);
   Until (Run >= fLineLen);
   //DebugLn(['TmscrScriptHighlighter.DirectiveProc Run=',Run,' fTokenPos=',fTokenPos,' fLineStr=',fLineStr,' Token=',GetToken]);
End;

Procedure TmscrScriptHighlighter.BraceOpenProc;

   Function ScanRegion: boolean;
   Var
      Txt: String;
      Idx, NestBrace, i, l: integer;
      InString: boolean;
   Begin
      Result := False;
      Txt := copy(fLine, Run, length(fLine));
      Idx := LineIndex;
      InString := False;
      NestBrace := 0;
      While True Do
      Begin
         i := 1;
         l := length(Txt);
         While i <= l Do
         Begin
            Case Txt[i] Of
               '{': Inc(NestBrace);
               '}': If NestBrace = 0 Then
                     exit
                  Else
                     Dec(NestBrace);
               '''': If (i + 1 <= l) And (Txt[i + 1] = '''') Then
                     Inc(i)
                  Else
                     InString := Not InString;
               '-', '/': If (Not InString) And (i + 4 <= l) And ((i = 1) Or (Txt[i - 1] In [' ', #9, #10, #13])) And (KeyCompEx(@Txt[i + 1], PChar('fold'), 4)) And
                  ((i + 4 = l) Or (Txt[i + 5] In [' ', #9, #10, #13, '}'])) Then
                     exit(True);
            End;
            Inc(i);
         End;
         Inc(Idx);
         If Idx < CurrentLines.Count Then
            Txt := CurrentLines[Idx]
         Else
            break;
      End;
   End;

   Procedure StartDirectiveFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType);
      Inline;
   Begin
      Dec(Run);
      Inc(fStringLen); // include $
      StartCustomCodeFoldBlock(ABlockType);
      Inc(Run);
   End;

   Procedure EndDirectiveFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType);
      Inline;
   Begin
      Dec(Run);
      Inc(fStringLen); // include $
      EndCustomCodeFoldBlock(ABlockType);
      Inc(Run);
   End;

   Procedure EndStartDirectiveFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType);
      Inline;
   Begin
      Dec(Run);
      Inc(fStringLen); // include $
      EndCustomCodeFoldBlock(ABlockType);
      StartCustomCodeFoldBlock(ABlockType);
      Inc(Run);
   End;

Var
   nd: PSynFoldNodeInfo;
Begin
   If (Run < fLineLen - 1) And (fLine[Run + 1] = '$') Then
   Begin
      // compiler directive
      fRange := fRange + [rsDirective];
      Inc(Run, 2);
      fToIdent := Run;
      KeyHash;
      If (fLine[Run] In ['i', 'I']) And (KeyComp('if') Or KeyComp('ifc') Or KeyComp('ifdef') Or KeyComp('ifndef') Or KeyComp('ifopt')) Then
         StartDirectiveFoldBlock(cfbtIfDef)
      Else If ((fLine[Run] In ['e', 'E']) And (KeyComp('endif') Or KeyComp('endc'))) Or KeyComp('ifend') Then
         EndDirectiveFoldBlock(cfbtIfDef)
      Else If (fLine[Run] In ['e', 'E']) And (KeyComp('else') Or KeyComp('elsec') Or KeyComp('elseif') Or KeyComp('elifc')) Then
         EndStartDirectiveFoldBlock(cfbtIfDef)
      Else If KeyComp('region') Then
      Begin
         StartDirectiveFoldBlock(cfbtRegion);
         If IsCollectingNodeInfo Then
            // Scan ahead
            If ScanRegion Then
            Begin
               nd := CollectingNodeInfoList.LastItemPointer;
               If nd <> Nil Then
                  nd^.FoldAction := nd^.FoldAction + [sfaDefaultCollapsed];
            End;
      End
      Else If KeyComp('endregion') Then
         EndDirectiveFoldBlock(cfbtRegion);
      DirectiveProc;
   End
   Else
   Begin
      // curly bracket open -> borland comment
      fStringLen := 1; // length of "{"
      Inc(Run);
      If (Run < fLineLen) And (fLine[Run] = '%') Then
      Begin
         fRange := fRange + [rsIDEDirective];
         // IDE directive {%xxx } rsIDEDirective
         Inc(Run);
         fToIdent := Run;
         KeyHash;
         If KeyComp('region') Then
         Begin
            StartDirectiveFoldBlock(cfbtRegion);
            If IsCollectingNodeInfo Then
               // Scan ahead
               If ScanRegion Then
               Begin
                  nd := CollectingNodeInfoList.LastItemPointer;
                  If nd <> Nil Then
                     nd^.FoldAction := nd^.FoldAction + [sfaDefaultCollapsed];
               End;
         End
         Else If KeyComp('endregion') Then
            EndDirectiveFoldBlock(cfbtRegion)
         Else
         Begin
            Dec(Run, 2);
            StartPascalCodeFoldBlock(cfbtBorCommand);
            Inc(Run);
         End;
      End
      Else
      Begin
         fRange := fRange + [rsBor];
         Dec(Run);
         StartPascalCodeFoldBlock(cfbtBorCommand);
         Inc(Run);
      End;
      BorProc;
   End;
End;

Procedure TmscrScriptHighlighter.ColonProc;
Begin
   fTokenID := tkSymbol;
   Inc(Run);
   If fLine[Run] = '=' Then
      Inc(Run) // ":="
   Else
   Begin
      fRange := fRange + [rsAfterEqualOrColon] - [rsAtCaseLabel];
      If (TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType]) Then
         fRange := fRange + [rsVarTypeInSpecification];
   End;
End;

Procedure TmscrScriptHighlighter.GreaterProc;
Begin
   fTokenID := tkSymbol;
   Inc(Run);
   If fLine[Run] = '=' Then
      Inc(Run);
End;

Procedure TmscrScriptHighlighter.CRProc;
Begin
   fTokenID := tkSpace;
   Inc(Run);
   If fLine[Run] = #10 Then Inc(Run);
End;

Procedure TmscrScriptHighlighter.IdentProc;
Begin
   fTokenID := IdentKind(Run);
   Inc(Run, fStringLen);
   While Identifiers[fLine[Run]] Do
      Inc(Run);
End;

Procedure TmscrScriptHighlighter.HexProc;
Begin
   Inc(Run);
   If (IsIntegerChar[FLine[Run]]) Then
   Begin
      fTokenID := tkNumber;
      While (IsIntegerChar[FLine[Run]]) Do
         Inc(Run);
   End
   Else
      fTokenID := tkSymbol;
End;

Procedure TmscrScriptHighlighter.BinaryProc;
Begin
   Inc(Run);
   If FLine[Run] In ['0'..'1'] Then
   Begin
      fTokenID := tkNumber;
      While FLine[Run] In ['0'..'1'] Do
         Inc(Run);
   End
   Else
      fTokenID := tkSymbol;
End;

Procedure TmscrScriptHighlighter.OctalProc;
Begin
   Inc(Run);
   If FLine[Run] In ['0'..'7'] Then
   Begin
      fTokenID := tkNumber;
      While FLine[Run] In ['0'..'7'] Do
         Inc(Run);
   End
   Else If FLine[Run] In ['A'..'Z', 'a'..'z', '_'] Then
   Begin
      fTokenID := tkIdentifier;
      While Identifiers[fLine[Run]] Do
         Inc(Run);
   End
   Else
      fTokenID := tkSymbol;
End;

Procedure TmscrScriptHighlighter.LFProc;
Begin
   fTokenID := tkSpace;
   Inc(Run);
End;

Procedure TmscrScriptHighlighter.LowerProc;
Begin
   fTokenID := tkSymbol;
   Inc(Run);
   If fLine[Run] In ['=', '>'] Then Inc(Run);
End;

Procedure TmscrScriptHighlighter.CaretProc;
Var
   t: TmscrPascalCodeFoldBlockType;
Begin
   Inc(Run);
   fTokenID := tkSymbol;

   t := TopPascalCodeFoldBlockType;
   If ((t In PascalStatementBlocks) Or //cfbtClass, cfbtClassSection,
      (((t In [cfbtVarType, cfbtLocalVarType]) Or ((t In [cfbtProcedure]) And (PasCodeFoldRange.BracketNestLevel > 0))) And (fRange * [rsInTypeBlock, rsAfterEqual] = [rsAfterEqual]))) And
      Not (rsAfterIdentifierOrValue In fRange) Then
   Begin
      If Run < fLineLen Then
      Begin
         If (Run + 1 < fLineLen) And (fLine[Run] = '{') And (fLine[Run + 1] = '$') Then
         Begin
            // "{$" directive takes precedence
            fTokenID := tkSymbol;
            exit;
         End;
         Inc(Run);
      End;
      fTokenID := tkString;
   End
   Else
      fRange := fRange + [rsAfterIdentifierOrValueAdd];
End;

Procedure TmscrScriptHighlighter.NullProc;
Begin
   If (Run = 0) And (rsSlash In fRange) Then
   Begin
      fRange := fRange - [rsSlash];
      If TopPascalCodeFoldBlockType = cfbtSlashComment Then
         EndPascalCodeFoldBlockLastLine;
   End;
   fTokenID := tkNull;
   If Run < fLineLen Then Inc(Run);
End;

Procedure TmscrScriptHighlighter.NumberProc;
Begin
   Inc(Run);
   fTokenID := tkNumber;
   If Run < fLineLen Then
   Begin
      While (IsNumberChar[FLine[Run]]) Do
         Inc(Run);
      If (FLine[Run] = '.') And Not (fLine[Run + 1] = '.') Then
      Begin
         Inc(Run);
         While (IsNumberChar[FLine[Run]]) Do
            Inc(Run);
      End;
      If (FLine[Run] = 'e') Or (fLine[Run] = 'E') Then
      Begin
         Inc(Run);
         If (FLine[Run] = '+') Or (fLine[Run] = '-') Then Inc(Run);
         While (IsNumberChar[FLine[Run]]) Do
            Inc(Run);
      End;
   End;
End;

Procedure TmscrScriptHighlighter.PointProc;
Begin
   fTokenID := tkSymbol;
   Inc(Run);
   If fLine[Run] In ['.', ')'] Then Inc(Run);
End;

Procedure TmscrScriptHighlighter.AnsiProc;
Begin
   fTokenID := tkComment;
   Repeat
      If fLine[Run] = #0 Then
         break
      Else If (fLine[Run] = '*') And (fLine[Run + 1] = ')') Then
      Begin
         Inc(Run, 2);
         If TopPascalCodeFoldBlockType = cfbtNestedComment Then
         Begin
            EndPascalCodeFoldBlock;
         End
         Else
         Begin
            fRange := fRange - [rsAnsi];
            If TopPascalCodeFoldBlockType = cfbtAnsiComment Then
               EndPascalCodeFoldBlock;
            break;
         End;
      End
      Else If NestedComments And (fLine[Run] = '(') And (fLine[Run + 1] = '*') Then
      Begin
         fStringLen := 2;
         StartPascalCodeFoldBlock(cfbtNestedComment);
         Inc(Run, 2);
      End
      Else
         Inc(Run);
   Until (Run >= fLineLen) Or (fLine[Run] In [#0, #10, #13]);
End;

Procedure TmscrScriptHighlighter.RoundOpenProc;
Begin
   Inc(Run);
   If Run >= fLineLen Then
   Begin
      fTokenID := tkSymbol;
      PasCodeFoldRange.IncBracketNestLevel;
      exit;
   End;
   Case fLine[Run] Of
      '*':
         Begin
            // We would not be here, if we were in a comment or directive already
            fRange := fRange + [rsAnsi];
            fTokenID := tkComment;
            fStringLen := 2; // length of "(*"
            Dec(Run);
            StartPascalCodeFoldBlock(cfbtAnsiComment);
            Inc(Run, 2);
            If Not (fLine[Run] In [#0, #10, #13]) Then
            Begin
               AnsiProc;
            End;
         End;
      '.':
         Begin
            Inc(Run);
            fTokenID := tkSymbol;
            PasCodeFoldRange.IncBracketNestLevel;
         End;
      Else
         Begin
            fTokenID := tkSymbol;
            PasCodeFoldRange.IncBracketNestLevel;
         End;
   End;
End;

Procedure TmscrScriptHighlighter.RoundCloseProc;
Begin
   Inc(Run);
   fTokenID := tkSymbol;
   fRange := fRange + [rsAfterIdentifierOrValueAdd];
   PasCodeFoldRange.DecBracketNestLevel;
   If (PasCodeFoldRange.BracketNestLevel = 0) Then
   Begin

      Exclude(fRange, rsInProcHeader);
   End;
End;

Procedure TmscrScriptHighlighter.SquareOpenProc;
Begin
   Inc(Run);
   fTokenID := tkSymbol;
   PasCodeFoldRange.IncBracketNestLevel;
End;

Procedure TmscrScriptHighlighter.SquareCloseProc;
Begin
   Inc(Run);
   fTokenID := tkSymbol;
   fRange := fRange + [rsAfterIdentifierOrValueAdd];
   PasCodeFoldRange.DecBracketNestLevel;
End;

Procedure TmscrScriptHighlighter.EqualSignProc;
Begin
   Inc(Run);
   fTokenID := tkSymbol;
   fRange := fRange + [rsAfterEqualOrColon, rsAfterEqual];
   If (TopPascalCodeFoldBlockType In [cfbtVarType, cfbtLocalVarType]) Then
      fRange := fRange + [rsVarTypeInSpecification];
End;

Procedure TmscrScriptHighlighter.SemicolonProc;
Var
   tfb: TmscrPascalCodeFoldBlockType;
   InSkipBlocks: boolean;
Begin
   fTokenID := tkSymbol;
   tfb := TopPascalCodeFoldBlockType;
   InSkipBlocks := rsSkipAllPasBlocks In fRange;
   Exclude(fRange, rsSkipAllPasBlocks);

   fStringLen := 1;
   If tfb = cfbtUses Then
      EndPascalCodeFoldBlock;

   While (tfb In [cfbtIfThen, cfbtIfElse, cfbtForDo, cfbtWhileDo, cfbtWithDo]) Do
   Begin
      EndPascalCodeFoldBlock(True);
      tfb := TopPascalCodeFoldBlockType;
   End;

   Inc(Run);

   If (tfb = cfbtCase) Then
      fRange := fRange + [rsAtCaseLabel];

   If (fRange * [rsVarTypeInSpecification] = [rsVarTypeInSpecification]) Then

      If (fRange * [rsInProcHeader] <> []) And (PasCodeFoldRange.BracketNestLevel = 0) Then
         fRange := fRange - [rsInProcHeader];
   fRange := fRange - [rsVarTypeInSpecification, rsAfterEqual];
End;

Procedure TmscrScriptHighlighter.SlashProc;
Begin
   If fLine[Run + 1] = '/' Then
   Begin
      fTokenID := tkComment;
      If FAtLineStart Then
      Begin
         fRange := fRange + [rsSlash];
         fStringLen := 2; // length of "//"
         If Not (TopPascalCodeFoldBlockType = cfbtSlashComment) Then
            StartPascalCodeFoldBlock(cfbtSlashComment);
      End;
      Inc(Run, 2);
      While Not (fLine[Run] In [#0, #10, #13]) Do
         Inc(Run);
   End
   Else
   Begin
      Inc(Run);
      fTokenID := tkSymbol;
   End;
End;

Procedure TmscrScriptHighlighter.SlashContinueProc;
Begin
   If (fLine[Run] = '/') And (fLine[Run + 1] = '/') Then
   Begin
      // Continue fold block
      fTokenID := tkComment;
      While Not (fLine[Run] In [#0, #10, #13]) Do
         Inc(Run);
      exit;
   End;

   fTokenID := tkUnknown;
   If IsSpaceChar[FLine[Run]] Then
   Begin
      fTokenID := tkSpace;
      Inc(Run);
      While IsSpaceChar[FLine[Run]] Do
         Inc(Run);
   End;

   If Not ((fLine[Run] = '/') And (fLine[Run + 1] = '/')) Then
   Begin
      fRange := fRange - [rsSlash];
      If TopPascalCodeFoldBlockType = cfbtSlashComment Then
         EndPascalCodeFoldBlockLastLine;
   End;

   If FTokenID = tkUnknown Then
      Next;
End;

Procedure TmscrScriptHighlighter.SpaceProc;
Begin
   Inc(Run);
   fTokenID := tkSpace;
   While IsSpaceChar[FLine[Run]] Do
      Inc(Run);
End;

Procedure TmscrScriptHighlighter.StringProc;
Begin
   fTokenID := tkString;
   Inc(Run);
   While (Not (fLine[Run] In [#0, #10, #13])) Do
   Begin
      If fLine[Run] = '''' Then
      Begin
         Inc(Run);
         If (fLine[Run] <> '''') Then
            break;
      End;
      Inc(Run);
   End;
End;

Procedure TmscrScriptHighlighter.SymbolProc;
Begin
   Inc(Run);
   fTokenID := tkSymbol;
End;

Procedure TmscrScriptHighlighter.UnknownProc;
Begin
   Inc(Run);
   While (fLine[Run] In [#128..#191]) Or // continued utf8 subcode
   ((fLine[Run] <> #0) And (fProcTable[fLine[Run]] = @UnknownProc)) Do
      Inc(Run);
   fTokenID := tkUnknown;
End;

Procedure TmscrScriptHighlighter.Next;
Var
   IsAtCaseLabel: boolean;
Begin
   fAsmStart := False;
   fTokenPos := Run;
   FTokenIsCaseLabel := False;
   If Run >= fLineLen Then
   Begin
      NullProc;
      exit;
   End;
   Case fLine[Run] Of
      #0: NullProc;
      #10: LFProc;
      #13: CRProc;
      Else
         If rsAnsi In fRange Then
            AnsiProc
         Else If fRange * [rsBor, rsIDEDirective] <> [] Then
            BorProc
         Else If rsDirective In fRange Then
            DirectiveProc
         Else If rsSlash In fRange Then
            SlashContinueProc
         Else
         Begin
            FOldRange := fRange;
            //if rsAtEqual in fRange then
            //  fRange := fRange + [rsAfterEqualOrColon] - [rsAtEqual]
            //else

            IsAtCaseLabel := rsAtCaseLabel In fRange;

            fProcTable[fLine[Run]];

            If (IsAtCaseLabel) And (rsAtCaseLabel In fRange) Then
            Begin
               FTokenIsCaseLabel := True;
               If (FTokenID = tkKey) Then
                  fRange := fRange - [rsAtCaseLabel];
            End;

            If Not (FTokenID In [tkSpace, tkComment, tkIDEDirective, tkDirective]) Then
            Begin
               If (PasCodeFoldRange.BracketNestLevel = 0) And Not (rsAtClosingBracket In fRange) Then

                  fRange := fRange - (FOldRange * [rsAfterEqualOrColon, rsAfterIdentifierOrValue, rsAfterEqualThenType]) - [rsAtClosingBracket];

            End
            Else
            Begin
               fRange := fRange - [rsAtClosingBracket];

            End;

            If (FTokenID = tkIdentifier) Or (rsAfterIdentifierOrValueAdd In fRange) Then
               fRange := fRange + [rsAfterIdentifierOrValue] - [rsAfterIdentifierOrValueAdd];
         End
   End;
   If FAtLineStart And Not (FTokenID In [tkSpace, tkComment, tkIDEDirective]) Then
      FAtLineStart := False;
   //DebugLn(['TmscrScriptHighlighter.Next Run=',Run,' fTokenPos=',fTokenPos,' fLineStr=',fLineStr,' Token=',GetToken]);
End;

Function TmscrScriptHighlighter.GetDefaultAttribute(Index: integer): TSynHighlighterAttributes;
Begin
   Case Index Of
      SYN_ATTR_COMMENT: Result := fCommentAttri;
      SYN_ATTR_IDENTIFIER: Result := fIdentifierAttri;
      SYN_ATTR_KEYWORD: Result := fKeyAttri;
      SYN_ATTR_STRING: Result := fStringAttri;
      SYN_ATTR_WHITESPACE: Result := fSpaceAttri;
      SYN_ATTR_NUMBER: Result := fNumberAttri;
      SYN_ATTR_DIRECTIVE: Result := fDirectiveAttri;
      Else
         Result := Nil;
   End;
End;

Function TmscrScriptHighlighter.GetEol: boolean;
Begin
   Result := (fTokenID = tkNull) And (Run >= fLineLen);
End;

Function TmscrScriptHighlighter.GetToken: String;
Var
   Len: longint;
Begin
   Len := Run - fTokenPos;
   SetLength(Result, Len);
   If Len > 0 Then
      System.Move(fLine[fTokenPos], Result[1], Len);
End;

Procedure TmscrScriptHighlighter.GetTokenEx(Out TokenStart: PChar; Out TokenLength: integer);
Begin
   TokenLength := Run - fTokenPos;
   If TokenLength > 0 Then
   Begin
      TokenStart := @fLine[fTokenPos];
   End
   Else
   Begin
      TokenStart := Nil;
   End;
End;

Function TmscrScriptHighlighter.GetTokenID: TmscrTokenKind;
Begin
   Result := fTokenId;
End;

Function TmscrScriptHighlighter.GetTokenAttribute: TSynHighlighterAttributes;
Var
   tid: TmscrTokenKind;
Begin
   tid := GetTokenID;
   Case tid Of

      tkComment: Result := fCommentAttri;
      tkIDEDirective:
         Begin
            FCurIDEDirectiveAttri.Assign(FCommentAttri);
            FCurIDEDirectiveAttri.Merge(FIDEDirectiveAttri);
            Result := FCurIDEDirectiveAttri;
         End;
      tkIdentifier: Result := fIdentifierAttri;
      tkKey: Result := fKeyAttri;
      tkNumber: Result := fNumberAttri;
      tkSpace: Result := fSpaceAttri;
      tkString: Result := fStringAttri;
      tkSymbol: Result := fSymbolAttri;
      tkDirective: Result := fDirectiveAttri;
      tkUnknown: Result := fSymbolAttri;
      Else
         Result := Nil;
   End;

   If FTokenIsCaseLabel And (tid In [tkIdentifier, tkKey, tkNumber, tkString]) Then
   Begin
      FCurCaseLabelAttri.Assign(Result);
      FCurCaseLabelAttri.Merge(FCaseLabelAttri);
      Result := FCurCaseLabelAttri;
   End;

   If (tid In [tkIdentifier, tkSymbol]) And (fRange * [rsInProcHeader, rsAfterEqualOrColon, rsAfterEqual] = [rsInProcHeader]) And (FOldRange * [rsAfterEqualOrColon, rsAfterEqual] = []) And (PasCodeFoldRange.BracketNestLevel = 0) Then
   Begin
      FCurProcedureHeaderNameAttr.Assign(Result);
      FCurProcedureHeaderNameAttr.Merge(FProcedureHeaderNameAttr);
      Result := FCurProcedureHeaderNameAttr;
   End;
End;

Function TmscrScriptHighlighter.GetTokenKind: integer;
Begin
   Result := Ord(GetTokenID);
End;

Function TmscrScriptHighlighter.GetTokenPos: integer;
Begin
   Result := fTokenPos;
End;

Function TmscrScriptHighlighter.GetRange: Pointer;
Begin
   // For speed reasons, we work with fRange instead of CodeFoldRange.RangeType
   // -> update now
   CodeFoldRange.RangeType := Pointer(PtrUInt(integer(fRange)));
   PasCodeFoldRange.TypeHelpers := TypeHelpers;
   // return a fixed copy of the current CodeFoldRange instance
   Result := Inherited GetRange;
End;

Procedure TmscrScriptHighlighter.SetRange(Value: Pointer);
Begin
   //DebugLn(['TmscrScriptHighlighter.SetRange START']);
   Inherited SetRange(Value);
   CompilerMode := PasCodeFoldRange.Mode;
   NestedComments := PasCodeFoldRange.NestedComments;
   TypeHelpers := PasCodeFoldRange.TypeHelpers;
   fRange := TmscrRangeStates(integer(PtrUInt(CodeFoldRange.RangeType)));
   FSynPasRangeInfo := TSynHighlighterPasRangeList(CurrentRanges).PasRangeInfo[LineIndex - 1];
End;

Procedure TmscrScriptHighlighter.StartAtLineIndex(LineNumber: integer);
Begin
   Inherited StartAtLineIndex(LineNumber);
End;

Function TmscrScriptHighlighter.GetEndOfLineAttribute: TSynHighlighterAttributes;
Begin
   If fRange * [rsAnsi, rsBor] <> [] Then
      Result := fCommentAttri
   Else
      Result := Inherited GetEndOfLineAttribute;
End;

Procedure TmscrScriptHighlighter.ResetRange;
Begin
   fRange := [];
   FStartCodeFoldBlockLevel := 0;
   FPasStartLevel := 0;
   With FSynPasRangeInfo Do
   Begin
      EndLevelIfDef := 0;
      MinLevelIfDef := 0;
      EndLevelRegion := 0;
      MinLevelRegion := 0;
   End;
   Inherited ResetRange;
   CompilerMode := pcmDelphi;
End;

Procedure TmscrScriptHighlighter.EnumUserSettings(settings: TStrings);
Begin
   { returns the user settings that exist in the registry }
   With TRegistry.Create Do
   Begin
      Try
         RootKey := HKEY_LOCAL_MACHINE;
{$IFNDEF SYN_TYPHON}
         // ToDo Registry
         If OpenKeyReadOnly('\SOFTWARE\Borland\Delphi') Then
         Begin
            Try
               GetKeyNames(settings);
            Finally
               CloseKey;
            End;
         End;
{$ENDIF}
      Finally
         Free;
      End;
   End;
End;

Function TmscrScriptHighlighter.IsLineStartingInDirective(ALineIndex: TLineIdx): boolean;
Var
   r: Pointer;
Begin
   Result := False;
   If ALineIndex < 1 Then exit;
   r := CurrentRanges[ALineIndex - 1];
   If (r <> Nil) And (r <> NullRange) Then
      Result := rsDirective In TmscrRangeStates(integer(PtrUInt(TmscrScriptHighlighterRange(r).RangeType)));
End;

Function TmscrScriptHighlighter.FoldBlockEndLevel(ALineIndex: TLineIdx; Const AFilter: TSynFoldBlockFilter): integer;
Var
   inf: TSynPasRangeInfo;
   r, r2: Pointer;
Begin
   Assert(CurrentRanges <> Nil, 'TSynCustomFoldHighlighter.FoldBlockEndLevel requires CurrentRanges');
   Result := 0;
   inf.EndLevelIfDef := 0;
   inf.MinLevelIfDef := 0;
   inf.EndLevelRegion := 0;
   inf.MinLevelRegion := 0;
   If (ALineIndex < 0) Or (ALineIndex >= CurrentLines.Count - 1) Then
      exit;

   If AFilter.FoldGroup In [0, FOLDGROUP_REGION, FOLDGROUP_IFDEF] Then
      inf := TSynHighlighterPasRangeList(CurrentRanges).PasRangeInfo[ALineIndex];

   If AFilter.FoldGroup In [0, FOLDGROUP_PASCAL] Then
   Begin
      // All or Pascal
      r := CurrentRanges[ALineIndex];
      If (r <> Nil) And (r <> NullRange) Then
      Begin
         r2 := TmscrScriptHighlighterRange(CurrentRanges[ALineIndex + 1]);
         If sfbIncludeDisabled In AFilter.Flags Then
         Begin
            Result := TmscrScriptHighlighterRange(r).NestFoldStackSize;
            If (r2 <> Nil) And (r2 <> NullRange) Then
               Result := Result + TmscrScriptHighlighterRange(r2).LastLineCodeFoldLevelFix;
         End
         Else
         Begin
            Result := TmscrScriptHighlighterRange(r).CodeFoldStackSize;
            If (r2 <> Nil) And (r2 <> NullRange) Then
               Result := Result + TmscrScriptHighlighterRange(r2).PasFoldFixLevel;
         End;
      End;
   End;

   If AFilter.FoldGroup In [0, FOLDGROUP_REGION] Then
   Begin
      // All or REGION
      If FFoldConfig[Ord(cfbtRegion)].Enabled Or (sfbIncludeDisabled In AFilter.Flags) Then
         Result := Result + inf.EndLevelRegion;
   End;

   If AFilter.FoldGroup In [0, FOLDGROUP_IFDEF] Then
   Begin
      // All or IFDEF
      If FFoldConfig[Ord(cfbtIfDef)].Enabled Or (sfbIncludeDisabled In AFilter.Flags) Then
         Result := Result + inf.EndLevelIfDef;
   End;
End;

Function TmscrScriptHighlighter.FoldBlockMinLevel(ALineIndex: TLineIdx; Const AFilter: TSynFoldBlockFilter): integer;
Var
   inf: TSynPasRangeInfo;
   r, r2: Pointer;
Begin
   Assert(CurrentRanges <> Nil, 'TSynCustomFoldHighlighter.FoldBlockMinLevel requires CurrentRanges');
   Result := 0;
   inf.EndLevelIfDef := 0;
   inf.MinLevelIfDef := 0;
   inf.EndLevelRegion := 0;
   inf.MinLevelRegion := 0;
   If (ALineIndex < 0) Or (ALineIndex >= CurrentLines.Count - 1) Then
      exit;

   If AFilter.FoldGroup In [0, FOLDGROUP_REGION, FOLDGROUP_IFDEF] Then
      inf := TSynHighlighterPasRangeList(CurrentRanges).PasRangeInfo[ALineIndex];

   If AFilter.FoldGroup In [0, FOLDGROUP_PASCAL] Then
   Begin
      // All or Pascal
    (* Range.EndLevel can be smaller. because Range.MinLevel does not know the LastLineFix
       Using a copy of FoldBlockEndLevel *)
      r := CurrentRanges[ALineIndex];
      If (r <> Nil) And (r <> NullRange) Then
      Begin
         r2 := TmscrScriptHighlighterRange(CurrentRanges[ALineIndex + 1]);
         If sfbIncludeDisabled In AFilter.Flags Then
         Begin
            Result := TmscrScriptHighlighterRange(r).NestFoldStackSize;
            If (r2 <> Nil) And (r2 <> NullRange) Then
               Result := Result + TmscrScriptHighlighterRange(r2).LastLineCodeFoldLevelFix;
            // now Result = FoldBlockEndLevel
            Result := Min(Result, TmscrScriptHighlighterRange(r).MinimumNestFoldBlockLevel);
         End
         Else
         Begin
            Result := TmscrScriptHighlighterRange(r).CodeFoldStackSize;
            If (r2 <> Nil) And (r2 <> NullRange) Then
               Result := Result + TmscrScriptHighlighterRange(r2).PasFoldFixLevel;
            // now Result = FoldBlockEndLevel
            Result := Min(Result, TmscrScriptHighlighterRange(r).MinimumCodeFoldBlockLevel);
         End;
      End;
   End;

   If AFilter.FoldGroup In [0, FOLDGROUP_REGION] Then
   Begin
      // All or REGION
      Result := Result + inf.MinLevelRegion;
   End;

   If AFilter.FoldGroup In [0, FOLDGROUP_IFDEF] Then
   Begin
      // All or IFDEF
      Result := Result + inf.MinLevelIfDef;
   End;
End;

Function TmscrScriptHighlighter.FoldBlockNestedTypes(ALineIndex: TLineIdx; ANestIndex: integer; Out AType: Pointer; Const AFilter: TSynFoldBlockFilter): boolean;
Var
   r, r2: Pointer;
   c: integer;
   Fold: TSynCustomCodeFoldBlock;
Begin
   Result := False;
   If (ALineIndex < 0) Or (ALineIndex >= CurrentLines.Count - 1) Then
      exit;
   If ANestIndex < 0 Then exit;

   If AFilter.FoldGroup = FOLDGROUP_PASCAL Then
   Begin
      If AFilter.Flags = [] Then
      Begin

         r := CurrentRanges[ALineIndex];
         If (r <> Nil) And (r <> NullRange) Then
         Begin
            r2 := TmscrScriptHighlighterRange(CurrentRanges[ALineIndex + 1]);
            c := TmscrScriptHighlighterRange(r).CodeFoldStackSize;
            If (r2 <> Nil) And (r2 <> NullRange) Then
               c := c + TmscrScriptHighlighterRange(r2).PasFoldFixLevel;

            If ANestIndex < c Then
            Begin
               c := TmscrScriptHighlighterRange(r).NestFoldStackSize - 1 - ANestIndex;
               Fold := TmscrScriptHighlighterRange(r).Top;
               While (Fold <> Nil) And ((c > 0) Or (Fold.BlockType >= CountPascalCodeFoldBlockOffset)) Do
               Begin
                  If (Fold.BlockType < CountPascalCodeFoldBlockOffset) Then
                     Dec(c);
                  Fold := Fold.Parent;
               End;
               If Fold <> Nil Then
               Begin
                  AType := Fold.BlockType;
                  If AType >= CountPascalCodeFoldBlockOffset Then
                     AType := AType - PtrUInt(CountPascalCodeFoldBlockOffset);
                  Result := True;
               End;
            End;

         End;

      End;
      If AFilter.Flags = [sfbIncludeDisabled] Then
      Begin

         r := CurrentRanges[ALineIndex];
         If (r <> Nil) And (r <> NullRange) Then
         Begin
            r2 := TmscrScriptHighlighterRange(CurrentRanges[ALineIndex + 1]);
            c := TmscrScriptHighlighterRange(r).NestFoldStackSize;
            If (r2 <> Nil) And (r2 <> NullRange) Then
               c := c + TmscrScriptHighlighterRange(r2).LastLineCodeFoldLevelFix;

            If ANestIndex < c Then
            Begin
               c := TmscrScriptHighlighterRange(r).NestFoldStackSize - 1 - ANestIndex;
               Fold := TmscrScriptHighlighterRange(r).Top;
               While (Fold <> Nil) And (c > 0) Do
               Begin
                  Fold := Fold.Parent;
                  Dec(c);
               End;
               If Fold <> Nil Then
               Begin
                  AType := Fold.BlockType;
                  If AType >= CountPascalCodeFoldBlockOffset Then
                     AType := AType - PtrUInt(CountPascalCodeFoldBlockOffset);
                  Result := True;
               End;
            End;

         End;
      End;
   End;
End;

Function TmscrScriptHighlighter.TopPascalCodeFoldBlockType(DownIndex: integer = 0): TmscrPascalCodeFoldBlockType;
Var
   p: Pointer;
Begin
   p := TopCodeFoldBlockType(DownIndex);
   If p >= CountPascalCodeFoldBlockOffset Then
      p := p - PtrUInt(CountPascalCodeFoldBlockOffset);
   Result := TmscrPascalCodeFoldBlockType(PtrUInt(p));
End;

Procedure TmscrScriptHighlighter.GetTokenBounds(Out LogX1, LogX2: integer);
Begin
   LogX1 := Run;
   LogX2 := LogX1 + fStringLen;
End;

Function TmscrScriptHighlighter.FoldTypeCount: integer;
Begin
   Result := 3;
End;

Function TmscrScriptHighlighter.FoldTypeAtNodeIndex(ALineIndex, FoldIndex: integer; UseCloseNodes: boolean): integer;
Var
   act: TSynFoldActions;
Begin
   act := [sfaOpenFold, sfaFold];
   If UseCloseNodes Then act := [sfaCloseFold, sfaFold];
   Case TmscrPascalCodeFoldBlockType(PtrUInt(FoldNodeInfo[ALineIndex].NodeInfoEx(FoldIndex, act).FoldType)) Of
      cfbtRegion:
         Result := 2;
      cfbtIfDef:
         Result := 3;
      Else
         Result := 1;
   End;
End;

Function TmscrScriptHighlighter.FoldLineLength(ALineIndex, FoldIndex: integer): integer;
Var
   atype: integer;
   node: TSynFoldNodeInfo;
Begin
   node := FoldNodeInfo[ALineIndex].NodeInfoEx(FoldIndex, [sfaOpenFold, sfaFold]);
   If sfaInvalid In node.FoldAction Then exit(-1);
   If sfaOneLineOpen In node.FoldAction Then exit(0);
   Case TmscrPascalCodeFoldBlockType(PtrUInt(node.FoldType)) Of
      cfbtRegion:
         atype := 2;
      cfbtIfDef:
         atype := 3;
      Else
         atype := 1;
   End;

   Result := FoldEndLine(ALineIndex, FoldIndex);
   // check if fold last line of block (not mixed "end begin")
   If (FoldBlockEndLevel(Result, atype) > FoldBlockMinLevel(Result, atype)) Then
      Dec(Result);
   // Amount of lines, that will become invisible (excludes the cfCollapsed line)
   Result := Result - ALineIndex;
End;

Function TmscrScriptHighlighter.FoldEndLine(ALineIndex, FoldIndex: integer): integer;
Var
   lvl, cnt, atype: integer;
   node: TSynFoldNodeInfo;
Begin
   node := FoldNodeInfo[ALineIndex].NodeInfoEx(FoldIndex, [sfaOpenFold, sfaFold]);
   If sfaInvalid In node.FoldAction Then exit(-1);
   If sfaOneLineOpen In node.FoldAction Then exit(ALineIndex);
   Case TmscrPascalCodeFoldBlockType(PtrUInt(node.FoldType)) Of
      cfbtRegion:
         atype := 2;
      cfbtIfDef:
         atype := 3;
      Else
         atype := 1;
   End;

   cnt := CurrentLines.Count;
   lvl := node.FoldLvlEnd;

   Result := ALineIndex + 1;
   While (Result < cnt) And (FoldBlockMinLevel(Result, atype) >= lvl) Do
      Inc(Result);
   // check if fold last line of block (not mixed "end begin")
   // and not lastlinefix
   If (Result = cnt) Then
      Dec(Result);
End;

Function TmscrScriptHighlighter.LastLinePasFoldLevelFix(Index: integer; AType: integer; AIncludeDisabled: boolean): integer;
Var
   r: TmscrScriptHighlighterRange;
Begin
   // AIncludeDisabled only works for Pascal Nodes
   Case AType Of
      2: Result := 0;
      3: Result := 0;
      Else
         Begin
            If (Index < 0) Or (Index >= CurrentLines.Count) Then
               exit(0);
            r := TmscrScriptHighlighterRange(CurrentRanges[Index]);
            If (r <> Nil) And (Pointer(r) <> NullRange) Then
            Begin
               If AIncludeDisabled Then
                  Result := r.LastLineCodeFoldLevelFix // all pascal nodes (incl. not folded)
               Else
                  Result := r.PasFoldFixLevel;
            End
            Else
               Result := 0;
         End;
   End;
End;

Procedure TmscrScriptHighlighter.DoInitNode(Var Node: TSynFoldNodeInfo; FinishingABlock: boolean; ABlockType: Pointer; aActions: TSynFoldActions; AIsFold: boolean);
Var
   PasBlockType: TmscrPascalCodeFoldBlockType;
   EndOffs: integer;
   OneLine: boolean;
   t: TSynCustomFoldConfig;
Begin
   PasBlockType := TmscrPascalCodeFoldBlockType(PtrUint(ABlockType));

   If FinishingABlock Then
      EndOffs := -1
   Else
      EndOffs := +1;

   aActions := aActions + [sfaMultiLine];

   If (Not FinishingABlock) And (ABlockType <> Nil) Then
   Begin
      If (PasBlockType In [cfbtIfThen, cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfElse]) Then

         Include(aActions, sfaOutlineKeepLevel);

      If (PasBlockType In [cfbtProcedure]) Then
      Begin
         t := FFoldConfig[Ord(cfbtTopBeginEnd)];
         If t.Enabled And (sfaOutline In t.FoldActions) Then
            aActions := aActions + [sfaOutlineKeepLevel, sfaOutlineNoColor];
      End;

      If (PasBlockType In [cfbtExcept]) Then
      Begin
         t := FFoldConfig[Ord(cfbtTry)];
         If t.Enabled And (sfaOutline In t.FoldActions) Then
            Include(aActions, sfaOutlineMergeParent);
      End;

   End;

   Node.LineIndex := LineIndex;
   Node.LogXStart := Run;
   Node.LogXEnd := Run + fStringLen;
   Node.FoldType := Pointer({%H-} PtrInt(ABlockType)); //Pointer(PtrUInt(PasBlockType));
   Node.FoldTypeCompatible := Pointer(PtrUInt(PascalFoldTypeCompatibility[PasBlockType]));
   Node.FoldAction := aActions;
   Case PasBlockType Of
      cfbtRegion:
         Begin
            node.FoldGroup := FOLDGROUP_REGION;
            If AIsFold Then
               Node.FoldLvlStart := FSynPasRangeInfo.EndLevelRegion
            Else
               Node.FoldLvlStart := 0;
            Node.NestLvlStart := FSynPasRangeInfo.EndLevelRegion;
            OneLine := FinishingABlock And (Node.FoldLvlStart > FSynPasRangeInfo.MinLevelRegion);
         End;
      cfbtIfDef:
         Begin
            node.FoldGroup := FOLDGROUP_IFDEF;
            If AIsFold Then
               Node.FoldLvlStart := FSynPasRangeInfo.EndLevelIfDef
            Else
               Node.FoldLvlStart := 0;
            Node.NestLvlStart := FSynPasRangeInfo.EndLevelIfDef;
            OneLine := FinishingABlock And (Node.FoldLvlStart > FSynPasRangeInfo.MinLevelIfDef);
         End;
      Else
         Begin
            //inherited DoInitNode(Node, FinishingABlock, ABlockType, aActions, AIsFold);
            node.FoldGroup := FOLDGROUP_PASCAL;
            If AIsFold Then
            Begin
               Node.FoldLvlStart := PasCodeFoldRange.CodeFoldStackSize;
               Node.NestLvlStart := PasCodeFoldRange.NestFoldStackSize;
               OneLine := FinishingABlock And (Node.FoldLvlStart > PasCodeFoldRange.MinimumCodeFoldBlockLevel); // MinimumCodeFoldBlockLevel);
            End
            Else
            Begin
               Node.FoldLvlStart := PasCodeFoldRange.NestFoldStackSize; // CodeFoldStackSize; ? but fails test // Todo: zero?
               Node.NestLvlStart := PasCodeFoldRange.NestFoldStackSize;
               OneLine := FinishingABlock And (Node.NestLvlStart > PasCodeFoldRange.MinimumNestFoldBlockLevel); // MinimumCodeFoldBlockLevel);
            End;
         End;
   End;
   Node.NestLvlEnd := Node.NestLvlStart + EndOffs;
   If Not (sfaFold In aActions) Then
      EndOffs := 0;
   Node.FoldLvlEnd := Node.FoldLvlStart + EndOffs;
   If OneLine Then // find opening node
      RepairSingleLineNode(Node);
End;

Procedure TmscrScriptHighlighter.StartCustomCodeFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType);
Var
   act: TSynFoldActions;
   nd: TSynFoldNodeInfo;
   FoldBlock, BlockEnabled: boolean;
Begin
   BlockEnabled := FFoldConfig[Ord(ABlockType)].Enabled;
   FoldBlock := BlockEnabled And (FFoldConfig[Ord(ABlockType)].Modes * [fmFold, fmHide] <> []);
   //if not FFoldConfig[ord(ABlockType)].Enabled then exit;
   If IsCollectingNodeInfo Then
   Begin // exclude subblocks, because they do not increase the foldlevel yet
      act := [sfaOpen, sfaOpenFold];
      If BlockEnabled Then
         act := act + FFoldConfig[Ord(ABlockType)].FoldActions;
      If Not FAtLineStart Then
         act := act - [sfaFoldHide];
      DoInitNode(nd {%H-}, False, Pointer(PtrUInt(ABlockType)), act, FoldBlock);
      CollectingNodeInfoList.Add(nd);
   End;
   //if not FoldBlock then
   //  exit;
   Case ABlockType Of
      cfbtIfDef:
         Inc(FSynPasRangeInfo.EndLevelIfDef);
      cfbtRegion:
         Inc(FSynPasRangeInfo.EndLevelRegion);
   End;
End;

Procedure TmscrScriptHighlighter.EndCustomCodeFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType);
Var
   act: TSynFoldActions;
   nd: TSynFoldNodeInfo;
   FoldBlock, BlockEnabled: boolean;
Begin
   FoldBlock := FFoldConfig[Ord(ABlockType)].Enabled;
   //if not FFoldConfig[ord(ABlockType)].Enabled then exit;
   If IsCollectingNodeInfo Then
   Begin // exclude subblocks, because they do not increase the foldlevel yet
      BlockEnabled := FFoldConfig[Ord(ABlockType)].Enabled;
      act := [sfaClose, sfaCloseFold];
      If BlockEnabled Then
         act := act + FFoldConfig[PtrUInt(ABlockType)].FoldActions;
      If Not FoldBlock Then
         act := act - [sfaFold, sfaFoldFold, sfaFoldHide];
      DoInitNode(nd {%H-}, True, Pointer(PtrUInt(ABlockType)), act, FoldBlock); // + FFoldConfig[ord(ABlockType)].FoldActions);
      CollectingNodeInfoList.Add(nd);
   End;
   //if not FoldBlock then
   //  exit;
   Case ABlockType Of
      cfbtIfDef:
         Begin
            If FSynPasRangeInfo.EndLevelIfDef > 0 Then
               Dec(FSynPasRangeInfo.EndLevelIfDef);
            If FSynPasRangeInfo.EndLevelIfDef < FSynPasRangeInfo.MinLevelIfDef Then
               FSynPasRangeInfo.MinLevelIfDef := FSynPasRangeInfo.EndLevelIfDef;
         End;
      cfbtRegion:
         Begin
            If FSynPasRangeInfo.EndLevelRegion > 0 Then
               Dec(FSynPasRangeInfo.EndLevelRegion);
            If FSynPasRangeInfo.EndLevelRegion < FSynPasRangeInfo.MinLevelRegion Then
               FSynPasRangeInfo.MinLevelRegion := FSynPasRangeInfo.EndLevelRegion;
         End;
   End;
End;

Procedure TmscrScriptHighlighter.CollectNodeInfo(FinishingABlock: boolean; ABlockType: Pointer; LevelChanged: boolean);
Begin
   // nothing
End;

Function TmscrScriptHighlighter.CreateFoldNodeInfoList: TLazSynFoldNodeInfoList;
Begin
   Result := TLazSynFoldNodeInfoList.Create;
End;

Procedure TmscrScriptHighlighter.ScanFoldNodeInfo();
Var
   nd: PSynFoldNodeInfo;
   i: integer;
Begin
   fStringLen := 0;
   Inherited ScanFoldNodeInfo;

   fStringLen := 0;
   i := LastLinePasFoldLevelFix(LineIndex + 1, FOLDGROUP_PASCAL, True); // all pascal nodes (incl. not folded)
   While i < 0 Do
   Begin
      EndPascalCodeFoldBlock;
      CollectingNodeInfoList.LastItemPointer^.FoldAction :=
         CollectingNodeInfoList.LastItemPointer^.FoldAction + [sfaCloseForNextLine];
      Inc(i);
   End;
   If LineIndex = CurrentLines.Count - 1 Then
   Begin
      // last LineIndex, close all folds
      // Run (for LogXStart) is at LineIndex-end
      i := CollectingNodeInfoList.CountAll;
      While TopPascalCodeFoldBlockType <> cfbtNone Do
         EndPascalCodeFoldBlock(True);
      While FSynPasRangeInfo.EndLevelIfDef > 0 Do
         EndCustomCodeFoldBlock(cfbtIfDef);
      While FSynPasRangeInfo.EndLevelRegion > 0 Do
         EndCustomCodeFoldBlock(cfbtRegion);
      While i < CollectingNodeInfoList.CountAll Do
      Begin
         nd := CollectingNodeInfoList.ItemPointer[i];
         nd^.FoldAction := nd^.FoldAction + [sfaLastLineClose];
         Inc(i);
      End;
   End;
End;

Function TmscrScriptHighlighter.StartPascalCodeFoldBlock(ABlockType: TmscrPascalCodeFoldBlockType; ForceDisabled: boolean): TSynCustomCodeFoldBlock;
Var
   p: PtrInt;
   FoldBlock, BlockEnabled: boolean;
   act: TSynFoldActions;
   nd: TSynFoldNodeInfo;
Begin
   If rsSkipAllPasBlocks In fRange Then exit(Nil);
   BlockEnabled := FFoldConfig[Ord(ABlockType)].Enabled;
   If (Not BlockEnabled) And (Not ForceDisabled) And (Not FFoldConfig[Ord(ABlockType)].IsEssential) Then
      exit(Nil);

   FoldBlock := BlockEnabled And (FFoldConfig[Ord(ABlockType)].Modes * [fmFold, fmHide] <> []);
   p := 0;
   // TODO: let inherited call CollectNodeInfo
   If IsCollectingNodeInfo Then
   Begin // exclude subblocks, because they do not increase the foldlevel yet
      act := [sfaOpen, sfaOpenFold]; //TODO: sfaOpenFold not for cfbtIfThen
      If BlockEnabled Then
         act := act + FFoldConfig[Ord(ABlockType)].FoldActions;
      If Not FAtLineStart Then
         act := act - [sfaFoldHide];
      DoInitNode(nd {%H-}, False, Pointer(PtrUInt(ABlockType)), act, FoldBlock);
      CollectingNodeInfoList.Add(nd);
   End;
   If Not FoldBlock Then
      p := PtrInt(CountPascalCodeFoldBlockOffset);
   Result := TSynCustomCodeFoldBlock(StartCodeFoldBlock(p + Pointer(PtrInt(ABlockType)), FoldBlock, True));
End;

Procedure TmscrScriptHighlighter.EndPascalCodeFoldBlock(NoMarkup: boolean; UndoInvalidOpen: boolean);
Var
   DecreaseLevel, BlockEnabled: boolean;
   act: TSynFoldActions;
   BlockType: TmscrPascalCodeFoldBlockType;
   nd: TSynFoldNodeInfo;
Begin
   Exclude(fRange, rsSkipAllPasBlocks);
   BlockType := TopPascalCodeFoldBlockType;
   If BlockType In [cfbtVarType, cfbtLocalVarType] Then
      fRange := fRange - [rsInTypeBlock];
   fRange := fRange - [rsAfterEqual];
   DecreaseLevel := TopCodeFoldBlockType < CountPascalCodeFoldBlockOffset;
   // TODO: let inherited call CollectNodeInfo
   If IsCollectingNodeInfo Then
   Begin // exclude subblocks, because they do not increase the foldlevel yet
      BlockEnabled := FFoldConfig[Ord(BlockType)].Enabled;
      act := [sfaClose, sfaCloseFold];
      If BlockEnabled Then
         act := act + FFoldConfig[Ord(BlockType)].FoldActions - [sfaFoldFold, sfaFoldHide]; // TODO: Why filter?
      If Not DecreaseLevel Then
         act := act - [sfaFold, sfaFoldFold, sfaFoldHide];
      If NoMarkup Then
         exclude(act, sfaMarkup);
      If UndoInvalidOpen Then
         act := act - [sfaMarkup, {sfaFold,} sfaFoldFold, sfaFoldHide, sfaOutline]; // sfaFold affects the EndOffset for the fold-lvl
      DoInitNode(nd {%H-}, True, Pointer(PtrUInt(BlockType)), act, DecreaseLevel);
      CollectingNodeInfoList.Add(nd);
   End;
   EndCodeFoldBlock(DecreaseLevel);
End;

Procedure TmscrScriptHighlighter.CloseBeginEndBlocksBeforeProc;
Begin
   If Not (TopPascalCodeFoldBlockType In PascalStatementBlocks) Then
      exit;
   While TopPascalCodeFoldBlockType In PascalStatementBlocks Do
      EndPascalCodeFoldBlockLastLine;
   If TopPascalCodeFoldBlockType = cfbtProcedure Then
      EndPascalCodeFoldBlockLastLine; // This procedure did have a begin/end block, so it must end too
End;

Procedure TmscrScriptHighlighter.SmartCloseBeginEndBlocks(SearchFor: TmscrPascalCodeFoldBlockType);
Var
   i: integer;
   t: TmscrPascalCodeFoldBlockType;
Begin
   // Close unfinished blocks, IF the expected type is found
   // Only check a limited deep. Otherwhise assume, that the "SearchFor"-End node may be misplaced
   i := 0;
   While (i <= 2) Do
   Begin
      t := TopPascalCodeFoldBlockType(i);
      If Not (t In PascalStatementBlocks + [SearchFor]) Then
         exit;
      If (t = SearchFor) Then break;
      Inc(i);
   End;
   If i > 2 Then
      exit;

   While i > 0 Do
   Begin
      EndPascalCodeFoldBlockLastLine;
      Dec(i);
   End;
End;

Procedure TmscrScriptHighlighter.EndPascalCodeFoldBlockLastLine;
Var
   i: integer;
   nd: PSynFoldNodeInfo;
Begin
   If IsCollectingNodeInfo Then
      i := CollectingNodeInfoList.CountAll;
   EndPascalCodeFoldBlock;
   If FAtLineStart Then
   Begin

      If (PasCodeFoldRange.NestFoldStackSize < FStartCodeFoldBlockLevel) And (FStartCodeFoldBlockLevel > 0) Then
      Begin
         PasCodeFoldRange.DecLastLineCodeFoldLevelFix;
         Dec(FStartCodeFoldBlockLevel);
         If IsCollectingNodeInfo Then CollectingNodeInfoList.Delete;
      End;

      If (PasCodeFoldRange.CodeFoldStackSize < FPasStartLevel) And (FPasStartLevel > 0) Then
      Begin
         PasCodeFoldRange.DecLastLinePasFoldFix;
         Dec(FPasStartLevel);
      End;
   End
   Else If IsCollectingNodeInfo And (CollectingNodeInfoList.CountAll > i) Then
   Begin
      nd := CollectingNodeInfoList.LastItemPointer;
      exclude(nd^.FoldAction, sfaMarkup);
      nd^.LogXEnd := 0;
   End;
End;

Function TmscrScriptHighlighter.GetDrawDivider(Index: integer): TSynDividerDrawConfigSetting;

   Function CheckFoldNestLevel(MaxDepth, StartLvl: integer; CountTypes, SkipTypes: TmscrPascalCodeFoldBlockTypes; Out ResultLvl: integer): boolean;
   Var
      i, j, m: integer;
      t: TmscrPascalCodeFoldBlockType;
   Begin
      i := 0;
      j := StartLvl;
      m := PasCodeFoldRange.NestFoldStackSize;
      ;
      t := TopPascalCodeFoldBlockType(j);
      While (i <= MaxDepth) And (j < m) And ((t In CountTypes) Or (t In SkipTypes)) Do
      Begin
         Inc(j);
         If t In CountTypes Then
            Inc(i);
         t := TopPascalCodeFoldBlockType(j);
      End;
      ResultLvl := i;
      Result := i <= MaxDepth;
   End;

Var
   cur: TmscrPascalCodeFoldBlockType;
   CloseCnt, ClosedByNextLine, ClosedInLastLine: integer;
   i, top, c: integer;
   t: TSynPasDividerDrawLocation;
Begin
   Result := Inherited;
   If (index = 0) Then exit;
   CloseCnt := FoldBlockClosingCount(Index, FOLDGROUP_PASCAL, [sfbIncludeDisabled]);
   If (CloseCnt = 0) Or (FoldBlockMinLevel(Index, FOLDGROUP_PASCAL, [sfbIncludeDisabled]) <> FoldBlockEndLevel(Index, FOLDGROUP_PASCAL, [sfbIncludeDisabled])) Then
      exit;

   ClosedByNextLine := -LastLinePasFoldLevelFix(Index + 1, FOLDGROUP_PASCAL, True);

   ClosedInLastLine := -LastLinePasFoldLevelFix(Index, FOLDGROUP_PASCAL, True);

   i := ClosedByNextLine - 1;
   If i >= 0 Then
   Begin
      SetRange(CurrentRanges[Index]);
      top := 0;
   End
   Else
   Begin
      SetRange(CurrentRanges[Index - 1]);
      i := CloseCnt - ClosedByNextLine + ClosedInLastLine - 1;
      top := ClosedInLastLine;
   End;

   cur := TopPascalCodeFoldBlockType(i + 1);
   While (i >= top) Do
   Begin

      cur := TopPascalCodeFoldBlockType(i);
      Result := FDividerDrawConfig[pddlUses].TopSetting;
      Case cur Of
         cfbtUnitSection:
            If FDividerDrawConfig[pddlUnitSection].MaxDrawDepth > 0 Then
               exit(FDividerDrawConfig[pddlUnitSection].TopSetting);
         cfbtUses:
            If FDividerDrawConfig[pddlUses].MaxDrawDepth > 0 Then
               exit(FDividerDrawConfig[pddlUses].TopSetting);
         cfbtLocalVarType:
            If CheckFoldNestLevel(FDividerDrawConfig[pddlVarLocal].MaxDrawDepth - 1, i + 2, [cfbtProcedure], cfbtAll, c) Then
            Begin
               If c = 0 Then
                  exit(FDividerDrawConfig[pddlVarLocal].TopSetting)
               Else
                  exit(FDividerDrawConfig[pddlVarLocal].NestSetting);
            End;
         cfbtVarType:
            If FDividerDrawConfig[pddlVarGlobal].MaxDrawDepth > 0 Then
               exit(FDividerDrawConfig[pddlVarGlobal].TopSetting);

         cfbtProcedure:
            If CheckFoldNestLevel(FDividerDrawConfig[pddlProcedure].MaxDrawDepth - 1, i + 1, [cfbtProcedure], cfbtAll, c) Then
            Begin
               If c = 0 Then
                  exit(FDividerDrawConfig[pddlProcedure].TopSetting)
               Else
                  exit(FDividerDrawConfig[pddlProcedure].NestSetting);
            End;
         cfbtTopBeginEnd:
            If FDividerDrawConfig[pddlBeginEnd].MaxDrawDepth > 0 Then
               exit(FDividerDrawConfig[pddlBeginEnd].TopSetting);
         cfbtBeginEnd, cfbtRepeat, cfbtCase:
            If CheckFoldNestLevel(FDividerDrawConfig[pddlBeginEnd].MaxDrawDepth - 2, i + 1, [cfbtBeginEnd, cfbtRepeat, cfbtCase], cfbtAll - [cfbtProcedure, cfbtTopBeginEnd], c) Then
               exit(FDividerDrawConfig[pddlBeginEnd].NestSetting);
         cfbtTry:
            If CheckFoldNestLevel(FDividerDrawConfig[pddlTry].MaxDrawDepth - 1, i + 1, [cfbtTry], cfbtAll - [cfbtProcedure], c) Then
            Begin
               If c = 0 Then
                  exit(FDividerDrawConfig[pddlTry].TopSetting)
               Else
                  exit(FDividerDrawConfig[pddlTry].NestSetting);
            End;
      End;
      Dec(i);
      If (i < top) And (ClosedByNextLine > 0) Then
      Begin

         SetRange(CurrentRanges[Index - 1]);
         i := CloseCnt - ClosedByNextLine + ClosedInLastLine - 1;
         ClosedByNextLine := 0;
         top := ClosedInLastLine;
         cur := TopPascalCodeFoldBlockType(i + 1);
      End;
   End;
   Result := Inherited;
End;

Procedure TmscrScriptHighlighter.CreateDividerDrawConfig;
Var
   i: TSynPasDividerDrawLocation;
Begin
   For i := low(TSynPasDividerDrawLocation) To high(TSynPasDividerDrawLocation) Do
   Begin
      FDividerDrawConfig[i] := TSynDividerDrawConfig.Create;
      FDividerDrawConfig[i].OnChange := @DefHighlightChange;
      FDividerDrawConfig[i].MaxDrawDepth := PasDividerDrawLocationDefaults[i];
   End;
End;

Procedure TmscrScriptHighlighter.DestroyDividerDrawConfig;
Var
   i: TSynPasDividerDrawLocation;
Begin
   For i := low(TSynPasDividerDrawLocation) To high(TSynPasDividerDrawLocation) Do
      FreeAndNil(FDividerDrawConfig[i]);
End;

Function TmscrScriptHighlighter.CreateFoldConfigInstance(Index: integer): TSynCustomFoldConfig;
Var
   m: TSynCustomFoldConfigModes;
Begin
   m := [];
   If TmscrPascalCodeFoldBlockType(Index) In PascalWordTripletRanges Then
      m := [fmMarkup];

   Case TmscrPascalCodeFoldBlockType(Index) Of
      cfbtRegion, cfbtNestedComment, cfbtAnsiComment, cfbtBorCommand, cfbtSlashComment:
         m := [fmFold, fmHide] + m;
      cfbtIfThen, cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfElse:
         m := m;
      cfbtFirstPrivate..high(TmscrPascalCodeFoldBlockType):
         m := [];
      Else
         m := [fmFold] + m;
   End;
   If Not (TmscrPascalCodeFoldBlockType(Index) In PascalNoOutlineRanges) Then
      m := m + [fmOutline];

   Result := TSynCustomFoldConfig.Create(m, TmscrPascalCodeFoldBlockType(Index) In cfbtEssential);
End;

Function TmscrScriptHighlighter.GetFoldConfigInstance(Index: integer): TSynCustomFoldConfig;
Var
   m: TSynCustomFoldConfigModes;
Begin
   Result := Inherited GetFoldConfigInstance(Index);

   m := Result.SupportedModes;

   If (TmscrPascalCodeFoldBlockType(Index) In [cfbtIfThen, cfbtForDo, cfbtWhileDo, cfbtWithDo, cfbtIfElse]) Then
      m := [];
   If TmscrPascalCodeFoldBlockType(Index) In [cfbtSlashComment] Then
      Result.Modes := [fmFold, fmHide] + m
   Else
      Result.Modes := [fmFold] + m;

   If Not (TmscrPascalCodeFoldBlockType(Index) In PascalNoOutlineRanges) Then
      Result.Modes := Result.Modes + [fmOutline];

   Result.Enabled := (TmscrPascalCodeFoldBlockType(Index) In [cfbtBeginEnd..cfbtLastPublic]) And (Result.Modes <> []);
End;

Function TmscrScriptHighlighter.CreateRangeList(ALines: TSynEditStringsBase): TSynHighlighterRangeList;
Begin
   Result := TSynHighlighterPasRangeList.Create;
End;

Function TmscrScriptHighlighter.UpdateRangeInfoAtLine(Index: integer): boolean;
Var
   r: TSynPasRangeInfo;
Begin
   Result := Inherited;
   r := TSynHighlighterPasRangeList(CurrentRanges).PasRangeInfo[Index];
   Result := Result Or (FSynPasRangeInfo.EndLevelIfDef <> r.EndLevelIfDef) Or (FSynPasRangeInfo.MinLevelIfDef <> r.MinLevelIfDef) Or (FSynPasRangeInfo.EndLevelRegion <>
      r.EndLevelRegion) Or (FSynPasRangeInfo.MinLevelRegion <> r.MinLevelRegion);
   TSynHighlighterPasRangeList(CurrentRanges).PasRangeInfo[Index] := FSynPasRangeInfo;
End;

Function TmscrScriptHighlighter.GetFoldConfigCount: integer;
Begin
   // excluded cfbtNone; // maybe ord(cfbtLastPublic)+1 ?
   Result := Ord(high(TmscrPascalCodeFoldBlockType)) - Ord(low(TmscrPascalCodeFoldBlockType)) - 1;
End;

Function TmscrScriptHighlighter.GetFoldConfigInternalCount: integer;
Begin

   Result := Ord(high(TmscrPascalCodeFoldBlockType)) - Ord(low(TmscrPascalCodeFoldBlockType)) + 1;
End;

Procedure TmscrScriptHighlighter.DoFoldConfigChanged(Sender: TObject);
Begin
   Inherited DoFoldConfigChanged(Sender);
End;

Function TmscrScriptHighlighter.GetDividerDrawConfig(Index: integer): TSynDividerDrawConfig;
Begin
   Result := FDividerDrawConfig[TSynPasDividerDrawLocation(Index)];
End;

Function TmscrScriptHighlighter.GetDividerDrawConfigCount: integer;
Begin
   Result := Ord(high(TSynPasDividerDrawLocation)) - Ord(low(TSynPasDividerDrawLocation)) + 1;
End;

Function TmscrScriptHighlighter.GetRangeClass: TSynCustomHighlighterRangeClass;
Begin
   Result := TmscrScriptHighlighterRange;
End;

Function TmscrScriptHighlighter.UseUserSettings(settingIndex: integer): boolean;

   Function ReadDelphiSettings(settingIndex: integer): boolean;

      Function ReadDelphiSetting(settingTag: String; attri: TSynHighlighterAttributes; key: String): boolean;

         Function ReadDelphi2Or3(settingTag: String; attri: TSynHighlighterAttributes; Name: String): boolean;
         Var
            i: integer;
         Begin
            For i := 1 To Length(Name) Do
               If Name[i] = ' ' Then Name[i] := '_';
            Result := attri.LoadFromBorlandRegistry(HKEY_CURRENT_USER, '\Software\Borland\Delphi\' + settingTag + '\Highlight', Name, True);
         End; { ReadDelphi2Or3 }

         Function ReadDelphi4OrMore(settingTag: String; attri: TSynHighlighterAttributes; key: String): boolean;
         Begin
            Result := attri.LoadFromBorlandRegistry(HKEY_CURRENT_USER, '\Software\Borland\Delphi\' + settingTag + '\Editor\Highlight', key, False);
         End;
         { ReadDelphi4OrMore }
      Begin { ReadDelphiSetting }
         Try
            If (settingTag[1] = '2') Or (settingTag[1] = '3') Then
               Result := ReadDelphi2Or3(settingTag, attri, key)
            Else
               Result := ReadDelphi4OrMore(settingTag, attri, key);
         Except
            Result := False;
         End;
      End;
      { ReadDelphiSetting }
   Var
      tmpStringAttri: TSynHighlighterAttributes;
      tmpNumberAttri: TSynHighlighterAttributes;
      tmpKeyAttri: TSynHighlighterAttributes;
      tmpSymbolAttri: TSynHighlighterAttributes;
      tmpAsmAttri: TSynHighlighterAttributes;
      tmpCommentAttri: TSynHighlighterAttributes;
      tmpDirectiveAttri: TSynHighlighterAttributes;
      tmpIdentifierAttri: TSynHighlighterAttributes;
      tmpSpaceAttri: TSynHighlighterAttributes;
      s: TStringList;
   Begin { ReadDelphiSettings }
      s := TStringList.Create;
      Try
         EnumUserSettings(s);
         If (settingIndex < 0) Or (settingIndex >= s.Count) Then
            Result := False
         Else
         Begin
            tmpStringAttri := TSynHighlighterAttributes.Create(Nil);
            tmpNumberAttri := TSynHighlighterAttributes.Create(Nil);
            tmpKeyAttri := TSynHighlighterAttributes.Create(Nil);
            tmpSymbolAttri := TSynHighlighterAttributes.Create(Nil);
            tmpAsmAttri := TSynHighlighterAttributes.Create(Nil);
            tmpCommentAttri := TSynHighlighterAttributes.Create(Nil);
            tmpDirectiveAttri := TSynHighlighterAttributes.Create(Nil);
            tmpIdentifierAttri := TSynHighlighterAttributes.Create(Nil);
            tmpSpaceAttri := TSynHighlighterAttributes.Create(Nil);
            tmpStringAttri.Assign(fStringAttri);
            tmpNumberAttri.Assign(fNumberAttri);
            tmpKeyAttri.Assign(fKeyAttri);
            tmpSymbolAttri.Assign(fSymbolAttri);
            tmpCommentAttri.Assign(fCommentAttri);
            tmpDirectiveAttri.Assign(fDirectiveAttri);
            tmpIdentifierAttri.Assign(fIdentifierAttri);
            tmpSpaceAttri.Assign(fSpaceAttri);
            Result := ReadDelphiSetting(s[settingIndex], fCommentAttri, 'Comment') And ReadDelphiSetting(s[settingIndex], fDirectiveAttri, 'Directive') And
               ReadDelphiSetting(s[settingIndex], fIdentifierAttri, 'Identifier') And ReadDelphiSetting(s[settingIndex], fKeyAttri, 'Reserved word') And
               ReadDelphiSetting(s[settingIndex], fNumberAttri, 'Number') And ReadDelphiSetting(s[settingIndex], fSpaceAttri, 'Whitespace') And ReadDelphiSetting(s[settingIndex], fStringAttri, 'string') And
               ReadDelphiSetting(s[settingIndex], fSymbolAttri, 'Symbol');
            If Not Result Then
            Begin
               fStringAttri.Assign(tmpStringAttri);
               fNumberAttri.Assign(tmpNumberAttri);
               fKeyAttri.Assign(tmpKeyAttri);
               fSymbolAttri.Assign(tmpSymbolAttri);
               fCommentAttri.Assign(tmpCommentAttri);
               FIDEDirectiveAttri.Assign(tmpCommentAttri);
               fDirectiveAttri.Assign(tmpDirectiveAttri);
               fIdentifierAttri.Assign(tmpIdentifierAttri);
               fSpaceAttri.Assign(tmpSpaceAttri);
            End;
            tmpStringAttri.Free;
            tmpNumberAttri.Free;
            tmpKeyAttri.Free;
            tmpSymbolAttri.Free;
            tmpAsmAttri.Free;
            tmpCommentAttri.Free;
            tmpDirectiveAttri.Free;
            tmpIdentifierAttri.Free;
            tmpSpaceAttri.Free;
         End;
      Finally
         s.Free;
      End;
   End;

Begin
   Result := ReadDelphiSettings(settingIndex);
End;

Function TmscrScriptHighlighter.GetIdentChars: TSynIdentChars;
Begin
   Result := ['&', '_', '0'..'9', 'a'..'z', 'A'..'Z'];
End;

Class Function TmscrScriptHighlighter.GetLanguageName: String;
Begin
   Result := SYNS_LangPascal;
End;

Class Function TmscrScriptHighlighter.GetCapabilities: TSynHighlighterCapabilities;
Begin
   Result := Inherited GetCapabilities + [hcUserSettings];
End;

{begin}//mh 2000-10-08

Function TmscrScriptHighlighter.IsFilterStored: boolean;
Begin
   Result := fDefaultFilter <> SYNS_FilterPascal;
End;

Procedure TmscrScriptHighlighter.CreateRootCodeFoldBlock;
Begin
   Inherited;
   RootCodeFoldBlock.InitRootBlockType(Pointer(PtrInt(cfbtNone)));
End;

Function TmscrScriptHighlighter.IsKeyword(Const AKeyword: String): boolean;
// returns true for some common keywords
// Note: this words are not always keywords (e.g. end), and some keywords are
// not listed here at all (e.g. static)
Var
   i: integer;
   m: TPascalCompilerMode;
Begin
   If KeywordsList = Nil Then
   Begin
      KeywordsList := TStringList.Create;
      For i := 1 To High(RESERVED_WORDS_TP) Do
         KeywordsList.AddObject(RESERVED_WORDS_TP[i], TObject(pcmTP));
      For i := 1 To High(RESERVED_WORDS_DELPHI) Do
         KeywordsList.AddObject(RESERVED_WORDS_DELPHI[i], TObject(pcmDelphi));
      For i := 1 To High(RESERVED_WORDS_FPC) Do
         KeywordsList.AddObject(RESERVED_WORDS_FPC[i], TObject(pcmFPC));
      KeywordsList.Sorted := True;
   End;
   Result := KeywordsList.Find(LowerCase(AKeyword), i);
   If Not Result Then exit;
   m := TPascalCompilerMode(PtrUInt(KeywordsList.Objects[i]));
   Case FCompilerMode Of
      pcmFPC, pcmObjFPC: ;
      pcmDelphi: Result := m In [pcmTP, pcmDelphi];
      Else
         Result := m = pcmTP;
   End;
End;

{end}//mh 2000-10-08

Procedure TmscrScriptHighlighter.SetD4syntax(Const Value: boolean);
Begin
   FD4syntax := Value;
End;

{ TSynFreePascalSyn }

Constructor TSynFreePascalSyn.Create(AOwner: TComponent);
Begin
   Inherited Create(AOwner);
   CompilerMode := pcmObjFPC;
End;

Procedure TSynFreePascalSyn.ResetRange;
Begin
   Inherited ResetRange;
   CompilerMode := pcmObjFPC;
End;

{ TmscrScriptHighlighterRange }

Procedure TmscrScriptHighlighterRange.Clear;
Begin
   Inherited Clear;
   FBracketNestLevel := 0;
   FLastLineCodeFoldLevelFix := 0;
   FPasFoldFixLevel := 0;
End;

Function TmscrScriptHighlighterRange.Compare(Range: TSynCustomHighlighterRange): integer;
Begin
   Result := Inherited Compare(Range);
   If Result <> 0 Then exit;

   Result := Ord(FMode) - Ord(TmscrScriptHighlighterRange(Range).FMode);
   If Result <> 0 Then exit;
   Result := Ord(FNestedComments) - Ord(TmscrScriptHighlighterRange(Range).FNestedComments);
   If Result <> 0 Then exit;
   Result := Ord(FTypeHelpers) - Ord(TmscrScriptHighlighterRange(Range).FTypeHelpers);
   If Result <> 0 Then exit;
   Result := FBracketNestLevel - TmscrScriptHighlighterRange(Range).FBracketNestLevel;
   If Result <> 0 Then exit;
   Result := FLastLineCodeFoldLevelFix - TmscrScriptHighlighterRange(Range).FLastLineCodeFoldLevelFix;
   If Result <> 0 Then exit;
   Result := FPasFoldFixLevel - TmscrScriptHighlighterRange(Range).FPasFoldFixLevel;
End;

Procedure TmscrScriptHighlighterRange.Assign(Src: TSynCustomHighlighterRange);
Begin
   If (Src <> Nil) And (Src <> TSynCustomHighlighterRange(NullRange)) Then
   Begin
      Inherited Assign(Src);
      FMode := TmscrScriptHighlighterRange(Src).FMode;
      FNestedComments := TmscrScriptHighlighterRange(Src).FNestedComments;
      FTypeHelpers := TmscrScriptHighlighterRange(Src).FTypeHelpers;
      FBracketNestLevel := TmscrScriptHighlighterRange(Src).FBracketNestLevel;
      FLastLineCodeFoldLevelFix := TmscrScriptHighlighterRange(Src).FLastLineCodeFoldLevelFix;
      FPasFoldFixLevel := TmscrScriptHighlighterRange(Src).FPasFoldFixLevel;
   End;
End;

Function TmscrScriptHighlighterRange.MaxFoldLevel: integer;
Begin
   // Protect from overly mem consumption, by too many nested folds
   Result := 100;
End;

Procedure TmscrScriptHighlighterRange.IncBracketNestLevel;
Begin
   Inc(FBracketNestLevel);
End;

Procedure TmscrScriptHighlighterRange.DecBracketNestLevel;
Begin
   Dec(FBracketNestLevel);
End;

Procedure TmscrScriptHighlighterRange.DecLastLineCodeFoldLevelFix;
Begin
   Dec(FLastLineCodeFoldLevelFix);
End;

Procedure TmscrScriptHighlighterRange.DecLastLinePasFoldFix;
Begin
   Dec(FPasFoldFixLevel);
End;

{ TSynHighlighterPasRangeList }

Function TSynHighlighterPasRangeList.GetTSynPasRangeInfo(Index: integer): TSynPasRangeInfo;
Begin
   If (Index < 0) Or (Index >= Count) Then
   Begin
      Result.MinLevelRegion := 0;
      Result.EndLevelRegion := 0;
      Result.MinLevelIfDef := 0;
      Result.EndLevelIfDef := 0;
      exit;
   End;
   Result := TSynPasRangeInfo((ItemPointer[Index] + FItemOffset)^);
End;

Procedure TSynHighlighterPasRangeList.SetTSynPasRangeInfo(Index: integer; Const AValue: TSynPasRangeInfo);
Begin
   TSynPasRangeInfo((ItemPointer[Index] + FItemOffset)^) := AValue;
End;

Constructor TSynHighlighterPasRangeList.Create;
Begin
   Inherited;
   FItemOffset := ItemSize;
   ItemSize := FItemOffset + SizeOf(TSynPasRangeInfo);
End;

Initialization
   MakeIdentTable;
   RegisterPlaceableHighlighter(TmscrScriptHighlighter);

Finalization
   FreeAndNil(KeywordsList);

End.
