{*************************************************************************
                Copyright (c) PilotLogic Software House

  Package pl_MagicScript
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: LGPL-2.1 with link exception (Modified LGPL)
 *
 * The contents of this file are subject to the
 * GNU LIBRARY GENERAL PUBLIC LICENSE Version 2.1 (the "License")
 * with the following modification:
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent modules,
 * and to copy and distribute the resulting executable under terms of your choice,
 * provided that you also meet, for each linked independent module, the terms
 * and conditions of the license of that module. An independent module is a
 * module which is not derived from or based on this library. If you modify this
 * library, you may extend this exception to your version of the library, but
 * you are not obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * ***** END LICENSE BLOCK *****
 
***************************************************************************}

Unit mscoreengine;

Interface

{$I magicscript.inc}

Uses
   SysUtils, Classes, Variants, SyncObjs,
   Dialogs, TypInfo, lazutf8,
   Controls, StdCtrls, Forms,
   msbaseobjects,
   mscoreconst;

Const
   cnCatStr   = 'catString';
   cnCatDate  = 'catDate';
   cnCatConv  = 'catConv';
   cnCatFormat= 'catFormat';
   cnCatMath  = 'catMath';
   cnCatOther = 'catOther';

Type
   fmscrInteger = PtrInt;

{$IFDEF CPU64}
   IntObject = NativeInt;
{$ELSE}
   IntObject = integer;
{$ENDIF}

   TmscrStatement = Class;
   TmscrVarForBase = Class;
   TmscrVarForClass = Class;
   TmscrVarForRecord = Class;
   TmscrVarForProc = Class;
   TmscrVarForMethod = Class;
   TmscrVarForParamItem = Class;
   TmscrEngProperty = Class;
   TmscrDScannerItem = Class;
   TmscrDScanner = Class;
   TmscrExpressionBase = Class;
   TmscrSetExpression = Class;
   TmscrScript = Class;

   TmscrEmitOp = (mseoNone, mseoCreate, mseoFree);

   TmscrDScannerKind = (msskOther, msskVariable, msskStringArray, msskArray);

   TmscrVarType = (msvtInt, msvtBool, msvtFloat, msvtChar, msvtString, msvtClass, msvtRecord,
                   msvtArray, msvtVariant, msvtEnum, msvtConstructor, msvtInt64);

   TmscrOperatorType = (msopNone, msopGreat, msopLess, msopLessEq, msopGreatEq, msopNonEq, msopEq,
                        msopPlus, msopMinus, msopOr, msopXor, msopMul, msopDivFloat, msopDivInt, msopMod, msopAnd,
                        msopShl, msopShr, msopLeftBracket, msopRightBracket, msopNot, msopUnMinus, msopIn, msopIs);

   TmscrVarRecArray = Array Of TVarRec;

   TmscrTypeRec = Packed Record
      Typ: TmscrVarType;
      TypeName: String[64];
   End;

   TmscrRecGetValueElement = Function(aInstance: Pointer; aTypeInfo: PTypeInfo; Const aPropName: String): variant Of Object;
   TmscrRecSetValueElement = Procedure(aInstance: Pointer; aTypeInfo: PTypeInfo; Const aPropName: String; aValue: variant) Of Object;

   TmscrClassGetProperty = Function(aInstance: TObject; aClassType: TClass; Const aPropName: String; aCallVar: TmscrEngProperty): variant Of Object;
   TmscrClassSetProperty = Procedure(aInstance: TObject; aClassType: TClass; Const aPropName: String; aValue: variant; aCallVar: TmscrEngProperty) Of Object;

   TmscrCallMethod = Function(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant Of Object;

   TmscrRunLineEvent = Procedure(aSender: TmscrScript; Const aUnitName, aSrcPosition: String) Of Object;
   TmscrGetUnitEvent = Procedure(aSender: TmscrScript; Const aUnitName: String; Var aUnitText: String) Of Object;
   TmscrGetVariableValueEvent = Function(aVarName: String; aVarTyp: TmscrVarType; OldValue: variant): variant Of Object;

   TmscrItemList = Class(TObject)
   Protected
      FItems: TList;
   Protected
      Procedure Clear; Virtual;
   Public
      Constructor Create;
      Destructor Destroy; Override;
      Function Count: integer;
      Procedure Add(Item: TObject);
      Procedure Remove(Item: TObject);
   End;

   TmscrScript = Class(TComponent)
   Protected
      FCalledBreak: boolean;
      FCalledContinue: boolean;
      FCalledExit: boolean;
      FAddedBy: TObject;
      FOnGetFinalUnit: TmscrGetUnitEvent;
      FOnGetUnit: TmscrGetUnitEvent;
      FOnRunLine: TmscrRunLineEvent;
      FOnGetVarValue: TmscrGetVariableValueEvent;
      FParent: TmscrScript;
      FRTTIAdded: boolean;
      FUseExtendCharset: boolean;
      FItems: TStringList;
      FLines: TStrings;
      FMacros: TStrings;
      FMainProj: boolean;
      FStatement: TmscrStatement;
      FTerminated: boolean;
      FUnitLines: TStringList;
      FIncludePath: TStrings;
      FUseClassLateBinding: boolean;
      FEvaluteRiseError: boolean;
      FClearLocalVars: boolean;
      FSyntaxType: String;
      FErrorMsg: String;
      FErrorPos: String;
      FErrorUnit: String;
      FLastSrcPosition: String;
      FProjName: String;
      FProjRunning: TmscrScript;
      FIsRunning: boolean;
      Function GetItem(Index: integer): TmscrVarForBase;
      Procedure RunLine(Const UnitName, Index: String);
      Function GetVariables(Index: String): variant;
      Procedure SetVariables(Index: String; Const Value: variant);
      Procedure SetLines(Const Value: TStrings);
      Function GetProgName: String;
      Procedure SetProgName(Const Value: String);
   Public
      Constructor Create(AOwner: TComponent); Override;
      Destructor Destroy; Override;
      Procedure Add(Const Name: String; Item: TObject);
      Procedure AddCodeLine(Const UnitName, APos: String);
      Procedure AddRTTI;
      Procedure Remove(Item: TObject);
      Procedure RemoveItems(Owner: TObject);
      Procedure Clear;
      Procedure ClearItems(Owner: TObject);
      Procedure ClearRTTI;
      Function Count: integer;

      Function Register_Class(AClass: TClass; Const ParentClassStr: String): TmscrVarForClass; Dynamic;
      Procedure Register_Object(Const Name: String; Obj: TObject); Dynamic;
      Procedure Register_Component(Form: TComponent); Dynamic;
      Procedure Register_Form(Form: TComponent); Dynamic;
      Function Register_Record(Const aTypeName: String; pti: PTypeInfo): TmscrVarForRecord; Dynamic;
      Procedure Register_Const(Const Name, Typ: String; Const Value: variant); Dynamic;
      Procedure Register_Enum(Const Typ, Names: String); Dynamic;
      Procedure Register_EnumSet(Const Typ, Names: String); Dynamic;
      Procedure Register_Method(Const Syntax: String; CallEvent: TmscrCallMethod; Const Category: String = ''; Const Description: String = ''); Dynamic;
      Procedure Register_Variable(Const Name, Typ: String; Const Value: variant); Dynamic;
      Procedure Register_Type(Const TypeName: String; ParentType: TmscrVarType); Dynamic;

      Function CallFunction(Const Name: String; Const Params: variant; sGlobal: boolean = False): variant;
      Function CallFunction1(Const Name: String; Var Params: variant; sGlobal: boolean = False): variant;
      Function CallFunction2(Const Func: TmscrVarForProc; Const Params: variant): variant;
      Function Compile: boolean;
      Procedure Execute;
      Function Run: boolean;
      Procedure Terminate;
      Function Evaluate(Const Expression: String): variant;
      Function IsExecutableLine(LineN: integer; Const UnitName: String = ''): boolean;
      Function SetFinalCode(Stream: TStream): boolean;
      Function GetFinalCode(Stream: TStream): boolean;

      Function Find(Const Name: String): TmscrVarForBase;
      Function FindClass(Const Name: String): TmscrVarForClass;
      Function FindRecord(Const Name: String): TmscrVarForRecord;
      Function FindLocal(Const Name: String): TmscrVarForBase;

      Property AddedBy: TObject Read FAddedBy Write FAddedBy;
      Property ClearLocalVars: boolean Read FClearLocalVars Write FClearLocalVars;
      Property ErrorMsg: String Read FErrorMsg Write FErrorMsg;
      Property ErrorPos: String Read FErrorPos Write FErrorPos;
      Property ErrorUnit: String Read FErrorUnit Write FErrorUnit;
      Property UseExtendCharset: boolean Read FUseExtendCharset Write FUseExtendCharset;
      Property Items[Index: integer]: TmscrVarForBase Read GetItem;
      Property Macros: TStrings Read FMacros;
      Property MainProj: boolean Read FMainProj Write FMainProj;
      Property Parent: TmscrScript Read FParent Write FParent;
      Property ProjRunning: TmscrScript Read FProjRunning;
      Property ProjName: String Read GetProgName Write SetProgName;
      Property Statement: TmscrStatement Read FStatement;
      Property Variables[Index: String]: variant Read GetVariables Write SetVariables;
      Property IncludePath: TStrings Read FIncludePath;
      Property UseClassLateBinding: boolean Read FUseClassLateBinding Write FUseClassLateBinding;
      Property EvaluteRiseError: boolean Read FEvaluteRiseError;
      Property IsRunning: boolean Read FIsRunning;
   Published
      Property Lines: TStrings Read FLines Write SetLines;
      Property SyntaxType: String Read FSyntaxType Write FSyntaxType;
      Property OnGetFinalUnit: TmscrGetUnitEvent Read FOnGetFinalUnit Write FOnGetFinalUnit;
      Property OnGetUnit: TmscrGetUnitEvent Read FOnGetUnit Write FOnGetUnit;
      Property OnRunLine: TmscrRunLineEvent Read FOnRunLine Write FOnRunLine;
      Property OnGetVarValue: TmscrGetVariableValueEvent Read FOnGetVarValue Write FOnGetVarValue;
   End;

   TmscrStatement = Class(TmscrItemList)
   Private
      FScrProject: TmscrScript;
      FSrcPosition: String;
      FUnitName: String;
      Function GetItem(Index: integer): TmscrStatement;
      Procedure RunLine;
   Public
      Constructor Create(aProject: TmscrScript; Const UnitName, SrcPosition: String); Virtual;
      Procedure Execute; Virtual;
      Property Items[Index: integer]: TmscrStatement Read GetItem;
   End;

   TmscrCodeStmAssign_Base = Class(TmscrStatement)
   Private
      FDScanner: TmscrDScanner;
      FExpression: TmscrExpressionBase;
      FVar: TmscrVarForBase;
      FExpr: TmscrVarForBase;
   Public
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Procedure Optimize;
      Property DScanner: TmscrDScanner Read FDScanner Write FDScanner;
      Property Expression: TmscrExpressionBase Read FExpression Write FExpression;
   End;

   TmscrCodeStmAssign_Plus = Class(TmscrCodeStmAssign_Base)
   Public
      Procedure Execute; Override;
   End;

   TmscrCodeStmAssign_Minus = Class(TmscrCodeStmAssign_Base)
   Public
      Procedure Execute; Override;
   End;

   TmscrCodeStmAssign_Mul = Class(TmscrCodeStmAssign_Base)
   Public
      Procedure Execute; Override;
   End;

   TmscrCodeStmAssign_Div = Class(TmscrCodeStmAssign_Base)
   Public
      Procedure Execute; Override;
   End;

   TmscrCodeStm_Call = Class(TmscrStatement)
   Private
      FDScanner: TmscrDScanner;
      FModificator: String;
   Public
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property DScanner: TmscrDScanner Read FDScanner Write FDScanner;
      Property Modificator: String Read FModificator Write FModificator;
   End;

   TmscrCodeStm_If = Class(TmscrStatement)
   Private
      FCondition: TmscrExpressionBase;
      FElsesrcstm: TmscrStatement;
   Public
      Constructor Create(aProject: TmscrScript; Const UnitName, SrcPosition: String); Override;
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property Condition: TmscrExpressionBase Read FCondition Write FCondition;
      Property Elsesrcstm: TmscrStatement Read FElsesrcstm Write FElsesrcstm;
   End;

   TmscrCaseSelector = Class(TmscrStatement)
   Private
      FSetExpression: TmscrSetExpression;
   Public
      Destructor Destroy; Override;
      Function Check(Const Value: variant): boolean;
      Property SetExpression: TmscrSetExpression Read FSetExpression Write FSetExpression;
   End;

   TmscrCodeStm_Case = Class(TmscrStatement)
   Private
      FCondition: TmscrExpressionBase;
      FElsesrcstm: TmscrStatement;
   Public
      Constructor Create(aProject: TmscrScript; Const UnitName, SrcPosition: String); Override;
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property Condition: TmscrExpressionBase Read FCondition Write FCondition;
      Property Elsesrcstm: TmscrStatement Read FElsesrcstm Write FElsesrcstm;
   End;

   TmscrCodeStm_Repeat = Class(TmscrStatement)
   Private
      FCondition: TmscrExpressionBase;
      FInverseCondition: boolean;
   Public
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property Condition: TmscrExpressionBase Read FCondition Write FCondition;
      Property InverseCondition: boolean Read FInverseCondition Write FInverseCondition;
   End;

   TmscrCodeStm_While = Class(TmscrStatement)
   Private
      FCondition: TmscrExpressionBase;
   Public
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property Condition: TmscrExpressionBase Read FCondition Write FCondition;
   End;

   TmscrCodeStm_For = Class(TmscrStatement)
   Private
      FValueBegin: TmscrExpressionBase;
      FValueEnd: TmscrExpressionBase;
      FDown: boolean;
      FVariable: TmscrVarForBase;
   Public
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property ValueBegin: TmscrExpressionBase Read FValueBegin Write FValueBegin;
      Property Down: boolean Read FDown Write FDown;
      Property ValueEnd: TmscrExpressionBase Read FValueEnd Write FValueEnd;
      Property Variable: TmscrVarForBase Read FVariable Write FVariable;
   End;

   TmscrCodesrcstmry = Class(TmscrStatement)
   Private
      FIsExcept: boolean;
      FExceptSrcStm: TmscrStatement;
   Public
      Constructor Create(aProject: TmscrScript; Const UnitName, SrcPosition: String); Override;
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property IsExcept: boolean Read FIsExcept Write FIsExcept;
      Property ExceptSrcStm: TmscrStatement Read FExceptSrcStm Write FExceptSrcStm;
   End;

   TmscrCodeStm_Break = Class(TmscrStatement)
   Public
      Procedure Execute; Override;
   End;

   TmscrCodeStm_Continue = Class(TmscrStatement)
   Public
      Procedure Execute; Override;
   End;

   TmscrCodeStm_Exit = Class(TmscrStatement)
   Public
      Procedure Execute; Override;
   End;

   TmscrCodeStm_With = Class(TmscrStatement)
   Private
      FDScanner: TmscrDScanner;
      FVariable: TmscrVarForBase;
   Public
      Destructor Destroy; Override;
      Procedure Execute; Override;
      Property DScanner: TmscrDScanner Read FDScanner Write FDScanner;
      Property Variable: TmscrVarForBase Read FVariable Write FVariable;
   End;

   TmscrVarForBase = Class(TmscrItemList)
   Private
      FName: String;      
      FDeclaration: String;
      FAddedBy: TObject;
      FNeedResult: boolean;
      FRefItem: TmscrVarForBase;
      FSrcPosition: String;
      FSourceUnit: String;
      FTyp: TmscrVarType;
      FTypeName: String;
      FUppercaseName: String;
      FValue: variant;
      FIsMacro: boolean;
      FIsReadOnly: boolean;
      FOnGetVarValue: TmscrGetVariableValueEvent;
      Function GetParam(Index: integer): TmscrVarForParamItem;
      Function GetPValue: PVariant;
   Protected
      Procedure SetValue(Const Value: variant); Virtual;
      Function GetValue: variant; Virtual;
   Public
      Constructor Create(Const AName: String; ATyp: TmscrVarType; Const ATypeName: String);
      Function GetFullTypeName: String;
      Function GetNumberOfRequiredParams: integer;
      Property AddedBy: TObject Read FAddedBy Write FAddedBy;
      Property IsMacro: boolean Read FIsMacro Write FIsMacro;
      Property IsReadOnly: boolean Read FIsReadOnly Write FIsReadOnly;
      Property Name: String Read FName;        
      Property Declaration: String Read FDeclaration Write FDeclaration;
      Property NeedResult: boolean Read FNeedResult Write FNeedResult;
      Property Params[Index: integer]: TmscrVarForParamItem Read GetParam; Default;
      Property PValue: PVariant Read GetPValue;
      Property RefItem: TmscrVarForBase Read FRefItem Write FRefItem;
      Property SrcPosition: String Read FSrcPosition Write FSrcPosition;
      Property SourceUnit: String Read FSourceUnit Write FSourceUnit;
      Property Typ: TmscrVarType Read FTyp Write FTyp;
      Property TypeName: String Read FTypeName Write FTypeName;
      Property Value: variant Read GetValue Write SetValue;
      Property OnGetVarValue: TmscrGetVariableValueEvent Read FOnGetVarValue Write FOnGetVarValue;
   End;

   TmscrVarForVariable = Class(TmscrVarForBase)
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   End;

   TmscrVarForType = Class(TmscrVarForBase)
   End;

   TmscrVarForString = Class(TmscrVarForVariable)
   Private
      FStr: String;
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   End;

   TmscrVarForParamItem = Class(TmscrVarForBase)
   Private
      FIsOptional: boolean;
      FIsVarParam: boolean;
      FDefValue: variant;
   Public
      Constructor Create(Const AName: String; ATyp: TmscrVarType; Const ATypeName: String; AIsOptional, AIsVarParam: boolean);
      Property DefValue: variant Read FDefValue Write FDefValue;
      Property IsOptional: boolean Read FIsOptional;
      Property IsVarParam: boolean Read FIsVarParam;
   End;

   TmscrEngLocalVariables = Class(TObject)
   Protected
      FValue: variant;
      FVariableLink: TmscrVarForBase;
   End;

   TmscrVarForProc = Class(TmscrVarForBase)
   Private
      FExecuting: boolean;
      FIsFunc: boolean;
      FScrProject: TmscrScript;
      FVarsStack: TList;
   Protected
      Function GetValue: variant; Override;
   Public
      Constructor Create(Const AName: String; ATyp: TmscrVarType; Const ATypeName: String; AParent: TmscrScript; AIsFunc: boolean = True);
      Destructor Destroy; Override;
      Function SaveLocalVariables: integer;
      Procedure RestoreLocalVariables(StackIndex: integer; bSkipVarParams: boolean = False; dItem: TmscrDScannerItem = Nil);
      Property Executing: boolean Read FExecuting;
      Property Project: TmscrScript Read FScrProject;
      Property IsFunc: boolean Read FIsFunc;
   End;

   TmscrExpressionBase = Class(TmscrVarForBase)
   End;

   TmscrEngCustom = Class(TmscrVarForBase)
   Private
      FParentRef: TmscrVarForBase;
      FParentValue: variant;
      FScrProject: TmscrScript;
   Public
      Property aProj: TmscrScript Read FScrProject Write FScrProject;
      Property ParentRef: TmscrVarForBase Read FParentRef Write FParentRef;
      Property ParentValue: variant Read FParentValue Write FParentValue;
   End;

   TmscrEngArray = Class(TmscrEngCustom)
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   Public
      Constructor Create(Const AName: String; DimCount: integer; Typ: TmscrVarType; Const TypeName: String);
      Destructor Destroy; Override;
   End;

   TmscrEngString = Class(TmscrEngCustom)
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   Public
      Constructor Create;
   End;

   TmscrEngProperty = Class(TmscrEngCustom)
   Private
      FClassRef: TClass;
      FIsPublished: boolean;
      FOnGetValue: TmscrClassGetProperty;
      FOnSetValue: TmscrClassSetProperty;
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   Public
      Property IsPublished: boolean Read FIsPublished;
      Property OnGetValue: TmscrClassGetProperty Read FOnGetValue Write FOnGetValue;
      Property OnSetValue: TmscrClassSetProperty Read FOnSetValue Write FOnSetValue;
   End;

   TmscrEngRecElement = Class(TmscrEngCustom)
   Private
      FOnGetValue: TmscrRecGetValueElement;
      FOnSetValue: TmscrRecSetValueElement;
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   Public
      Property OnGetValue: TmscrRecGetValueElement Read FOnGetValue Write FOnGetValue;
      Property OnSetValue: TmscrRecSetValueElement Read FOnSetValue Write FOnSetValue;
   End;

   TmscrVarForMethod = Class(TmscrEngCustom)
   Private
      FCategory: String;
      FClassRef: TClass;
      FDescription: String;
      FIndexMethod: boolean;
      FOnCall: TmscrCallMethod;
      FSetValue: variant;
      FSyntax: String;
      FVarArray: variant;
      Function GetVParam(Index: integer): variant;
      Procedure SetVParam(Index: integer; Const Value: variant);
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   Public
      Constructor Create(Const Syntax: String; Script: TmscrScript);
      Destructor Destroy; Override;
      Property Params[Index: integer]: variant Read GetVParam Write SetVParam; Default;
      Property Category: String Read FCategory Write FCategory;
      Property Description: String Read FDescription Write FDescription;
      Property IndexMethod: boolean Read FIndexMethod;
      Property Syntax: String Read FSyntax;
      Property OnCall: TmscrCallMethod Read FOnCall Write FOnCall;
   End;

   TmscrEngComponent = Class(TmscrEngCustom)
   Private
      FComponent: TComponent;
   Protected
      Function GetValue: variant; Override;
   Public
      Constructor Create(Component: TComponent);
   End;

   TmscrCustomEvent = Class(TObject)
   Private
      FInstance: TObject;
      FHandler: TmscrVarForProc;
   Protected
      Procedure CallHandler(Params: Array Of Const);
   Public
      Constructor Create(AObject: TObject; AHandler: TmscrVarForProc); Virtual;
      Function GetMethod: Pointer; Virtual; Abstract;
      class function GetDeclaration: String; Virtual;
      Property Handler: TmscrVarForProc Read FHandler;
      Property Instance: TObject Read FInstance;
   End;

   TmscrCustomEventClass = Class Of TmscrCustomEvent;

   TmscrEngEvent = Class(TmscrEngCustom)
   Private
      FEvent: TmscrCustomEventClass;
      FClassRef: TClass;
   Protected
      Procedure SetValue(Const Value: variant); Override;
      Function GetValue: variant; Override;
   Public
      Constructor Create(Const Name: String; AEvent: TmscrCustomEventClass);
   End;

   TmscrVarForClass = Class(TmscrVarForBase)
   Private
      FParentClassStr: String;
      FClassRef: TClass;
      FDefProperty: TmscrEngCustom;
      FMembers: TmscrItemList;
      FScrProject: TmscrScript;
      Procedure AddComponent(c: TComponent);
      Procedure AddPublishedProperties(AClass: TClass);
      Function GetMembers(Index: integer): TmscrEngCustom;
      Function GetMembersCount: integer;
   Protected
      Function GetValue: variant; Override;
   Public
      Constructor Create(AClass: TClass; Const ParentClassStr: String);
      Destructor Destroy; Override;

      Procedure Register_Constructor(Syntax: String; CallEvent: TmscrCallMethod);
      Procedure Register_Property(Const Name, Typ: String; GetEvent: TmscrClassGetProperty; SetEvent: TmscrClassSetProperty = Nil);
      Procedure Register_DefaultProperty(Const Name, Params, Typ: String; CallEvent: TmscrCallMethod; AReadOnly: boolean = False);
      Procedure Register_IndexProperty(Const Name, Params, Typ: String; CallEvent: TmscrCallMethod; AReadOnly: boolean = False);
      Procedure Register_Method(Const Syntax: String; CallEvent: TmscrCallMethod); Overload;
      Procedure Register_Event(Const Name: String; AEvent: TmscrCustomEventClass; Const ADeclaration: String='');

      Function Find(Const Name: String): TmscrEngCustom;
      Property ClassRef: TClass Read FClassRef;
      Property ParentClassStr: String Read FParentClassStr;
      Property DefProperty: TmscrEngCustom Read FDefProperty;
      Property MembersCount: integer Read GetMembersCount;
      Property Members[Index: integer]: TmscrEngCustom Read GetMembers;
   End;

   TmscrVarForRecord = Class(TmscrVarForBase)
   Protected
      FTypeInfoStr: String;
      FTypeInfo: PTypeInfo;
      FRecData: Pointer;
      FRecDataSize: int64;
      FDefProperty: TmscrEngCustom;
      FMembers: TmscrItemList;
      FScrProject: TmscrScript;
      Procedure AddComponent(c: TComponent);
      Function  GetMembers(Index: integer): TmscrEngCustom;
      Function  GetMembersCount: integer;
   Public
      Constructor Create(ATypeInfo: PTypeInfo; Const aTypeInfoStr: String);
      Destructor  Destroy; Override;

      Procedure Register_Element(Const Name, Typ: String; GetEvent: TmscrRecGetValueElement; SetEvent: TmscrRecSetValueElement = Nil);

      Function Find(Const Name: String): TmscrEngCustom;
      Property TypeInfo: PTypeInfo Read FTypeInfo;
      Property TypeInfoStr: String Read FTypeInfoStr;
      Property DefProperty: TmscrEngCustom Read FDefProperty;
      Property MembersCount: integer Read GetMembersCount;
      Property Members[Index: integer]: TmscrEngCustom Read GetMembers;
   End;

   TmscrDScannerItem = Class(TmscrItemList)
   Private
      FRef: TmscrVarForBase;
      FFlag: boolean;
      FSrcPosition: String;
      Function GetItem(Index: integer): TmscrExpressionBase;
   Public
      Property Items[Index: integer]: TmscrExpressionBase Read GetItem; Default;
      Property Flag: boolean Read FFlag Write FFlag;
      Property Ref: TmscrVarForBase Read FRef Write FRef;
      Property SrcPosition: String Read FSrcPosition Write FSrcPosition;
   End;

   TmscrDScanner = Class(TmscrVarForBase)
   Private
      FKind: TmscrDScannerKind;
      FMainProj: TmscrScript;
      FScrProject: TmscrScript;
      FRef1: TmscrVarForBase;
      FRef2: TmscrDScannerItem;
      FLateBindingDATASource: TmscrDATAItem;
      Procedure CheckLateBinding;
      Function  DoCalc(Const AValue: variant; Flag: boolean): variant;
      Function  GetItem(Index: integer): TmscrDScannerItem;
   Protected
      Function  GetValue: variant; Override;
      Procedure SetValue(Const Value: variant); Override;
   Public
      Constructor Create(aProject: TmscrScript);
      Destructor Destroy; Override;
      Procedure Borrow(ADScanner: TmscrDScanner);
      Procedure Finalize;
      Property Items[Index: integer]: TmscrDScannerItem Read GetItem; Default;
      Property Kind: TmscrDScannerKind Read FKind;
      Property LateBindingDATASource: TmscrDATAItem Read FLateBindingDATASource Write FLateBindingDATASource;
   End;

   TmscrVarForVariableDScanner = Class(TmscrDScanner)
   Protected
      Function GetValue: variant; Override;
      Procedure SetValue(Const Value: variant); Override;
   End;

   TmscrStringDScanner = Class(TmscrDScanner)
   Protected
      Function GetValue: variant; Override;
      Procedure SetValue(Const Value: variant); Override;
   End;

   TmscrArrayDScanner = Class(TmscrDScanner)
   Protected
      Function GetValue: variant; Override;
      Procedure SetValue(Const Value: variant); Override;
   End;

   TmscrSetExpression = Class(TmscrVarForBase)
   Private
      Function GetItem(Index: integer): TmscrExpressionBase;
   Protected
      Function GetValue: variant; Override;
   Public
      Function Check(Const Value: variant): boolean;
      Property Items[Index: integer]: TmscrExpressionBase Read GetItem;
   End;

   TmscrRTTILibrary = Class(TObject)
   Private
      FScript: TmscrScript;
   Public
      Constructor Create(AScript: TmscrScript); Virtual;
      Property Script: TmscrScript Read FScript;
   End;

//============ System Events ========================
Type
   TmscrNotifyEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject);
      Function GetMethod: Pointer; Override;  
      class function GetDeclaration: String; Override;
   End;

   TmscrMouseEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
      Function GetMethod: Pointer; Override;   
      class function GetDeclaration: String; Override;
   End;

   TmscrMouseMoveEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject; Shift: TShiftState; X, Y: integer);
      Function GetMethod: Pointer; Override;   
      class function GetDeclaration: String; Override;
   End;

   TmscrKeyEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject; Var Key: word; Shift: TShiftState);
      Function GetMethod: Pointer; Override;    
      class function GetDeclaration: String; Override;
   End;

   TmscrKeyPressEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject; Var Key: char);
      Function GetMethod: Pointer; Override;  
      class function GetDeclaration: String; Override;
   End;

   TmscrCloseEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject; Var Action: TCloseAction);
      Function GetMethod: Pointer; Override; 
      class function GetDeclaration: String; Override;
   End;

   TmscrCloseQueryEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject; Var CanClose: boolean);
      Function GetMethod: Pointer; Override; 
      class function GetDeclaration: String; Override;
   End;

   TmscrCanResizeEvent = Class(TmscrCustomEvent)
   Public
      Procedure DoEvent(Sender: TObject; Var NewWidth, NewHeight: integer; Var Resize: boolean);
      Function GetMethod: Pointer; Override; 
      class function GetDeclaration: String; Override;
   End;

//=======================================

   TmscrExpressionNode = Class(TmscrVarForBase)
   Private
      FParent, FLeft, FRight: TmscrExpressionNode;
      Procedure AddNode(Node: TmscrExpressionNode);
      Procedure RemoveNode(Node: TmscrExpressionNode);
   Public
      Destructor Destroy; Override;
      Function Priority: integer; Virtual; Abstract;
   End;

   TmscrActionNode = Class(TmscrExpressionNode)
   Public
      Constructor Create(Const AValue: variant);
      Function Priority: integer; Override;
   End;

   TmscrActionBase = Class(TmscrExpressionNode)
   Private
      FOp: TmscrOperatorType;
      FOptimizeInt: boolean;
      FOptimizeBool: boolean;
   Public
      Constructor Create(Op: TmscrOperatorType);
      Function Priority: integer; Override;
   End;

   TmscrDScannerNode = Class(TmscrActionNode)
   Private
      FDScanner: TmscrDScanner;
      FVar: TmscrVarForBase;
   Protected
      Function GetValue: variant; Override;
   Public
      Constructor Create(ADScanner: TmscrDScanner);
      Destructor Destroy; Override;
   End;

   TmscrSetNode = Class(TmscrActionNode)
   Private
      FSetExpression: TmscrSetExpression;
   Protected
      Function GetValue: variant; Override;
   Public
      Constructor Create(ASet: TmscrSetExpression);
      Destructor Destroy; Override;
   End;

   TmscrExpression = Class(TmscrExpressionBase)
   Private
      FCurNode: TmscrExpressionNode;
      FNode: TmscrExpressionNode;
      FScript: TmscrScript;
      FSource: String;
      Procedure AddOperand(Node: TmscrExpressionNode);
   Protected
      Function GetValue: variant; Override;
      Procedure SetValue(Const Value: variant); Override;
   Public
      Constructor Create(Script: TmscrScript);
      Destructor Destroy; Override;
      Procedure Register_Const(Const AValue: variant);
      Procedure Register_ConstWithType(Const AValue: variant; aTyp: TmscrVarType);
      Procedure AddDScanner(ADScanner: TmscrDScanner);
      Procedure AddOperator(Const Op: String);
      Procedure AddSet(ASet: TmscrSetExpression);
      Function Finalize: String;
      Function Optimize(DScanner: TmscrDScanner): String;
      Function SingleItem: TmscrVarForBase;
      Property Source: String Read FSource Write FSource;
   End;

//======================================

   TmscrEngParser = Class(TObject)
   Private
      FErrorPos: String;
      FGrammar: TmscrDATADocument;
      FFinalScript: TmscrDATADocument;
      FNeedDeclareVars: boolean;
      FParser: TmscrParser;
      FScrProject: TmscrScript;
      FProjRoot: TmscrDATAItem;
      FRoot: TmscrDATAItem;
      FUnitName: String;
      FUsesList: TStrings;
      FWithList: TStringList;
      Function  propPos(xi: TmscrDATAItem): String;
      Procedure ErrorPos(xi: TmscrDATAItem);
      Function  FindError(Const aErr: integer): String;
      Function  FindClass(Const TypeName: String): TmscrVarForClass;
      Function  FindRecord(Const TypeName: String): TmscrVarForRecord;
      Function  FindVar(aProj: TmscrScript; Const Name: String): TmscrVarForBase;
      Function  FindType(s: String): TmscrVarType;
      Procedure CheckIdent(aProj: TmscrScript; Const Name: String);
      Procedure CheckTypeCompatibility(Var1, Var2: TmscrVarForBase);
      Function  CreateVar(xi: TmscrDATAItem; aProj: TmscrScript; Const Name: String; Statement: TmscrStatement = Nil; CreateParam: boolean = False; IsVarParam: boolean = False): TmscrVarForBase;

      Function  Compile_Set(xi: TmscrDATAItem; aProj: TmscrScript): TmscrSetExpression;
      Function  Compile_Expression(xi: TmscrDATAItem; aProj: TmscrScript): TmscrExpression;
      Procedure Compile_Uses(xi: TmscrDATAItem; aProj: TmscrScript);
      Procedure Compile_Var(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Const(xi: TmscrDATAItem; aProj: TmscrScript);
      Procedure Compile_Parameters(xi: TmscrDATAItem; v: TmscrVarForProc);
      Procedure Compile_ProcA(xi: TmscrDATAItem; aProj: TmscrScript);
      Procedure Compile_ProcB(xi: TmscrDATAItem; aProj: TmscrScript);
      Procedure Compile_FuncA(xi: TmscrDATAItem; aProj: TmscrScript);
      Procedure Compile_FuncB(xi: TmscrDATAItem; aProj: TmscrScript);
      Procedure Compile_Assign(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Call(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_If(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_For(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_While(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Repeat(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Case(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Try(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Break(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Continue(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Exit(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Return(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_With(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Delete(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_mscodecompound(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_SrcStm(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
      Procedure Compile_Project(xi: TmscrDATAItem; aProj: TmscrScript);
   Public
      Constructor Create(aProject: TmscrScript);
      Destructor Destroy; Override;
      Procedure InitGrammar;
      Function FinalScript_Build(Const Text: String): boolean;
      Procedure FinalScript_Parse;
      Function DoDScanner(xi: TmscrDATAItem; aProj: TmscrScript; EmitOp: TmscrEmitOp = mseoNone): TmscrDScanner;
      Property FinalScript: TmscrDATADocument Read FFinalScript;
   End;

{$IFDEF MAGSCR_USEOLE}
Type
   TmscrEngOLE = Class(TmscrEngCustom)
   Private
      Function DispatchInvoke(Const ParamArray: Variant; ParamCount: Integer; Flags: Word): Variant;
   Protected
      Procedure SetValue(Const Value: Variant); Override;
      Function GetValue: Variant; Override;
   Public
      Constructor Create(Const AName: String);
   End;
{$ENDIF}

Type
   TmscrSysFunctions = Class(TmscrRTTILibrary)
   Private
      FCatConv: String;
      FCatDate: String;
      FCatFormat: String;
      FCatMath: String;
      FCatOther: String;
      FCatStr: String;
      Function CallMethod1(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
      Function CallMethod2(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
      Function CallMethod3(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
      Function CallMethod4(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
      Function CallMethod5(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
      Function CallMethod6(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
      Function CallMethod7(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
   Public
      Constructor Create(AScript: TmscrScript); Override;
   End;

   TmscrPascalScript = Class(TmscrScript)
   End;

//====================================================

Procedure GenerateDATAContents(aProj: TmscrScript; Item: TmscrDATAItem; FunctionsOnly: boolean = False);
Procedure GenerateMembers(aProj: TmscrScript; cl: TClass; Item: TmscrDATAItem);
Function  TypesCompatible(Typ1, Typ2: TmscrTypeRec; Script: TmscrScript): boolean;
Function  AssignCompatible(Var1, Var2: TmscrVarForBase; Script: TmscrScript): boolean;
Function  VarRecToVariant(v: TVarRec): variant;
Function  StrToVarType(Const TypeName: String; Script: TmscrScript): TmscrVarType;
Procedure VariantToVarRec(v: variant; Var ar: TmscrVarRecArray; Var sPtrList: TList);
Procedure ClearVarRec(Var ar: TmscrVarRecArray; Var sPtrList: TList);
Function  ParserStringToVariant(s: String): variant;
Function  ParseMethodSyntax(Const Syntax: String; Script: TmscrScript): TmscrVarForBase;

Function MSCR_PosToPoint(Const ErrorPos: String): TPoint;
Function MSCR_GlobalScripter: TmscrScript;
Function MSCR_IsGlobalScripterExist: boolean;
Function MSCR_RTTILibraries: TList;

//====================================================

Function VarToObject(Const v: variant): TObject; //  Converts a variant type to TObject.
Function ObjectToVar(Const obj: TObject): variant; //  Converts a TObject to Variant type

Function VarToClass(Const v: variant): TClass; //  Converts a variant type to TClass.
Function ClassToVar(Const obj: TClass): variant; //  Converts a TClass to Variant type

Function PointerToVar(p: pointer): variant; //  Converts a Pointer to Variant type
Function VarToPointer(v: variant): pointer; //  Converts a variant type to Pointer.

Function VarToIntObject(Const v: variant): IntObject; //  Converts a variant type to an integer (32 or 64 bit) value that represents the object
Function VarToInteger(v: variant): integer; //  Converts a variant type to integer.
Function VarToInt64(v: variant): int64; //  Converts a variant type to Int64.
Function VarToUInt64(v: variant): uint64; //  Converts a variant type to UInt64.
Function VarToFloat(v: variant): double; //  Converts a variant to double.
Function VarToBoolean(v: variant): boolean; //  Converts a variant to boolean. Value is false if variant is zero, true otherwise.
Function VarToWideChar(v: variant): widechar; //  Converts a variant to WideChar.
Function VarToChar(v: variant): char; //  Converts a variant to Char.
Function VarContent(v: variant): String; //  Returns a text description describing the content of the variant.

//  Converts a variant type to an enumerated value. The variant can have the enumerated name
//  in string type. For example, 'alClient'. You must provide the PropType of enumerated type so
//  such conversion is possible.
Function VarToEnumInteger(v: variant; PropType: PTypeInfo): integer;

//  Converts a variant type to a set type. The variant is an array of string with the names of enumerated
//  types belonging to the set. The PropType parameter must receive the RTTI info about the enumerated type.
Function VarToSet(v: variant; PropType: PTypeInfo): integer;

//======================================
Var
   VarMSCR_ConcoleOut: TMemo = Nil;

   //===================================================
   //===================================================
   //===================================================

Implementation

{$IFDEF MAGSCR_USEOLE}
Uses
   Windows, ActiveX, ComObj;
{$ENDIF}

Var
   FGlobalScripter: TmscrScript = Nil;
   FGlobalScripterIsNIL: boolean = False;

   FRTTILibraries: TList = Nil;
   FRTTILibrariesIsNIL: boolean = False;

Const
   _SyntaxType = 'ObjcPascalScript';

   _MAINGRAMMAR = '<?xml version="1.0"?><language text="ObjcPascalScript"><types/><empty/><program><optional>' + '<keyword text="PROGRAM"/><ident errornum="1"/><char text=";" errornum="5"/></optional>' +
      '<optional><usesclause/></optional><block/><char text="." errornum="6"/></program>' + '<usesclause node="uses"><keyword text="USES"/><loop text=","><string add="file" errornum="1"/>' +
      '</loop><char text=";" errornum="5"/></usesclause><block><optloop><declsection/></optloop>' + '<mscodecompound errornum="10"/></block><declsection><switch><constsection/><varsection/><proceduredeclsection/>' +
      '</switch></declsection><constsection><keyword text="CONST"/><loop><constantdecl/></loop></constsection>' + '<constantdecl node="const"><ident add="ident" errornum="1" term="1"/><char text="=" errornum="9"/>' +
      '<expression errornum="2"/><char text=";" errornum="5"/></constantdecl><varsection><keyword text="VAR"/>' + '<loop><varlist/><char text=";" errornum="5"/></loop></varsection><varlist node="var"><loop text=",">' +
      '<ident add="ident" errornum="1" term="1"/></loop><char text=":" errornum="4"/><typeident errornum="1"/>' + '<optional><initvalue/></optional></varlist><typeident><optional><array/></optional><ident add="type"/>' +
      '</typeident><array node="array"><keyword text="ARRAY"/><optional><char text="["/><loop text=",">' + '<arraydim errornum="2"/></loop><char text="]" errornum="8"/></optional><keyword text="OF" errornum="12"/>' +
      '</array><arraydim node="dim"><switch><dojob><expression/><char text="."/><char text="." skip="0"/>' + '<expression/></dojob><expression/></switch></arraydim><initvalue node="init"><char text="="/>' +
      '<expression errornum="2"/></initvalue><expression node="expr"><simpleexpression/><optloop><relop/>' + '<simpleexpression/></optloop></expression><simpleexpression><optional><char text="-" add="op" addtext="unminus"/>' +
      '</optional><term/><optloop><addop/><term/></optloop></simpleexpression><term><factor/>' + '<optloop><mulop/><factor/></optloop></term><factor><switch><DScanner/><number add="number"/>' +
      '<string add="string"/><dojob><char text="(" add="op"/><expression errornum="2"/><char text=")" add="op" errornum="7"/>' + '</dojob><dojob><keyword text="NOT" add="op" addtext="not"/><factor errornum="2"/></dojob>' +
      '<dojob><char text="["/><setconstructor errornum="2"/><char text="]" errornum="8"/></dojob><dojob>' + '<char text="&#60;"/><frstring/><char text="&#62;" errornum="20"/></dojob></switch></factor><setconstructor node="set">' +
      '<loop text=","><setnode/></loop></setconstructor><setnode><expression/><optional><char text="." add="range"/>' + '<char text="." errornum="6"/><expression/></optional></setnode><relop><switch><dojob><char text="&#62;"/>' +
      '<char text="=" skip="0" add="op" addtext="&#62;="/></dojob><dojob><char text="&#60;"/>' + '<char text="&#62;" skip="0" add="op" addtext="&#60;&#62;"/></dojob><dojob><char text="&#60;"/>' +
      '<char text="=" skip="0" add="op" addtext="&#60;="/></dojob><char text="&#62;" add="op" addtext="&#62;"/>' + '<char text="&#60;" add="op" addtext="&#60;"/><char text="=" add="op" addtext="="/><keyword text="IN" add="op" addtext="in"/>' +
      '<keyword text="IS" add="op" addtext="is"/></switch></relop><addop><switch><char text="+" add="op"/>' + '<char text="-" add="op"/><keyword text="OR" add="op" addtext="or"/><keyword text="XOR" add="op" addtext="xor"/>' +
      '</switch></addop><mulop><switch><char text="*" add="op"/><char text="/" add="op"/><keyword text="DIV" add="op" addtext="div"/>' +
      '<keyword text="MOD" add="op" addtext="mod"/><keyword text="AND" add="op" addtext="and"/><keyword text="SHL" add="op" addtext="shl"/>' +
      '<keyword text="SHR" add="op" addtext="shr"/></switch></mulop><DScanner node="dsgn"><optional><char text="@" add="addr"/>' +
      '</optional><ident add="node"/><optloop><switch><dojob><char text="."/><keyword add="node" errornum="1"/></dojob><dojob>' + '<char text="[" add="node"/><exprlist errornum="2"/><char text="]" errornum="8"/></dojob><dojob><char text="("/>' +
      '<exprlist errornum="2"/><char text=")" errornum="7"/></dojob></switch></optloop></DScanner><exprlist><optloop text=",">' + '<expression/></optloop></exprlist><statement><optswitch><simplestatement/><structsrcstm/></optswitch>' +
      '</statement><srcstmlist><loop text=";"><statement/></loop></srcstmlist><simplestatement><switch><codeassign/><mscodecall/>' +
      '<switch><keyword text="BREAK" node="break"/><keyword text="CONTINUE" node="continue"/><keyword text="EXIT" node="exit"/>' +
      '</switch></switch></simplestatement><codeassign node="codeassign"><DScanner/><char text=":"/><char text="=" skip="0" errornum="9"/>' +
      '<expression errornum="2"/></codeassign><mscodecall node="mscodecall"><DScanner/></mscodecall><structsrcstm><switch><mscodecompound/>' +
      '<conditionalsrcstm/><loopsrcstm/><mscodetry/><withsrcstm/></switch></structsrcstm><mscodecompound node="mscodecompound">' +
      '<keyword text="BEGIN"/><srcstmlist/><keyword text="END" errornum="5"/></mscodecompound><conditionalsrcstm><switch><mscodeif/>' +
      '<mscodecases/></switch></conditionalsrcstm><mscodeif node="mscodeif"><keyword text="IF"/><expression errornum="2"/>' +
      '<keyword text="THEN" errornum="13"/><statement node="thensrcstm"/><optional><keyword text="ELSE"/><statement node="elsesrcstm"/>' +
      '</optional></mscodeif><mscodecases node="mscodecases"><keyword text="CASE"/><expression errornum="2"/><keyword text="OF" errornum="12"/>' +
      '<loop text=";" skip="1"><caseselector/></loop><optional><keyword text="ELSE"/><statement/></optional><optional>' + '<char text=";"/></optional><keyword text="END" errornum="11"/></mscodecases><caseselector node="caseselector">' +
      '<setconstructor errornum="2"/><char text=":" errornum="4"/><statement/></caseselector><loopsrcstm><switch><mscoderepeat/>' +
      '<mscodewhile/><mscodefor/></switch></loopsrcstm><mscoderepeat node="mscoderepeat"><keyword text="REPEAT"/><srcstmlist/>' +
      '<keyword text="UNTIL" errornum="14"/><expression errornum="2"/></mscoderepeat><mscodewhile node="mscodewhile"><keyword text="WHILE"/>' +
      '<expression errornum="2"/><keyword text="DO" errornum="16"/><statement/></mscodewhile><mscodefor node="mscodefor"><keyword text="FOR"/>' +
      '<ident add="ident" errornum="1"/><char text=":" errornum="4"/><char text="=" skip="0" errornum="9"/><expression errornum="2"/>' +
      '<switch errornum="15"><keyword text="TO"/><keyword text="DOWNTO" add="downto"/></switch><expression errornum="2"/>' +
      '<keyword text="DO" errornum="16"/><statement/></mscodefor><mscodetry node="mscodetry"><keyword text="TRY"/><srcstmlist/><switch errornum="17">' +
      '<dojob><keyword text="FINALLY"/><srcstmlist node="finallysrcstm"/></dojob><dojob><keyword text="EXCEPT"/>' + '<srcstmlist node="ExceptSrcStm"/></dojob></switch><keyword text="END" errornum="11"/></mscodetry><withsrcstm node="with">' +
      '<keyword text="WITH"/><loop text=","><DScanner errornum="2"/></loop><keyword text="DO" errornum="16"/><statement/>' +
      '</withsrcstm><proceduredeclsection><switch><proceduredecl/><functiondecl/></switch></proceduredeclsection><proceduredecl node="procedure">' +
      '<procedureheading/><char text=";" errornum="5"/><block/><char text=";" errornum="5"/></proceduredecl><procedureheading><keyword text="PROCEDURE"/>' +
      '<ident add="name" errornum="1"/><optional><formalparameters/></optional></procedureheading><functiondecl node="function"><functionheading/>' +
      '<char text=";" errornum="5"/><block/><char text=";" errornum="5"/></functiondecl><functionheading><keyword text="FUNCTION"/>' +
      '<ident add="name" errornum="1"/><optional><formalparameters/></optional><char text=":" errornum="4"/><ident add="type" errornum="1"/>' +
      '</functionheading><formalparameters node="parameters"><char text="("/><optloop text=";"><formalparam/></optloop>' +
      '<char text=")" errornum="7"/></formalparameters><formalparam><optswitch><keyword text="VAR" node="varparams"/><keyword text="CONST"/>' + '</optswitch><varlist/></formalparam></language>';

   //==========================================================

Function StrVal(s: String): double;
Var
   p: integer;
Begin
   p := pos(',', s);
   If p > 0 Then
      s[p] := '.';
   val(s, Result, p);
End;

Function VarToInteger(v: variant): integer;
Begin
   Case VarType(v) Of
      varSmallint, varInteger, varByte, varError, varWord, varLongWord:
         Result := v;
      varSingle, varDouble, varCurrency, varDate:
         Result := round(double(v));
      varBoolean:
         If v = True Then
            Result := 1
         Else
            Result := 0;
      varString,
         varUString,
         varOleStr:
         Result := round(StrVal(v));
      varUnknown, varDispatch:
         Result := 0;
      Else
         If VarIsNull(v) Then
            Result := 0
         Else
            Result := VarAsType(v, varInteger);
   End;
End;

Function VarToInt64(v: variant): int64;
Begin
   Case VarType(v) Of
      varSmallint, varInteger, varByte, varError, varInt64, varWord, varLongWord:
         Result := v;
      varSingle, varDouble, varCurrency, varDate:
         Result := round(double(v));
      varBoolean:
         If v = True Then
            Result := 1
         Else
            Result := 0;
      varString,
         varUString,
         varOleStr:
         Result := round(StrVal(v));
      varUnknown, varDispatch:
         Result := 0;
      Else
         If VarIsNull(v) Then
            Result := 0
         Else
            Result := VarAsType(v, varInteger);
   End;
End;

Function VarToUInt64(v: variant): uint64;
Begin
   Case VarType(v) Of
      varSmallint, varInteger, varByte, varError,
         varUInt64, varWord, varLongWord:
         Result := v;
      varSingle, varDouble, varCurrency, varDate:
         Result := round(double(v));
      varBoolean:
         If v = True Then
            Result := 1
         Else
            Result := 0;
      varString,
         varUString,
         varOleStr:
         Result := round(StrVal(v));
      varUnknown, varDispatch:
         Result := 0;
      Else
         If VarIsNull(v) Then
            Result := 0
         Else
{$IFDEF DELPHIXE2_LVL}
            Result := VarAsType(v, varUInt64);
{$ELSE}
            Result := VarAsType(v, varInt64);
{$ENDIF}
   End;
End;

Function VarToIntObject(Const v: variant): IntObject;
Begin
{$IFDEF CPU64BITS}
   Result := VarToInt64(v);
{$ELSE}
   Result := VarToInteger(v);
{$ENDIF}
End;

Function VarToClass(Const v: variant): TClass;
Begin
   Result := TClass(VarToIntObject(v));
End;

Function ClassToVar(Const obj: TClass): variant;
Begin
   Result := IntObject(obj);
End;

Function VarToObject(Const v: variant): TObject;
Begin
   Result := TObject(VarToIntObject(v));
End;

Function ObjectToVar(Const obj: TObject): variant;
Begin
   Result := IntObject(obj);
End;

Function PointerToVar(p: pointer): variant;
Begin
   Result := IntObject(p);
End;

Function VarToPointer(v: variant): pointer;
Begin
   Result := pointer(VarToIntObject(v));
End;

Function VarToChar(v: variant): char;
Begin
   Case VarType(v) Of
      varSmallint, varInteger, varByte, varError, varWord, varLongWord:
         Result := Chr(integer(v) And $FF);

      varSingle, varDouble, varCurrency, varDate:
         Result := Chr(PtrInt(v) And $FF);

      varBoolean:
         If v = True Then
            Result := #1
         Else
            Result := #0;

      varString,
         varUString,
         varOleStr:
         If Length(v) > 0 Then
            Result := String(v)[1]
         Else
            Result := #0;

      Else
         Result := #0;
   End;
End;

Function VarToWideChar(v: variant): widechar;
Begin
   Case VarType(v) Of
      varSmallint, varInteger, varByte, varError, varWord, varLongWord:
         Result := widechar(integer(v) And $FFFF);

      varSingle, varDouble, varCurrency, varDate:
         Result := widechar(PtrInt(v) And $FFFF);

      varBoolean:
         If v = True Then
            Result := #1
         Else
            Result := #0;

      varString,
         varUString,
         varOleStr:
         If Length(v) > 0 Then
            Result := WideString(v)[1]
         Else
            Result := #0;

      Else
         Result := #0;
   End;
End;

Function VarToFloat(v: variant): double;
Begin
   Case VarType(v) Of
      varSmallint, varInteger, varByte, varError, varSingle, varDouble, varCurrency, varDate,
         varWord, varLongWord:
         Result := v;
      varBoolean:
         If v = True Then
            Result := 1
         Else
            Result := 0;
      varString,
         varUString,
         varOleStr:
         Result := StrVal(v);
      varUnknown, varDispatch:
         Result := 0;
      Else
         If VarIsNull(v) Then
            Result := 0
         Else
            Result := VarAsType(v, varDouble);
   End;
End;

Function VarToBoolean(v: variant): boolean;
Begin
   Result := (VarToInteger(v) <> 0);
End;

Function VarContent(v: variant): String;
Begin
   Try
      Case VarType(v) Of
         varEmpty:
            Result := 'Empty';
         varNull:
            Result := 'Null';
         varSmallint:
            Result := 'Smallint:' + VarToStr(v);
         varInteger:
            Result := 'Integer:' + VarToStr(v);
         varSingle:
            Result := 'Single:' + VarToStr(v);
         varDouble:
            Result := 'Double:' + VarToStr(v);
         varCurrency:
            Result := 'Currency:' + VarToStr(v);
         varDate:
            Result := 'Date:' + VarToStr(v);
         varOleStr:
            Result := 'OleStr:' + VarToStr(v);
         varDispatch:
            Result := 'Dispatch:' + VarToStr(v);
         varError:
            Result := 'Error:' + VarToStr(v);
         varBoolean:
            Result := 'Boolean:' + VarToStr(v);
         varVariant:
            Result := 'Variant:' + VarToStr(v);
         varUnknown:
            Result := 'Unknown:' + VarToStr(v);
         varByte:
            Result := 'Byte:' + VarToStr(v);
         varString:
            Result := 'String:' + VarToStr(v);
         varUString:
            Result := 'UnicodeString:' + VarToStr(v);
         varArray:
            Result := 'Array:' + VarToStr(v);
         varByRef:
            Result := 'ByRef:' + VarToStr(v);
         varWord:
            Result := 'Word:' + VarToStr(v);
         varLongWord:
            Result := 'Longword:' + VarToStr(v);
         Else
            Result := VarToStr(v);
      End;
   Except
      Result := '(unknown)';
   End;
End;

Function VarToEnumInteger(v: variant; PropType: PTypeInfo): integer;
Begin
   Case VarType(v) Of
      varSmallint, varInteger, varByte, varError, varWord, varLongWord:
         Result := v;
      varSingle, varDouble, varCurrency, varDate:
         Result := round(double(v));
      varBoolean:
         If v = True Then
            Result := 1
         Else
            Result := 0;
      varString,
         varUString,
         varOleStr:
         Result := GetEnumValue(PropType, v);
      varUnknown, varDispatch:
         Result := 0;
      Else
         If VarIsNull(v) Then
            Result := 0
         Else
            Result := VarAsType(v, varInteger);
   End;
End;

Function VariantToSet(v: variant; PropType: PTypeInfo): integer;
Var
   c, d: integer;
Begin
   d := VarArrayDimCount(v);
   If d = 1 Then
   Begin
      { monta um conjunto atrav?s dos itens passados no vetor de variants
        os elementos s?o tratados como bits e o conjunto ? formado pela uni?o
        desses bits }
      Result := 0;
      If Assigned(PropType) Then
      Begin
         PropType := GetTypeData(PropType).CompType;
         For c := VarArrayLowBound(v, 1) To VarArrayHighBound(v, 1) Do
            Result := Result Or (1 Shl VarToEnumInteger(v[c], PropType));
      End
      Else
         For c := VarArrayLowBound(v, 1) To VarArrayHighBound(v, 1) Do
            Result := Result Or (1 Shl VarToInteger(v[c]));
   End
   Else
      Case VarType(v) Of
         varSmallint, varInteger, varByte, varError, varWord, varLongWord:
            Result := v;
         varSingle, varDouble, varCurrency, varDate:
            Result := round(double(v));
         varBoolean:
            If v = True Then
               Result := 1
            Else
               Result := 0;
         varString,
            varUString,
            varOleStr:
            Result := GetEnumValue(PropType, v);
         varUnknown, varDispatch:
            Result := 0;
         Else
            If VarIsNull(v) Then
               Result := 0
            Else
               Result := VarAsType(v, varInteger);
      End;
End;

Function VarToSet(v: variant; PropType: PTypeInfo): integer;
Begin
   Result := VariantToSet(v, PropType);
End;

//============= TmscrItemsList =============================

Procedure TmscrItemList.Clear;
Begin
   While FItems.Count > 0 Do
   Begin
      TObject(FItems[0]).Free;
      FItems.Delete(0);
   End;
End;

Function TmscrItemList.Count: integer;
Begin
   Result := FItems.Count;
End;

Procedure TmscrItemList.Remove(Item: TObject);
Begin
   FItems.Remove(Item);
End;

Procedure TmscrItemList.Add(Item: TObject);
Begin
   FItems.Add(Item);
End;

Constructor TmscrItemList.Create;
Begin
   FItems := TList.Create;
End;

Destructor TmscrItemList.Destroy;
Begin
   Clear;
   FItems.Free;
   Inherited;
End;

//============= TmscrVarForBase =============

Function TmscrVarForBase.GetParam(Index: integer): TmscrVarForParamItem;
Begin
   Result := FItems[Index];
End;

Function TmscrVarForBase.GetPValue: PVariant;
Begin
   Result := @FValue;
End;

Function TmscrVarForBase.GetValue: variant;
Begin
   Result := FValue;
End;

Procedure TmscrVarForBase.SetValue(Const Value: variant);
Begin
   If Not FIsReadOnly Then
      FValue := Value;
End;

Function TmscrVarForBase.GetFullTypeName: String;
Begin
   Result := 'Variant';
   Case FTyp Of
      msvtInt: Result := 'Integer';
      msvtInt64: Result := 'Int64';
      msvtBool: Result := 'Boolean';
      msvtFloat: Result := 'Extended';
      msvtChar: Result := 'Char';
      msvtString: Result := 'String';
      msvtClass: Result := FTypeName;
      msvtRecord: Result := FTypeName;
      msvtArray: Result := 'Array';
      msvtEnum: Result := FTypeName;
   End;
End;

Function TmscrVarForBase.GetNumberOfRequiredParams: integer;
Var
   i: integer;
Begin
   Result := 0;
   For i := 0 To Count - 1 Do
      If Not Params[i].IsOptional Then
         Inc(Result);
End;

Constructor TmscrVarForBase.Create(Const AName: String; ATyp: TmscrVarType; Const ATypeName: String);
Begin
   Inherited Create;
   FName := AName;
   FTyp := ATyp;
   FTypeName := ATypeName;
   FValue := Null;
   FNeedResult := True;
   FUppercaseName := AnsiUpperCase(FName);
End;

//============= TmscrVarForString =============

Function TmscrVarForString.GetValue: variant;
Begin
   Result := FStr;
   If Assigned(FOnGetVarValue) Then
   Begin
      Result := FOnGetVarValue(FName, FTyp, FStr);
      If Result = null Then
         Result := FStr;
   End;
End;

Procedure TmscrVarForString.SetValue(Const Value: variant);
Begin
   FStr := Value;
End;

//============= TmscrVarForParamItem =============

Constructor TmscrVarForParamItem.Create(Const AName: String; ATyp: TmscrVarType; Const ATypeName: String; AIsOptional, AIsVarParam: boolean);
Begin
   Inherited Create(AName, ATyp, ATypeName);
   FIsOptional := AIsOptional;
   FIsVarParam := AIsVarParam;
   FDefValue := Null;
End;

//============= TmscrVarForProc =============

Function TmscrVarForProc.GetValue: variant;
Var
   Temp: boolean;
   ParentProg, SaveProg: TmscrScript;
   i: integer;
Begin
   Temp := FExecuting;
   FExecuting := True;
   If FIsFunc Then
      FRefItem.Value := Unassigned;

   ParentProg := FScrProject;
   SaveProg := Nil;

   While ParentProg <> Nil Do
      If ParentProg.FMainProj Then
      Begin
         SaveProg := ParentProg.FProjRunning;
         ParentProg.FProjRunning := FScrProject;
         break;
      End
      Else
         ParentProg := ParentProg.FParent;

   Try
      With FScrProject Do
      Begin
         FCalledExit := False;
         FTerminated := False;
         FIsRunning := True;
         FProjName := Self.FName;
         Try
            FStatement.Execute;
         Finally
            FCalledExit := False;
            FTerminated := False;
            FIsRunning := False;
            FProjName := '';
         End;
      End;

      If FIsFunc Then
         Result := FRefItem.Value
      Else
         Result := Null;

   Finally
      If ParentProg <> Nil Then
         ParentProg.FProjRunning := SaveProg;
      FExecuting := Temp;
      If (ParentProg <> Nil) And ParentProg.FClearLocalVars Then
         For i := 0 To Project.Count - 1 Do
            If (Project.Items[i] Is TmscrVarForVariable) And (CompareText('Result', Project.Items[i].Name) <> 0) And Not VarIsClear(Project.Items[i].Value) Then
               Case TmscrVarForVariable(Project.Items[i]).Typ Of
                  msvtString:
                     TmscrVarForVariable(Project.Items[i]).Value := '';
                  msvtInt, msvtFloat, msvtChar:
                     TmscrVarForVariable(Project.Items[i]).Value := 0;
                  msvtVariant:
                     TmscrVarForVariable(Project.Items[i]).Value := Null;
               End;
   End;
End;

Constructor TmscrVarForProc.Create(Const AName: String; ATyp: TmscrVarType; Const ATypeName: String; AParent: TmscrScript; AIsFunc: boolean = True);
Begin
   Inherited Create(AName, ATyp, ATypeName);
   FIsReadOnly := True;
   FVarsStack := TList.Create;
   FIsFunc := AIsFunc;
   FScrProject := TmscrScript.Create(Nil);
   FScrProject.Parent := AParent;
   If FScrProject.Parent <> Nil Then
      FScrProject.UseClassLateBinding := FScrProject.Parent.UseClassLateBinding;

   If FIsFunc Then
   Begin
      FRefItem := TmscrVarForVariable.Create('Result', ATyp, ATypeName);
      FScrProject.Add('Result', FRefItem);
   End;
End;

Destructor TmscrVarForProc.Destroy;
Var
   i: integer;
Begin
   For i := 0 To Count - 1 Do
      FScrProject.FItems.Delete(FScrProject.FItems.IndexOfObject(Params[i]));

   FScrProject.Free;
   FVarsStack.Free;
   Inherited;
End;

Procedure TmscrVarForProc.RestoreLocalVariables(StackIndex: integer; bSkipVarParams: boolean; dItem: TmscrDScannerItem);
Var
   i: integer;
   LocalVars: TList;
   StackItem: TmscrEngLocalVariables;
   bIsVar: boolean;
   Temp1: Array Of variant;
Begin
   If (FVarsStack.Count < StackIndex) Or (StackIndex < 0) Then Exit;

   LocalVars := TList(FVarsStack[StackIndex]);
   SetLength(Temp1, Count);
   Try
      If Assigned(dItem) Then
         For i := 0 To Count - 1 Do
            If Params[i].IsVarParam Then
               Temp1[i] := Params[i].Value;
      For i := 0 To LocalVars.Count - 1 Do
      Begin
         StackItem := TmscrEngLocalVariables(LocalVars[i]);
         bIsVar := TmscrVarForParamItem(StackItem.FVariableLink).FIsVarParam;
         If Not (bSkipVarParams And (StackItem.FVariableLink Is TmscrVarForParamItem) And bIsVar) Then
            StackItem.FVariableLink.Value := StackItem.FValue;
         StackItem.Free;
      End;
      If Assigned(dItem) Then
         For i := 0 To Count - 1 Do
            If Params[i].IsVarParam Then
               dItem[i].Value := Temp1[i];
   Finally
      Temp1 := Nil;
      LocalVars.Free;
      FVarsStack.Delete(StackIndex);
   End;
End;

Function TmscrVarForProc.SaveLocalVariables: integer;
Var
   i: integer;
   LocalVars: TList;
   StackItem: TmscrEngLocalVariables;
Begin
   LocalVars := TList.Create;
   FVarsStack.Add(LocalVars);
   Result := FVarsStack.Count - 1;
   For i := 0 To Project.Count - 1 Do
      If (Project.Items[i] Is TmscrVarForVariable) Or (Project.Items[i] Is TmscrVarForParamItem) Then
      Begin
         StackItem := TmscrEngLocalVariables.Create;
         StackItem.FValue := Project.Items[i].Value;
         StackItem.FVariableLink := Project.Items[i];
         LocalVars.Add(StackItem);
      End;
End;

//============= TmscrEngProperty =============

Procedure TmscrEngProperty.SetValue(Const Value: variant);
Var
   p: PPropInfo;
   Instance: TObject;
   IntVal: fmscrInteger;
Begin

   If IsReadOnly Then Exit;
   Instance := TObject(fmscrInteger(ParentValue));

   If Instance = Nil Then
      Raise Exception.CreateFmt(msEClassCreateError, [FClassRef.ClassName]);

   If FIsPublished Then
   Begin
      p := GetPropInfo(Instance.ClassInfo, Name);
      If p <> Nil Then
         Case p.PropType^.Kind Of
            tkInteger, tkSet, tkEnumeration, tkClass, tkBool:
               Begin
                  Begin
                     If Typ = msvtBool Then
                        If Value = True Then
                           IntVal := 1
                        Else
                           IntVal := 0
                     Else
                        IntVal := fmscrInteger(Value);

                     SetOrdProp(Instance, p, IntVal);
                  End;
               End;
            tkInt64: SetInt64Prop(Instance, p, Value);
            tkFloat: SetFloatProp(Instance, p, extended(Value));
            tkChar, tkWChar: SetOrdProp(Instance, p, Ord(String(Value)[1]));
            tkVariant: SetVariantProp(Instance, p, Value);

            tkString, tkLString,
               tkWString, tkAString: SetStrProp(Instance, p, String(Value));
         End;
   End
   Else If Assigned(FOnSetValue) Then
      FOnSetValue(Instance, FClassRef, FUppercaseName, Value, Self);

End;

Function TmscrEngProperty.GetValue: variant;
Var
   p: PPropInfo;
   Instance: TObject;
Begin
   Result := Null;
   Instance := TObject(fmscrInteger(ParentValue));

   If Instance = Nil Then
      Raise Exception.CreateFmt(msEClassCreateError, [FClassRef.ClassName]);

   If FIsPublished And Assigned(Instance) Then
   Begin
      p := GetPropInfo(Instance.ClassInfo, Name);
      If p <> Nil Then
         Case p.PropType^.Kind Of
            tkInteger, tkSet, tkEnumeration,
               tkClass, tkBool: Result := GetOrdProp(Instance, p);

            tkInt64: Result := GetInt64Prop(Instance, p);
            tkFloat: Result := GetFloatProp(Instance, p);
            tkChar, tkWChar: Result := Chr(GetOrdProp(Instance, p));
            tkVariant: Result := GetVariantProp(Instance, p);

            tkString, tkLString,
               tkWString, tkAString: Result := GetStrProp(Instance, p);
         End;
   End
   Else If Assigned(FOnGetValue) Then
      Result := FOnGetValue(Instance, FClassRef, FUppercaseName, Self);

   If Typ = msvtBool Then
      If Result = 0 Then
         Result := False
      Else
         Result := True;

End;

//============= TmscrEngRecElement =============

Function TmscrEngRecElement.GetValue: variant;
Var
   Instance: Pointer;
Begin
   Result := Null;

   Instance := TmscrVarForRecord(FParentRef).FRecData;

   If Assigned(FOnGetValue) Then
      Result := FOnGetValue(Instance, TmscrVarForRecord(FParentRef).TypeInfo, FUppercaseName);

   If Typ = msvtBool Then
      If Result = 0 Then
         Result := False
      Else
         Result := True;
End;

Procedure TmscrEngRecElement.SetValue(Const Value: variant);
Var
   Instance: Pointer;
Begin

   Instance := TmscrVarForRecord(FParentRef).FRecData;

   If Assigned(FOnSetValue) Then
      FOnSetValue(Instance, TmscrVarForRecord(FParentRef).TypeInfo, FUppercaseName, Value);
End;

//============= TmscrVarForMethod =============

Procedure TmscrVarForMethod.SetVParam(Index: integer; Const Value: variant);
Begin
   TmscrVarForParamItem(FItems[Index]).Value := Value;
End;

Function TmscrVarForMethod.GetValue: variant;
Var
   i: integer;
   Instance: TObject;
Begin
   If Assigned(FOnCall) Then
   Begin
      Instance := Nil;
      If Not VarIsNull(ParentValue) Then
         Instance := TObject(fmscrInteger(ParentValue));

      If FIndexMethod Then
         Result := FOnCall(Instance, FClassRef, FUppercaseName + '.GET', Self)
      Else
         Result := FOnCall(Instance, FClassRef, FUppercaseName, Self);

   End
   Else
      Result := 0;
End;

Procedure TmscrVarForMethod.SetValue(Const Value: variant);
Var
   v: variant;
   i: integer;
Begin
   If FIndexMethod Then
      If Assigned(FOnCall) Then
      Begin
         FSetValue := Value;
         FOnCall(TObject(fmscrInteger(ParentValue)), FClassRef, FUppercaseName + '.SET', Self);
         FSetValue := Null;
      End;
End;

Function TmscrVarForMethod.GetVParam(Index: integer): variant;
Begin
   If Index = Count Then
      Result := FSetValue
   Else
      Result := TmscrVarForParamItem(FItems[Index]).Value;
End;

Constructor TmscrVarForMethod.Create(Const Syntax: String; Script: TmscrScript);
Var
   i: integer;
   v: TmscrVarForBase;
Begin
   v := ParseMethodSyntax(Syntax, Script);
   Inherited Create(v.Name, v.Typ, v.TypeName);
   FIsReadOnly := True;
   FSyntax := Syntax;
   IsMacro := v.IsMacro;

   For i := 0 To v.Count - 1 Do
      Add(v.Params[i]);

   While v.Count > 0 Do
      v.FItems.Delete(0);

   v.Free;

   If Count > 0 Then
      FVarArray := VarArrayCreate([0, Count - 1], varVariant);
End;

Destructor TmscrVarForMethod.Destroy;
Begin
   FVarArray := Null;
   Inherited;
End;

//============= TmscrEngComponent =============

Function TmscrEngComponent.GetValue: variant;
Begin
   Result := fmscrInteger(FComponent);
End;

Constructor TmscrEngComponent.Create(Component: TComponent);
Begin
   Inherited Create(Component.Name, msvtClass, Component.ClassName);
   FComponent := Component;
End;

//============= TmscrEngEvent =============

Procedure TmscrEngEvent.SetValue(Const Value: variant);
Var
   Instance: TPersistent;
   v: TmscrVarForBase;
   e: TmscrCustomEvent;
   p: PPropInfo;
   m: TMethod;
Begin
   Instance := TPersistent(fmscrInteger(ParentValue));

   If VarToStr(Value) = '0' Then
   Begin
      m.Code := Nil;
      m.Data := Nil;
   End Else
   Begin
      v := FScrProject.Find(Value);
      If (v = Nil) Or Not (v Is TmscrVarForProc) Then
         Raise Exception.Create(msEEventError);

      e := TmscrCustomEvent(FEvent.NewInstance);
      e.Create(Instance, TmscrVarForProc(v));
      FScrProject.Add('', e);
      m.Code := e.GetMethod;
      m.Data := e;
   End;

   //=== ct9999-zzz ===================
   If Instance = Nil Then
      Raise Exception.CreateFmt(msEClassCreateError, [FClassRef.ClassName]);

   p := GetPropInfo(Instance.ClassInfo, Name);
   SetMethodProp(Instance, p, m);
End;

Function TmscrEngEvent.GetValue: variant;
Begin
   Result := '';
End;

Constructor TmscrEngEvent.Create(Const Name: String; AEvent: TmscrCustomEventClass);
Begin
   Inherited Create(Name, msvtString, '');
   FEvent := AEvent;
End;

//============= TmscrVarForClass =============

Function TmscrVarForClass.GetMembers(Index: integer): TmscrEngCustom;
Begin
   Result := FMembers.FItems[Index];
End;

Function TmscrVarForClass.GetMembersCount: integer;
Begin
   Result := FMembers.Count;
End;

Procedure TmscrVarForClass.Register_Constructor(Syntax: String; CallEvent: TmscrCallMethod);
Var
   i: integer;
Begin
   i := Pos(' ', Syntax);
   Delete(Syntax, 1, i - 1);
   Syntax := 'function' + Syntax + ': ' + 'Constructor';
   Register_Method(Syntax, CallEvent);
End;

Procedure TmscrVarForClass.Register_Method(Const Syntax: String; CallEvent: TmscrCallMethod);
Var
   m: TmscrVarForMethod;
Begin
   m := TmscrVarForMethod.Create(Syntax, FScrProject);
   m.FOnCall := CallEvent;
   m.FClassRef := FClassRef;
   FMembers.Add(m);
End;

Procedure TmscrVarForClass.Register_Property(Const Name, Typ: String; GetEvent: TmscrClassGetProperty; SetEvent: TmscrClassSetProperty);
Var
   p: TmscrEngProperty;
Begin
   p := TmscrEngProperty.Create(Name, StrToVarType(Typ, FScrProject), Typ);
   p.FClassRef := FClassRef;
   p.FOnGetValue := GetEvent;
   p.FOnSetValue := SetEvent;
   p.IsReadOnly := Not Assigned(SetEvent);
   FMembers.Add(p);
End;

Procedure TmscrVarForClass.Register_Event(Const Name: String; AEvent: TmscrCustomEventClass; Const ADeclaration: String='');
Var
   e: TmscrEngEvent;
Begin
   e := TmscrEngEvent.Create(Name, AEvent);
   e.FClassRef := FClassRef;
   if ADeclaration='' then
     e.Declaration:= AEvent.GetDeclaration else
     e.Declaration:= ADeclaration;

   FMembers.Add(e);
End;

Procedure TmscrVarForClass.Register_DefaultProperty(Const Name, Params, Typ: String; CallEvent: TmscrCallMethod; AReadOnly: boolean);
Begin
   Register_IndexProperty(Name, Params, Typ, CallEvent, AReadOnly);
   FDefProperty := Members[FMembers.Count - 1];
End;

Procedure TmscrVarForClass.Register_IndexProperty(Const Name, Params, Typ: String; CallEvent: TmscrCallMethod; AReadOnly: boolean);
Var
   i: integer;
   sl: TStringList;
   s: String;
Begin
   sl := TStringList.Create;
   sl.CommaText := Params;

   s := '';
   For i := 0 To sl.Count - 1 Do
      s := s + 'p' + IntToStr(i) + ': ' + sl[i] + '; ';

   SetLength(s, Length(s) - 2);
   Try
      Register_Method('function ' + Name + '(' + s + '): ' + Typ, CallEvent);

      With TmscrVarForMethod(Members[FMembers.Count - 1]) Do
      Begin
         IsReadOnly := AReadOnly;
         FIndexMethod := True;
      End;

   Finally
      sl.Free;
   End;
End;

Procedure TmscrVarForClass.AddComponent(c: TComponent);
Begin
   FMembers.Add(TmscrEngComponent.Create(c));
End;

Procedure TmscrVarForClass.AddPublishedProperties(AClass: TClass);
Var
   TypeInfo: PTypeInfo;
   propCount: integer;
   propList: PPropList;
   i: integer;
   cl: String;
   t: TmscrVarType;
   FClass: TClass;
   p: TmscrEngProperty;
Begin
   TypeInfo := AClass.ClassInfo;
   If TypeInfo = Nil Then Exit;

   propCount := GetPropList(TypeInfo, tkProperties, Nil);
   GetMem(PropList, PropCount * SizeOf(PPropInfo));
   GetPropList(TypeInfo, tkProperties, PropList);

   Try
      For i := 0 To PropCount - 1 Do
      Begin
         t := msvtInt;
         cl := '';

         Case PropList[i].PropType^.Kind Of
            tkInteger:
               t := msvtInt;
            tkInt64:
               t := msvtInt64;
            tkSet:
               Begin
                  t := msvtEnum;
                  cl := String(PropList[i].PropType^.Name);
               End;
            tkEnumeration:
               Begin
                  t := msvtEnum;
                  cl := String(PropList[i].PropType^.Name);
                  If (CompareText(cl, 'Boolean') = 0) Or (CompareText(cl, 'bool') = 0) Then
                     t := msvtBool;
               End;
            tkBool:
               t := msvtBool;
            tkFloat:
               t := msvtFloat;
            tkChar, tkWChar:
               t := msvtChar;
            tkString, tkLString, tkWString, tkAString:
               t := msvtString;
            tkVariant:
               t := msvtVariant;
            tkRecord:
               t := msvtRecord;
            tkClass:
               Begin
                  t := msvtClass;
                  FClass := GetTypeData(PropList[i].PropType).ClassType;
                  cl := FClass.ClassName;
               End;
         End;

         p := TmscrEngProperty.Create(String(PropList[i].Name), t, cl);
         p.FClassRef := FClassRef;
         p.FIsPublished := True;
         FMembers.Add(p);
      End;

   Finally
      FreeMem(PropList, PropCount * SizeOf(PPropInfo));
   End;
End;

Function TmscrVarForClass.Find(Const Name: String): TmscrEngCustom;
Var
   cl: TmscrVarForClass;

   //.........................
   Function DoFind(Const Name: String): TmscrEngCustom;
   Var
      i: integer;
   Begin
      Result := Nil;
      For i := 0 To FMembers.Count - 1 Do
         If CompareText(Name, Members[i].Name) = 0 Then
         Begin
            Result := Members[i];
            Exit;
         End;
   End;

   //.........................
Begin
   Result := DoFind(Name);
   If Result = Nil Then
   Begin
      cl := FScrProject.FindClass(FParentClassStr);
      If cl <> Nil Then
         Result := cl.Find(Name);
   End;
End;

Function TmscrVarForClass.GetValue: variant;
Begin

   If Params[0].Value = Null Then
      Result := fmscrInteger(FClassRef.NewInstance)
   Else
      Result := Params[0].Value;

   Params[0].Value := Null;
End;

Constructor TmscrVarForClass.Create(AClass: TClass; Const ParentClassStr: String);
Begin
   Inherited Create(AClass.ClassName, msvtClass, AClass.ClassName);
   FMembers := TmscrItemList.Create;
   FParentClassStr := ParentClassStr;
   FClassRef := AClass;

   AddPublishedProperties(AClass);
   Add(TmscrVarForParamItem.Create('', msvtVariant, '', True, False));
End;

Destructor TmscrVarForClass.Destroy;
Begin
   FMembers.Free;
   Inherited;
End;

//============= TmscrVarForRecord =============

Constructor TmscrVarForRecord.Create(ATypeInfo: PTypeInfo; Const aTypeInfoStr: String);
Begin
   Inherited Create(aTypeInfoStr, msvtRecord, ATypeInfo.Name);

   FMembers := TmscrItemList.Create;
   FTypeInfoStr := TypeInfoStr;
   FTypeInfo := ATypeInfo;
   FRecData := Nil;
   FRecDataSize := 0;
   FValue := Nil;

   Add(TmscrVarForParamItem.Create('', msvtVariant, '', True, False));

   //..................

   FRecDataSize := SizeOf(FTypeInfo^);
   GetMem(FRecData, FRecDataSize);
   FValue := fmscrInteger(FRecData);
End;

Destructor TmscrVarForRecord.Destroy;
Begin
   If FRecDataSize > 0 Then FreeMem(FRecData, FRecDataSize);
   FMembers.Free;
   Inherited;
End;

Function TmscrVarForRecord.GetMembers(Index: integer): TmscrEngCustom;
Begin
   Result := FMembers.FItems[Index];
End;

Function TmscrVarForRecord.GetMembersCount: integer;
Begin
   Result := FMembers.Count;
End;

Procedure TmscrVarForRecord.Register_Element(Const Name, Typ: String; GetEvent: TmscrRecGetValueElement; SetEvent: TmscrRecSetValueElement = Nil);
Var
   p: TmscrEngRecElement;
Begin
   p := TmscrEngRecElement.Create(Name, StrToVarType(Typ, FScrProject), Typ);
   p.ParentRef := Self;
   p.ParentValue := FValue;
   p.FOnGetValue := GetEvent;
   p.FOnSetValue := SetEvent;
   p.IsReadOnly := Not Assigned(SetEvent);
   FMembers.Add(p);
End;

Procedure TmscrVarForRecord.AddComponent(c: TComponent);
Begin
   FMembers.Add(TmscrEngComponent.Create(c));
End;

Function TmscrVarForRecord.Find(Const Name: String): TmscrEngCustom;
Var
   cl: TmscrVarForRecord;

   //.........................
   Function DoFind(Const Name: String): TmscrEngCustom;
   Var
      i: integer;
   Begin
      Result := Nil;
      For i := 0 To FMembers.Count - 1 Do
         If CompareText(Name, Members[i].Name) = 0 Then
         Begin
            Result := Members[i];
            Exit;
         End;
   End;

   //.........................
Begin
   Result := DoFind(Name);
   If Result = Nil Then
   Begin
      cl := FScrProject.FindRecord(FTypeInfoStr);
      If cl <> Nil Then
         Result := cl.Find(Name);
   End;
End;

//============= TmscrDScannerItem =============

Function TmscrDScannerItem.GetItem(Index: integer): TmscrExpressionBase;
Begin
   Result := FItems[Index];
End;

//=============TmscrDScanner =============

Constructor TmscrDScanner.Create(aProject: TmscrScript);
Var
   ParentProg: TmscrScript;
Begin

{$IFDEF CPU64}
   Inherited Create('', msvtInt64, '');
{$ELSE}
   Inherited Create('', msvtInt, '');
{$ENDIF}

   FScrProject := aProject;
   FMainProj := FScrProject;
   ParentProg := FScrProject;

   While ParentProg <> Nil Do
      If ParentProg.FMainProj Then
      Begin
         FMainProj := ParentProg;
         break;
      End
      Else
      Begin
         ParentProg := ParentProg.FParent;
      End;

   FScrProject.UseClassLateBinding := FMainProj.UseClassLateBinding;
End;

Destructor TmscrDScanner.Destroy;
Begin
   If FLateBindingDATASource <> Nil Then
      FLateBindingDATASource.Free;
   Inherited;
End;

Procedure TmscrDScanner.Borrow(ADScanner: TmscrDScanner);
Var
   SaveItems: TList;
Begin
   SaveItems := FItems;
   FItems := ADScanner.FItems;
   ADScanner.FItems := SaveItems;
   FKind := ADScanner.FKind;
   FRef1 := ADScanner.FRef1;
   FRef2 := ADScanner.FRef2;
   FTyp := ADScanner.Typ;
   FTypeName := ADScanner.TypeName;
   FIsReadOnly := ADScanner.IsReadOnly;
   RefItem := ADScanner.RefItem;
End;

Procedure TmscrDScanner.Finalize;
Var
   Item: TmscrDScannerItem;
Begin
   Item := Items[Count - 1];
   FTyp := Item.Ref.Typ;
   FTypeName := Item.Ref.TypeName;

   If FTyp = msvtConstructor Then
   Begin
      FTyp := msvtClass;
      FTypeName := Items[Count - 2].Ref.TypeName;
   End;

   FIsReadOnly := Item.Ref.IsReadOnly;

   If (Count = 1) And (Items[0].Ref Is TmscrVarForVariable) Then
   Begin
      RefItem := Items[0].Ref;
      FKind := msskVariable;
   End
   Else If (Count = 2) And (Items[0].Ref Is TmscrVarForString) Then
   Begin
      RefItem := Items[0].Ref;
      FRef1 := Items[1][0];
      FKind := msskStringArray;
   End
   Else If (Count = 2) And (Items[0].Ref Is TmscrVarForVariable) And (Items[0].Ref.Typ = msvtArray) Then
   Begin
      RefItem := Items[0].Ref;
      FRef1 := RefItem.RefItem;
      FRef2 := Items[1];
      FKind := msskArray;
   End
   Else
      FKind := msskOther;
End;

Function TmscrDScanner.GetItem(Index: integer): TmscrDScannerItem;
Begin
   Result := FItems[Index];
End;

Function TmscrDScanner.DoCalc(Const AValue: variant; Flag: boolean): variant;
Var
   i, j: integer;
   Item: TmscrVarForBase;
   Val: variant;
   Ref: TmscrVarForBase;
   Temp1: Array Of variant;
   StackIndex: integer;
Begin
   Ref := Nil;
   Val := Null;
   StackIndex := -1;
   For i := 0 To Count - 1 Do
   Begin
      Item := Items[i].Ref;

      If Item Is TmscrDScanner Then
      Begin
         Ref := Item;
         Val := Item.Value;
         continue;
      End;

      Try
         If (Item Is TmscrVarForProc) And TmscrVarForProc(Item).Executing Then
            StackIndex := TmscrVarForProc(Item).SaveLocalVariables;

         If Item.Count > 0 Then
         Begin
            SetLength(Temp1, Item.Count);

            Try
               For j := 0 To Item.Count - 1 Do
                  If Item.IsMacro Then
                     Temp1[j] := TmscrExpression(Items[i][j]).Source
                  Else
                     Temp1[j] := Items[i][j].Value;

               For j := 0 To Item.Count - 1 Do
                  Item.Params[j].Value := Temp1[j];

            Finally
               Temp1 := Nil;
            End;
         End;

         If Item Is TmscrEngCustom Then
         Begin
            TmscrEngCustom(Item).ParentRef := Ref;
            TmscrEngCustom(Item).ParentValue := Val;
            TmscrEngCustom(Item).aProj := FScrProject;
         End;

         Ref := Item;

         If Flag And (i = Count - 1) Then
         Begin
            Item.Value := AValue;
         End
         Else
         Begin
            Item.NeedResult := (i <> Count - 1) Or NeedResult;
            Val := Item.Value;
         End;

         For j := 0 To Item.Count - 1 Do
            If Item.Params[j].IsVarParam Then
               Items[i][j].Value := Item.Params[j].Value;

      Finally
         If (Item Is TmscrVarForProc) And TmscrVarForProc(Item).Executing Then
            TmscrVarForProc(Item).RestoreLocalVariables(StackIndex, False, Items[i]);
      End;
   End;

   Result := Val;
End;

Procedure TmscrDScanner.CheckLateBinding;
Var
   NewDScanner: TmscrDScanner;
   Parser: TmscrEngParser;
Begin
   If FLateBindingDATASource <> Nil Then
   Begin
      Parser := TmscrEngParser.Create(FScrProject);
      Try
         NewDScanner := Parser.DoDScanner(FLateBindingDATASource, FScrProject);
         Borrow(NewDScanner);
         NewDScanner.Free;
      Finally
         Parser.Free;
         FLateBindingDATASource.Free;
         FLateBindingDATASource := Nil;
      End;
   End;
End;

Function TmscrDScanner.GetValue: variant;
Begin
   CheckLateBinding;
   Result := DoCalc(Null, False);
End;

Procedure TmscrDScanner.SetValue(Const Value: variant);
Begin
   CheckLateBinding;
   DoCalc(Value, True);
End;

//=============TmscrVarForVariableDScanner =============

Function TmscrVarForVariableDScanner.GetValue: variant;
Begin
   Result := RefItem.Value;
End;

Procedure TmscrVarForVariableDScanner.SetValue(Const Value: variant);
Begin
   RefItem.Value := Value;
End;

//=============TmscrStringDScanner =============

Function TmscrStringDScanner.GetValue: variant;
Begin
   Result := TmscrVarForString(RefItem).FStr[integer(FRef1.Value)];
End;

Procedure TmscrStringDScanner.SetValue(Const Value: variant);
Begin
   TmscrVarForString(RefItem).FStr[integer(FRef1.Value)] := VarToStr(Value)[1];
End;

//============= TmscrArrayDScanner =============

Function TmscrArrayDScanner.GetValue: variant;
Var
   i: integer;
Begin
   TmscrEngCustom(FRef1).ParentRef := RefItem;
   For i := 0 To FRef2.Count - 1 Do
      FRef1.Params[i].Value := FRef2[i].Value;
   Result := FRef1.Value;
End;

Procedure TmscrArrayDScanner.SetValue(Const Value: variant);
Var
   i: integer;
Begin
   TmscrEngCustom(FRef1).ParentRef := RefItem;
   For i := 0 To FRef2.Count - 1 Do
      FRef1.Params[i].Value := FRef2[i].Value;
   FRef1.Value := Value;
End;

//============= TmscrSetExpression =============

Function TmscrSetExpression.Check(Const Value: variant): boolean;
Var
   i: integer;
   Expr: TmscrExpressionBase;
Begin
   Result := False;

   i := 0;
   While i < Count Do
   Begin
      Expr := Items[i];

      If (i < Count - 1) And (Items[i + 1] = Nil) Then
      Begin
         Result := (Value >= Expr.Value) And (Value <= Items[i + 2].Value);
         Inc(i, 2);
      End
      Else
      Begin
         Result := Value = Expr.Value;
      End;

      If Result Then break;
      Inc(i);
   End;
End;

Function TmscrSetExpression.GetItem(Index: integer): TmscrExpressionBase;
Begin
   Result := FItems[Index];
End;

Function TmscrSetExpression.GetValue: variant;
Var
   i: integer;
Begin
   Result := VarArrayCreate([0, Count - 1], varVariant);

   For i := 0 To Count - 1 Do
      If Items[i] = Nil Then
         Result[i] := Null
      Else
         Result[i] := Items[i].Value;
End;

//=============TmscrScript =============

Constructor TmscrScript.Create(AOwner: TComponent);
Begin
   Inherited;
   FEvaluteRiseError := False;
   FItems := TStringList.Create;
   FItems.Sorted := True;
   FItems.Duplicates := dupAccept;
   FLines := TStringList.Create;
   FMacros := TStringList.Create;
   FIncludePath := TStringList.Create;
   FIncludePath.Add('');
   FStatement := TmscrStatement.Create(Self, '', '');
   FSyntaxType := _SyntaxType;
   FUnitLines := TStringList.Create;
   FUseClassLateBinding := False;
End;

Destructor TmscrScript.Destroy;
Begin
   Inherited;
   Clear;
   ClearRTTI;
   FItems.Free;
   FLines.Free;
   FMacros.Free;
   FIncludePath.Free;
   FStatement.Free;
   FUnitLines.Free;
End;

Procedure TmscrScript.Add(Const Name: String; Item: TObject);
Begin
   FItems.AddObject(Name, Item);
   If Item Is TmscrVarForBase Then
      TmscrVarForBase(Item).AddedBy := FAddedBy;
End;

Function TmscrScript.Count: integer;
Begin
   Result := FItems.Count;
End;

Procedure TmscrScript.Remove(Item: TObject);
Begin
   FItems.Delete(FItems.IndexOfObject(Item));
End;

Function TmscrScript.GetItem(Index: integer): TmscrVarForBase;
Begin
   Result := TmscrVarForBase(FItems.Objects[Index]);
End;

Function TmscrScript.GetProgName: String;
Begin
   If Assigned(ProjRunning) Then
      Result := ProjRunning.ProjName
   Else
      Result := FProjName;
End;

Function TmscrScript.Find(Const Name: String): TmscrVarForBase;
Begin
   Result := FindLocal(Name);

   If (Result = Nil) And (FParent <> Nil) Then
      Result := FParent.Find(Name);
End;

Function TmscrScript.FindLocal(Const Name: String): TmscrVarForBase;
Var
   i: integer;
Begin
   Result := Nil;
   i := FItems.IndexOf(Name);
   If (i <> -1) And (FItems.Objects[i] Is TmscrVarForBase) Then
      Result := TmscrVarForBase(FItems.Objects[i]);
End;

Procedure TmscrScript.ClearItems(Owner: TObject);
Var
   i: integer;
Begin
   RemoveItems(Owner);
   FStatement.Clear;
   For i := 0 To FUnitLines.Count - 1 Do
      FUnitLines.Objects[i].Free;
   FUnitLines.Clear;
End;

Procedure TmscrScript.RemoveItems(Owner: TObject);
Var
   i: integer;
Begin
   For i := Count - 1 Downto 0 Do
      If FItems.Objects[i] Is TmscrVarForBase Then
         If Items[i].AddedBy = Owner Then
         Begin
            Items[i].Free;
            Remove(Items[i]);
         End;
End;

Procedure TmscrScript.Clear;
Var
   i: integer;
   item: TObject;
Begin
   i := 0;
   While i < FItems.Count Do
   Begin
      item := FItems.Objects[i];
      If (item Is TmscrRTTILibrary) Or ((item Is TmscrVarForBase) And (TmscrVarForBase(item).AddedBy = TObject(1))) Then
      Begin
         Inc(i);
      End
      Else
      Begin
         item.Free;
         FItems.Delete(i);
      End;
   End;

   FStatement.Clear;
   For i := 0 To FUnitLines.Count - 1 Do
      FUnitLines.Objects[i].Free;
   FUnitLines.Clear;
   FErrorPos := '';
   FErrorMsg := '';
   FErrorUnit := '';
End;

Procedure TmscrScript.Execute;
Begin
   FCalledExit := False;
   FTerminated := False;
   FIsRunning := True;
   FMainProj := True;
   Try
      FStatement.Execute;
   Finally
      FCalledExit := False;
      FTerminated := False;
      FIsRunning := False;
   End;
End;

Function TmscrScript.Run: boolean;
Begin
   Result := Compile;
   If Result Then Execute;
End;

Function TmscrScript.Compile: boolean;
Var
   p: TmscrEngParser;
Begin
   Result := False;
   FErrorMsg := '';
   p := TmscrEngParser.Create(Self);
   Try
      p.InitGrammar;
      If p.FinalScript_Build(FLines.Text) Then p.FinalScript_Parse;
   Finally
      p.Free;
   End;
   If FErrorMsg = '' Then
   Begin
      Result := True;
      FErrorPos := '';
   End;
End;

Function TmscrScript.GetFinalCode(Stream: TStream): boolean;
Var
   p: TmscrEngParser;
Begin
   Result := False;
   FErrorMsg := '';

   p := TmscrEngParser.Create(Self);
   Try
      p.InitGrammar;
      If p.FinalScript_Build(FLines.Text) Then
         p.FinalScript.SaveToStream(Stream);
   Finally
      p.Free;
   End;

   If FErrorMsg = '' Then
   Begin
      Result := True;
      FErrorPos := '';
   End;
End;

Function TmscrScript.SetFinalCode(Stream: TStream): boolean;
Var
   p: TmscrEngParser;
Begin
   Result := False;
   FErrorMsg := '';

   p := TmscrEngParser.Create(Self);
   Try
      p.InitGrammar;
      p.FinalScript.LoadFromStream(Stream);
      p.FinalScript_Parse;
   Finally
      p.Free;
   End;

   If FErrorMsg = '' Then
   Begin
      Result := True;
      FErrorPos := '';
   End;
End;

Procedure TmscrScript.Register_Type(Const TypeName: String; ParentType: TmscrVarType);
Var
   v: TmscrVarForType;
Begin
   If Find(TypeName) <> Nil Then
      Exit;
   v := TmscrVarForType.Create(TypeName, ParentType, '');
   Add(TypeName, v);
End;

Function TmscrScript.Register_Record(Const aTypeName: String; pti: PTypeInfo): TmscrVarForRecord;
Var
   rc: TmscrVarForRecord;
Begin
   Result := Nil;
   If pti = Nil Then Exit;
   If Find(aTypeName) <> Nil Then Exit;
   If pti.Kind <> tkRecord Then Exit;

   Result := TmscrVarForRecord.Create(pti, aTypeName);
   Result.FScrProject := Self;
   Add(Result.Name, Result);

   rc := TmscrVarForRecord(Find(aTypeName));
   If rc <> Nil Then Result.FDefProperty := rc.DefProperty;
End;

Function TmscrScript.Register_Class(AClass: TClass; Const ParentClassStr: String): TmscrVarForClass;
Var
   cl: TmscrVarForClass;
Begin
   Result := Nil;
   If Find(AClass.ClassName) <> Nil Then Exit;

   Result := TmscrVarForClass.Create(AClass, ParentClassStr);
   Result.FScrProject := Self;
   Add(Result.Name, Result);

   cl := TmscrVarForClass(Find(ParentClassStr));
   If cl <> Nil Then Result.FDefProperty := cl.DefProperty;

   If AClass.InheritsFrom(TPersistent) Then
      Classes.RegisterClass(TPersistentClass(AClass));
End;

Procedure TmscrScript.Register_Const(Const Name, Typ: String; Const Value: variant);
Var
   v: TmscrVarForVariable;
Begin
   If Find(Name) <> Nil Then Exit;

   v := TmscrVarForVariable.Create(Name, StrToVarType(Typ, Self), Typ);
   v.Value := Value;
   v.IsReadOnly := True;
   Add(v.Name, v);
End;

Procedure TmscrScript.Register_Enum(Const Typ, Names: String);
Var
   i: integer;
   v: TmscrVarForVariable;
   sl: TStringList;
Begin
   v := TmscrVarForVariable.Create(Typ, msvtEnum, Typ);
   Add(v.Name, v);

   sl := TStringList.Create;
   sl.CommaText := Names;

   Try
      For i := 0 To sl.Count - 1 Do
      Begin
         v := TmscrVarForVariable.Create(Trim(sl[i]), msvtEnum, Typ);
         v.Value := i;
         v.IsReadOnly := True;
         Add(v.Name, v);
      End;
   Finally
      sl.Free;
   End;
End;

Procedure TmscrScript.Register_EnumSet(Const Typ, Names: String);
Var
   sl: TStringList;
   i, j: integer;
   v: TmscrVarForVariable;
Begin
   v := TmscrVarForVariable.Create(Typ, msvtEnum, Typ);
   Add(v.Name, v);
   sl := TStringList.Create;
   sl.CommaText := Names;
   Try
      j := 1;
      For i := 0 To sl.Count - 1 Do
      Begin
         v := TmscrVarForVariable.Create(Trim(sl[i]), msvtEnum, Typ);
         v.Value := j;
         v.IsReadOnly := True;
         Add(v.Name, v);
         j := j * 2;
      End;
   Finally
      sl.Free;
   End;
End;

Procedure TmscrScript.Register_Method(Const Syntax: String; CallEvent: TmscrCallMethod; Const Category: String = ''; Const Description: String = '');
Var
   v: TmscrVarForMethod;
Begin
   v := TmscrVarForMethod.Create(Syntax, Self);
   v.FOnCall := CallEvent;
   If Description = '' Then
      v.FDescription := v.Name
   Else
      v.FDescription := Description;
   v.FCategory := Category;
   Add(v.Name, v);
End;

Procedure TmscrScript.Register_Object(Const Name: String; Obj: TObject);
Begin
   Register_Variable(Name, Obj.ClassName, fmscrInteger(Obj));
End;

Procedure TmscrScript.Register_Variable(Const Name, Typ: String; Const Value: variant);
Var
   v: TmscrVarForVariable;
Begin
   If Find(Name) <> Nil Then Exit;
   v := TmscrVarForVariable.Create(Name, StrToVarType(Typ, Self), Typ);
   v.Value := Value;
   v.OnGetVarValue := FOnGetVarValue;
   Add(v.Name, v);
End;

Procedure TmscrScript.Register_Form(Form: TComponent);
Begin
   Register_Component(Form);
End;

Procedure TmscrScript.Register_Component(Form: TComponent);
{$IFNDEF MAGSCR_NOFORMS}
Var
   i: integer;
   v: TmscrVarForClass;
{$ENDIF}
Begin
{$IFNDEF MAGSCR_NOFORMS}
   v := FindClass(Form.ClassName);
   If v = Nil Then
   Begin
      If Form.InheritsFrom(TForm) Then
         Register_Class(Form.ClassType, 'TForm')
      Else If Form.InheritsFrom(TDataModule) Then
         Register_Class(Form.ClassType, 'TDataModule')
      Else If Form.InheritsFrom(TFrame) Then
         Register_Class(Form.ClassType, 'TFrame')
      Else
         Exit;
      v := FindClass(Form.ClassName);
   End;
   For i := 0 To Form.ComponentCount - 1 Do
      v.AddComponent(Form.Components[i]);

   Register_Object(Form.Name, Form);
{$ENDIF}
End;

Procedure TmscrScript.ClearRTTI;
Var
   i: integer;
   item: TObject;
Begin
   If Not FRTTIAdded Then Exit;
   i := 0;
   While i < FItems.Count Do
   Begin
      item := FItems.Objects[i];
      If (item Is TmscrRTTILibrary) Or ((item Is TmscrVarForBase) And (TmscrVarForBase(item).AddedBy = TObject(1))) Then
      Begin
         item.Free;
         FItems.Delete(i);
      End
      Else
      Begin
         Inc(i);
      End;
   End;
   FRTTIAdded := False;
End;

Procedure TmscrScript.AddRTTI;
Var
   i: integer;
   rtti: TmscrRTTILibrary;
   obj: TClass;
Begin
   If FRTTIAdded Then Exit;
   AddedBy := TObject(1);
   For i := 0 To FRTTILibraries.Count - 1 Do
   Begin
      obj := TClass(FRTTILibraries[i]);
      rtti := TmscrRTTILibrary(obj.NewInstance);
      rtti.Create(Self);
      Add('', rtti);
   End;
   AddedBy := Nil;
   FRTTIAdded := True;
End;

Function TmscrScript.CallFunction(Const Name: String; Const Params: variant; sGlobal: boolean): variant;
Var
   p: TmscrVarForProc;
   i: integer;
   v: TmscrVarForBase;
Begin
   If sGlobal Then
      v := Find(Name)
   Else
      v := FindLocal(Name);

   If (v <> Nil) And (v Is TmscrVarForProc) Then
   Begin
      p := TmscrVarForProc(v);
      If VarIsArray(Params) Then
         For i := 0 To VarArrayHighBound(Params, 1) Do
            p.Params[i].Value := Params[i];

      Result := p.Value;
   End
   Else
   Begin
      Result := Null;
   End;
End;

Function TmscrScript.CallFunction1(Const Name: String; Var Params: variant; sGlobal: boolean): variant;
Var
   i: integer;
   v: TmscrVarForBase;
   p: TmscrVarForProc;
Begin
   If sGlobal Then
      v := Find(Name)
   Else
      v := FindLocal(Name);

   If (v <> Nil) And (v Is TmscrVarForProc) Then
   Begin
      p := TmscrVarForProc(v);
      If VarIsArray(Params) Then
         For i := 0 To VarArrayHighBound(Params, 1) Do
            p.Params[i].Value := Params[i];
      Result := p.Value;
      If VarIsArray(Params) Then
         For i := 0 To VarArrayHighBound(Params, 1) Do
            Params[i] := p.Params[i].Value;
   End
   Else
   Begin
      Result := Null;
   End;
End;

Function TmscrScript.CallFunction2(Const Func: TmscrVarForProc; Const Params: variant): variant;
Var
   i: integer;
Begin
   If (Func <> Nil) Then
   Begin
      If VarIsArray(Params) Then
         For i := 0 To VarArrayHighBound(Params, 1) Do
            Func.Params[i].Value := Params[i];
      Result := Func.Value;
   End
   Else
   Begin
      Result := Null;
   End;
End;

Function TmscrScript.Evaluate(Const Expression: String): variant;
Var
   p: TmscrScript;
   aProj: TmscrScript;
   SaveEvent: TmscrRunLineEvent;
Begin
   FEvaluteRiseError := False;
   Result := Null;

   If FProjRunning = Nil Then
      p := Self
   Else
      p := FProjRunning;

   aProj := TmscrScript.Create(Nil);
   If Not p.FRTTIAdded Then aProj.AddRTTI;
   aProj.Parent := p;
   aProj.OnGetVarValue := p.OnGetVarValue;
   SaveEvent := FOnRunLine;
   FOnRunLine := Nil;
   Try
      aProj.SyntaxType := SyntaxType;

      If CompareText(SyntaxType, _SyntaxType) = 0 Then
         aProj.Lines.Text := 'function mscrEvaluateFUNC: Variant; begin Result:= ' + Expression + ' end; begin end.';

      If Not aProj.Compile Then
      Begin
         Result := aProj.ErrorMsg;
         FEvaluteRiseError := True;
      End
      Else
      Begin
         Result := aProj.FindLocal('mscrEvaluateFUNC').Value;
      End;

   Finally
      aProj.Free;
      FOnRunLine := SaveEvent;
   End;
End;

Function TmscrScript.FindClass(Const Name: String): TmscrVarForClass;
Var
   Item: TmscrVarForBase;
Begin
   Item := Find(Name);
   If (Item <> Nil) And (Item Is TmscrVarForClass) Then
      Result := TmscrVarForClass(Item)
   Else
      Result := Nil;
End;

Function TmscrScript.FindRecord(Const Name: String): TmscrVarForRecord;
Var
   Item: TmscrVarForBase;
Begin
   Item := Find(Name);
   If (Item <> Nil) And (Item Is TmscrVarForRecord) Then
      Result := TmscrVarForRecord(Item)
   Else
      Result := Nil;
End;

Procedure TmscrScript.RunLine(Const UnitName, Index: String);
Var
   p: TmscrScript;
Begin
   p := Self;
   While p <> Nil Do
      If Assigned(p.FOnRunLine) Then
      Begin
         p.FOnRunLine(Self, UnitName, Index);
         break;
      End
      Else
      Begin
         p := p.FParent;
      End;
End;

Function TmscrScript.GetVariables(Index: String): variant;
Var
   v: TmscrVarForBase;
Begin
   v := Find(Index);
   If v <> Nil Then
      Result := v.Value
   Else
      Result := Null;
End;

Procedure TmscrScript.SetVariables(Index: String; Const Value: variant);
Var
   v: TmscrVarForBase;
Begin
   v := Find(Index);
   If v <> Nil Then
      v.Value := Value
   Else
      Register_Variable(Index, 'Variant', Value);
End;

Procedure TmscrScript.SetLines(Const Value: TStrings);
Begin
   FLines.Assign(Value);
End;

Procedure TmscrScript.SetProgName(Const Value: String);
Begin
   If Assigned(FProjRunning) Then
      FProjRunning.FProjName := Value
   Else
      FProjName := Value;
End;

Procedure TmscrScript.Terminate;

//.......................
   Procedure TerminateAll(Script: TmscrScript);
   Var
      i: integer;
   Begin
      Script.FCalledExit := True;
      Script.FTerminated := True;
      For i := 0 To Script.Count - 1 Do
         If Script.Items[i] Is TmscrVarForProc Then
            TerminateAll(TmscrVarForProc(Script.Items[i]).Project);
   End;

   //......................
Begin
   TerminateAll(Self);
End;

Procedure TmscrScript.AddCodeLine(Const UnitName, APos: String);
Var
   sl: TStringList;
   LineN: String;
   i: integer;
Begin
   i := FUnitLines.IndexOf(UnitName);

   If (i = -1) Then
   Begin
      sl := TStringList.Create;
      sl.Sorted := True;
      FUnitLines.AddObject(UnitName, sl);
   End
   Else
   Begin
      sl := TStringList(FUnitLines.Objects[i]);
   End;

   LineN := Copy(APos, 1, Pos(':', APos) - 1);
   If sl.IndexOf(LineN) = -1 Then
   Begin
      sl.Add(LineN);
   End;
End;

Function TmscrScript.IsExecutableLine(LineN: integer; Const UnitName: String = ''): boolean;
Var
   sl: TStringList;
   i: integer;
Begin
   Result := False;

   i := FUnitLines.IndexOf(UnitName);
   If (i = -1) Then Exit;

   sl := TStringList(FUnitLines.Objects[i]);
   If sl.IndexOf(IntToStr(LineN)) <> -1 Then
      Result := True;
End;

//=============TmscrStatement =============

Constructor TmscrStatement.Create(aProject: TmscrScript; Const UnitName, SrcPosition: String);
Begin
   Inherited Create;
   FScrProject := aProject;
   FSrcPosition := SrcPosition;
   FUnitName := UnitName;
End;

Function TmscrStatement.GetItem(Index: integer): TmscrStatement;
Begin
   Result := FItems[Index];
End;

Procedure TmscrStatement.Execute;
Var
   i: integer;
Begin
   FScrProject.ErrorPos := '';
   For i := 0 To Count - 1 Do
   Begin
      If FScrProject.FTerminated Then
         break;
      Try
         FScrProject.FLastSrcPosition := Items[i].FSrcPosition;
         Items[i].Execute;
      Except
         On E: Exception Do
         Begin
            If FScrProject.ErrorPos = '' Then
               FScrProject.ErrorPos := FScrProject.FLastSrcPosition;
            Raise;
         End;
      End;
      If FScrProject.FCalledBreak Or FScrProject.FCalledContinue Or FScrProject.FCalledExit Then
         break;
   End;
End;

Procedure TmscrStatement.RunLine;
Begin
   FScrProject.RunLine(FUnitName, FSrcPosition);
End;

//=============TmscrCodeStmAssign_Base =============

Destructor TmscrCodeStmAssign_Base.Destroy;
Begin
   FDScanner.Free;
   FExpression.Free;
   Inherited;
End;

Procedure TmscrCodeStmAssign_Base.Optimize;
Begin
   FVar := FDScanner;
   FExpr := FExpression;

   If FDScanner Is TmscrVarForVariableDScanner Then
      FVar := FDScanner.RefItem;
   If TmscrExpression(FExpression).SingleItem <> Nil Then
      FExpr := TmscrExpression(FExpression).SingleItem;
End;

Procedure TmscrCodeStmAssign_Base.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;
   FVar.Value := FExpr.Value;
End;

Procedure TmscrCodeStmAssign_Plus.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;
   FVar.Value := FVar.Value + FExpr.Value;
End;

Procedure TmscrCodeStmAssign_Minus.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;
   FVar.Value := FVar.Value - FExpr.Value;
End;

Procedure TmscrCodeStmAssign_Mul.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;
   FVar.Value := FVar.Value * FExpr.Value;
End;

Procedure TmscrCodeStmAssign_Div.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;
   FVar.Value := FVar.Value / FExpr.Value;
End;

//=============TmscrCodeStm_Call =============

Destructor TmscrCodeStm_Call.Destroy;
Begin
   FDScanner.Free;
   Inherited;
End;

Procedure TmscrCodeStm_Call.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then
      Exit;
   If FModificator = '' Then
   Begin
      FDScanner.NeedResult := False;
      FDScanner.Value;
   End
   Else If FModificator = '+' Then
      FDScanner.Value := FDScanner.Value + 1
   Else If FModificator = '-' Then
      FDScanner.Value := FDScanner.Value - 1;
End;

//=============TmscrCodeStm_If =============

Constructor TmscrCodeStm_If.Create(aProject: TmscrScript; Const UnitName, SrcPosition: String);
Begin
   Inherited;
   FElsesrcstm := TmscrStatement.Create(FScrProject, UnitName, SrcPosition);
End;

Destructor TmscrCodeStm_If.Destroy;
Begin
   FCondition.Free;
   FElsesrcstm.Free;
   Inherited;
End;

Procedure TmscrCodeStm_If.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;

   If boolean(FCondition.Value) = True Then
      Inherited Execute
   Else
      FElsesrcstm.Execute;
End;

//=============TmscrCodeStm_While =============

Destructor TmscrCodeStm_While.Destroy;
Begin
   FCondition.Free;
   Inherited;
End;

Procedure TmscrCodeStm_While.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;
   While boolean(FCondition.Value) = True Do
   Begin
      Inherited Execute;
      If FScrProject.FCalledBreak Or FScrProject.FCalledExit Then break;
      FScrProject.FCalledContinue := False;
   End;
   FScrProject.FCalledBreak := False;
End;

//=============TmscrCodeStm_For =============

Destructor TmscrCodeStm_For.Destroy;
Begin
   FValueBegin.Free;
   FValueEnd.Free;
   Inherited;
End;

Procedure TmscrCodeStm_For.Execute;
Var
   i, bValue, eValue: integer;
Begin
   Try
      bValue := FValueBegin.Value;
      eValue := FValueEnd.Value;
   Finally
      RunLine;
   End;
   If FScrProject.FTerminated Then Exit;

   If FDown Then
      For i := bValue Downto eValue Do
      Begin
         FVariable.FValue := i;
         Inherited Execute;
         If FScrProject.FCalledBreak Or FScrProject.FCalledExit Then break;
         FScrProject.FCalledContinue := False;
      End
   Else
      For i := bValue To eValue Do
      Begin
         FVariable.FValue := i;
         Inherited Execute;
         If FScrProject.FCalledBreak Or FScrProject.FCalledExit Then break;
         FScrProject.FCalledContinue := False;
      End;
   FScrProject.FCalledBreak := False;
End;

//=============TmscrCodeStm_Repeat =============

Procedure TmscrCodeStm_Repeat.Execute;
Begin
   RunLine;
   If FScrProject.FTerminated Then Exit;
   Repeat
      Inherited Execute;
      If FScrProject.FCalledBreak Or FScrProject.FCalledExit Then break;
      FScrProject.FCalledContinue := False;
   Until boolean(FCondition.Value) = Not FInverseCondition;
   FScrProject.FCalledBreak := False;
End;

Destructor TmscrCodeStm_Repeat.Destroy;
Begin
   FCondition.Free;
   Inherited;
End;

//=============TmscrCaseSelector =============

Destructor TmscrCaseSelector.Destroy;
Begin
   FSetExpression.Free;
   Inherited;
End;

Function TmscrCaseSelector.Check(Const Value: variant): boolean;
Begin
   Result := FSetExpression.Check(Value);
End;

//=============TmscrCodeStm_Case =============

Constructor TmscrCodeStm_Case.Create(aProject: TmscrScript; Const UnitName, SrcPosition: String);
Begin
   Inherited;
   FElsesrcstm := TmscrStatement.Create(FScrProject, UnitName, SrcPosition);
End;

Destructor TmscrCodeStm_Case.Destroy;
Begin
   FCondition.Free;
   FElsesrcstm.Free;
   Inherited;
End;

Procedure TmscrCodeStm_Case.Execute;
Var
   i: integer;
   Value: variant;
   Executed: boolean;
Begin
   Value := FCondition.Value;
   Executed := False;
   RunLine;
   If FScrProject.FTerminated Then Exit;

   For i := 0 To Count - 1 Do
      If TmscrCaseSelector(Items[i]).Check(Value) Then
      Begin
         Items[i].Execute;
         Executed := True;
         break;
      End;

   If Not Executed Then FElsesrcstm.Execute;
End;

//=============TmscrCodesrcstmry =============

Constructor TmscrCodesrcstmry.Create(aProject: TmscrScript; Const UnitName, SrcPosition: String);
Begin
   Inherited;
   FExceptSrcStm := TmscrStatement.Create(aProject, UnitName, SrcPosition);
End;

Destructor TmscrCodesrcstmry.Destroy;
Begin
   FExceptSrcStm.Free;
   Inherited;
End;

Procedure TmscrCodesrcstmry.Execute;
Var
   SaveExitCalled: boolean;
Begin
   RunLine;
   If FScrProject.FTerminated Then
      Exit;
   If IsExcept Then
   Begin
      Try
         Inherited Execute;
      Except
         On E: Exception Do
         Begin
            FScrProject.SetVariables('ExceptionClassName', E.ClassName);
            FScrProject.SetVariables('ExceptionMessage', E.Message);
            FScrProject.ErrorPos := FScrProject.FLastSrcPosition;
            ExceptSrcStm.Execute;
         End;
      End;
   End
   Else
   Begin
      Try
         Inherited Execute;
      Finally
         SaveExitCalled := FScrProject.FCalledExit;
         FScrProject.FCalledExit := False;
         ExceptSrcStm.Execute;
         FScrProject.FCalledExit := SaveExitCalled;
      End;
   End;
End;

//=============TmscrCodeStm_Break =============

Procedure TmscrCodeStm_Break.Execute;
Begin
   FScrProject.FCalledBreak := True;
End;

//=============TmscrCodeStm_Continue =============

Procedure TmscrCodeStm_Continue.Execute;
Begin
   FScrProject.FCalledContinue := True;
End;

//=============TmscrCodeStm_Exit =============

Procedure TmscrCodeStm_Exit.Execute;
Begin
   RunLine;
   FScrProject.FCalledExit := True;
End;

//=============TmscrCodeStm_With =============

Destructor TmscrCodeStm_With.Destroy;
Begin
   FDScanner.Free;
   Inherited;
End;

Procedure TmscrCodeStm_With.Execute;
Begin
   Inherited;
   FVariable.Value := FDScanner.Value;
End;

//=============TmscrEngArray =============

Constructor TmscrEngArray.Create(Const AName: String; DimCount: integer; Typ: TmscrVarType; Const TypeName: String);
Var
   i: integer;
Begin
   Inherited Create(AName, Typ, TypeName);

   If DimCount <> -1 Then
   Begin
      For i := 0 To DimCount - 1 Do

{$IFDEF CPU64}
         Add(TmscrVarForParamItem.Create('', msvtInt64, '', False, False));
{$ELSE}
         Add(TmscrVarForParamItem.Create('', msvtInt, '', False, False));
{$ENDIF}

   End
   Else
      For i := 0 To 2 Do

{$IFDEF CPU64}
         Add(TmscrVarForParamItem.Create('', msvtInt64, '', i > 0, False));
{$ELSE}
         Add(TmscrVarForParamItem.Create('', msvtInt, '', i > 0, False));
{$ENDIF}

End;

Destructor TmscrEngArray.Destroy;
Begin
   Inherited;
End;

Function TmscrEngArray.GetValue: variant;
Var
   DimCount: integer;
Begin
   DimCount := VarArrayDimCount(ParentRef.PValue^);
   Case DimCount Of
      1: Result := ParentRef.PValue^[Params[0].Value];
      2: Result := ParentRef.PValue^[Params[0].Value, Params[1].Value];
      3: Result := ParentRef.PValue^[Params[0].Value, Params[1].Value, Params[2].Value];
      Else
         Result := Null;
   End;
End;

Procedure TmscrEngArray.SetValue(Const Value: variant);
Var
   DimCount: integer;
Begin
   DimCount := VarArrayDimCount(ParentRef.PValue^);
   Case DimCount Of
      1: ParentRef.PValue^[Params[0].Value] := Value;
      2: ParentRef.PValue^[Params[0].Value, Params[1].Value] := Value;
      3: ParentRef.PValue^[Params[0].Value, Params[1].Value, Params[2].Value] := Value;
   End;
End;

//=============TmscrEngString =============

Constructor TmscrEngString.Create;
Begin
   Inherited Create('__HelperForString', msvtChar, '');
   Add(TmscrVarForParamItem.Create('', msvtInt, '', False, False));
End;

Function TmscrEngString.GetValue: variant;
Begin
   Result := String(ParentValue)[integer(Params[0].Value)];
End;

Procedure TmscrEngString.SetValue(Const Value: variant);
Var
   s: String;
Begin
   s := ParentValue;
   s[integer(Params[0].Value)] := String(Value)[1];
   TmscrVarForBase(fmscrInteger(ParentRef)).Value := s;
End;

//============= TmscrRTTILibrary =============

Constructor TmscrRTTILibrary.Create(AScript: TmscrScript);
Begin
   FScript := AScript;
End;

Function MSCR_GlobalScripter: TmscrScript;
Begin
   If (FGlobalScripter = Nil) And Not FGlobalScripterIsNIL Then
   Begin
      FGlobalScripter := TmscrScript.Create(Nil);
      FGlobalScripter.AddRTTI;
   End;
   Result := FGlobalScripter;
End;

Function MSCR_IsGlobalScripterExist: boolean;
Begin
   Result := Assigned(FGlobalScripter);
End;

Function MSCR_RTTILibraries: TList;
Begin
   If (FRTTILibraries = Nil) And Not FRTTILibrariesIsNIL Then
   Begin
      FRTTILibraries := TList.Create;
      FRTTILibraries.Add(TmscrSysFunctions);
   End;
   Result := FRTTILibraries;
End;

//============= TmscrVarForVariable =============

Function TmscrVarForVariable.GetValue: variant;
Begin
   Result := Inherited GetValue;
   If Assigned(FOnGetVarValue) Then
   Begin
      Result := FOnGetVarValue(FName, FTyp, FValue);
      If Result = null Then
         Result := FValue;
   End;
End;

Procedure TmscrVarForVariable.SetValue(Const Value: variant);
Begin
   If Not FIsReadOnly Then
      Case FTyp Of
         msvtInt: FValue := VarAsType(Value, varInteger);
         msvtInt64: FValue := VarAsType(Value, varInt64);
         msvtBool: FValue := VarAsType(Value, varBoolean);

         msvtFloat: If (VarType(Value) = varDate) Then
               FValue := VarAsType(Value, varDate)
            Else
               FValue := VarAsType(Value, varDouble);

         msvtString: FValue := VarAsType(Value, varString);

         Else
            FValue := Value;
      End;
End;

//==========================================

Type
   TByteSet = Set Of 0..7;
   PByteSet = ^TByteSet;

//********************* System Events **********************************

//============= TmscrCustomEvent =========================

Constructor TmscrCustomEvent.Create(AObject: TObject; AHandler: TmscrVarForProc);
Begin
   FInstance := AObject;
   FHandler := AHandler;
End;

Procedure TmscrCustomEvent.CallHandler(Params: Array Of Const);
Var
   i, StackIndex: integer;
Begin
   StackIndex := -1;
   If FHandler.Executing Then
      StackIndex := FHandler.SaveLocalVariables;
   Try
      For i := 0 To FHandler.Count - 1 Do
         FHandler.Params[i].Value := VarRecToVariant(Params[i]);
      FHandler.Value;
   Finally
      FHandler.RestoreLocalVariables(StackIndex, True);
   End;
End;

class function TmscrCustomEvent.GetDeclaration: String;
begin
   Result:='Nothing here yet...'
end;

//============= TmscrNotifyEvent =============

Procedure TmscrNotifyEvent.DoEvent(Sender: TObject);
Begin
   CallHandler([Sender]);
End;

Function TmscrNotifyEvent.GetMethod: Pointer;
Begin
   Result := @TmscrNotifyEvent.DoEvent;
End;

class function TmscrNotifyEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender: TObject);';
end;

//============= TmscrMouseEvent =============

Procedure TmscrMouseEvent.DoEvent(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
Var
   b: byte;
Begin
   b := byte(PByteSet(@Shift)^);
   CallHandler([Sender, integer(Button), b, X, Y]);
End;

Function TmscrMouseEvent.GetMethod: Pointer;
Begin
   Result := @TmscrMouseEvent.DoEvent;
End; 

class function TmscrMouseEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);';
end;

//=============TmscrMouseMoveEvent =============

Procedure TmscrMouseMoveEvent.DoEvent(Sender: TObject; Shift: TShiftState; X, Y: integer);
Var
   b: byte;
Begin
   b := byte(PByteSet(@Shift)^);
   CallHandler([Sender, b, X, Y]);
End;

Function TmscrMouseMoveEvent.GetMethod: Pointer;
Begin
   Result := @TmscrMouseMoveEvent.DoEvent;
End;

class function TmscrMouseMoveEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender: TObject; Shift: TShiftState; X, Y: Integer);';
end;

//=============TmscrKeyEvent =============

Procedure TmscrKeyEvent.DoEvent(Sender: TObject; Var Key: word; Shift: TShiftState);
Var
   b: byte;
Begin
   b := byte(PByteSet(@Shift)^);
   CallHandler([Sender, Key, b]);
   Key := Handler.Params[1].Value;
End;

Function TmscrKeyEvent.GetMethod: Pointer;
Begin
   Result := @TmscrKeyEvent.DoEvent;
End;   

class function TmscrKeyEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender: TObject; var Key: Word; Shift: TShiftState);';
end;

//=============TmscrKeyPressEvent =============

Procedure TmscrKeyPressEvent.DoEvent(Sender: TObject; Var Key: char);
Begin
   CallHandler([Sender, Key]);
   Key := String(Handler.Params[1].Value)[1];
End;

Function TmscrKeyPressEvent.GetMethod: Pointer;
Begin
   Result := @TmscrKeyPressEvent.DoEvent;
End;     

class function TmscrKeyPressEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender: TObject; var Key: char)';
end;

//=============TmscrCloseEvent =============

Procedure TmscrCloseEvent.DoEvent(Sender: TObject; Var Action: TCloseAction);
Begin
   CallHandler([Sender, integer(Action)]);
   Action := Handler.Params[1].Value;
End;

Function TmscrCloseEvent.GetMethod: Pointer;
Begin
   Result := @TmscrCloseEvent.DoEvent;
End;

class function TmscrCloseEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender: TObject; var CloseAction: TCloseAction);';
end;

//=============TmscrCloseQueryEvent =============

Procedure TmscrCloseQueryEvent.DoEvent(Sender: TObject; Var CanClose: boolean);
Begin
   CallHandler([Sender, CanClose]);
   CanClose := Handler.Params[1].Value;
End;

Function TmscrCloseQueryEvent.GetMethod: Pointer;
Begin
   Result := @TmscrCloseQueryEvent.DoEvent;
End;  

class function TmscrCloseQueryEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender : TObject; var CanClose: Boolean) of object;';
end;

//=============TmscrCanResizeEvent =============

Procedure TmscrCanResizeEvent.DoEvent(Sender: TObject; Var NewWidth, NewHeight: integer; Var Resize: boolean);
Begin
   CallHandler([Sender, NewWidth, NewHeight, Resize]);
   NewWidth := Handler.Params[1].Value;
   NewHeight := Handler.Params[2].Value;
   Resize := Handler.Params[3].Value;
End;

Function TmscrCanResizeEvent.GetMethod: Pointer;
Begin
   Result := @TmscrCanResizeEvent.DoEvent;
End;  

class function TmscrCanResizeEvent.GetDeclaration: String;
begin
   Result:='procedure(Sender: TObject; var NewSize: Integer; var Accept: Boolean);';
end;

//====================================================================

Function StrToVarType(Const TypeName: String; Script: TmscrScript): TmscrVarType;
Var
   v: TmscrVarForBase;
Begin
   v := Script.Find(TypeName);
   If v = Nil Then
      Result := msvtClass
   Else
      Result := v.Typ;
End;

Function ClassesCompatible(Const Class1, Class2: String; Script: TmscrScript): boolean;
Var
   cl1, cl2: TmscrVarForClass;
Begin
   Result := False;
   cl1 := Script.FindClass(Class1);
   cl2 := Script.FindClass(Class2);
   If (cl1 <> Nil) And (cl2 <> Nil) Then
      Result := cl2.ClassRef.InheritsFrom(cl1.ClassRef);
End;

Function RecordsCompatible(Const Record1, Record2: String; Script: TmscrScript): boolean;
Var
   cl1, cl2: TmscrVarForRecord;
Begin
   Result := False;
   cl1 := Script.FindRecord(Record1);
   cl2 := Script.FindRecord(Record2);
   If (cl1 <> Nil) And (cl2 <> Nil) Then
      Result := (cl1.TypeInfo = cl2.TypeInfo);
End;

Function TypesCompatible(Typ1, Typ2: TmscrTypeRec; Script: TmscrScript): boolean;
Begin
   Result := False;

   Case Typ1.Typ Of
      msvtInt:
         Result := Typ2.Typ In [msvtInt, msvtFloat, msvtVariant, msvtEnum, msvtInt64];
      msvtInt64:
         Result := Typ2.Typ In [msvtInt64, msvtInt, msvtFloat, msvtVariant, msvtEnum];
      msvtFloat:
         Result := Typ2.Typ In [msvtInt, msvtFloat, msvtVariant];
      msvtBool:
         Result := Typ2.Typ In [msvtBool, msvtVariant];
      msvtChar, msvtString:
         Result := Typ2.Typ In [msvtChar, msvtString, msvtVariant];
      msvtClass:
         Result := (Typ2.Typ = msvtVariant) Or ((Typ2.Typ = msvtClass) And ClassesCompatible(Typ1.TypeName, Typ2.TypeName, Script));
      msvtRecord:
         Result := (Typ2.Typ = msvtVariant) Or ((Typ2.Typ = msvtRecord) And RecordsCompatible(Typ1.TypeName, Typ2.TypeName, Script));
      msvtArray:
         Result := Typ2.Typ In [msvtArray, msvtVariant];
      msvtVariant:
         Result := True;
      msvtEnum:
         Begin
            Result := Typ2.Typ In [msvtInt, msvtInt64, msvtVariant, msvtEnum];
            If Typ2.Typ = msvtEnum Then
               Result := AnsiCompareText(Typ1.TypeName, Typ2.TypeName) = 0;
         End;
   End;
End;

Function AssignCompatible(Var1, Var2: TmscrVarForBase; Script: TmscrScript): boolean;
Var
   t1, t2: TmscrTypeRec;
Begin
   t1.Typ := Var1.Typ;
   t1.TypeName := Var1.TypeName;
   t2.Typ := Var2.Typ;
   t2.TypeName := Var2.TypeName;

   Result := TypesCompatible(t1, t2, Script);
   If Result And (Var1.Typ = msvtInt) And (Var2.Typ = msvtFloat) Then
      Result := False;
End;

Function VarRecToVariant(v: TVarRec): variant;
Begin
   With v Do
      Case VType Of
         vtInt64:
            Result := VInt64^;

{$IFDEF CPU64}
         vtInteger:
            Result := VInteger;
         vtObject:
            Result := fmscrInteger(VObject);
{$ELSE}
         vtInteger, vtObject:
            Result := VInteger;
{$ENDIF}
         vtBoolean:
            Result := VBoolean;
         vtExtended, vtCurrency:
            Result := VExtended^;
         vtChar:
            Result := VChar;
         vtString:
            Result := VString^;
         vtAnsiString:
            Result := ansistring(VAnsiString);
         vtVariant:
            Result := VVariant^;
         Else
            Result := Null;
      End;
End;

Procedure VariantToVarRec(v: variant; Var ar: TmscrVarRecArray; Var sPtrList: TList);
Var
   i: integer;
   pAStr: PAnsiString;
Begin
   SetLength(ar, VarArrayHighBound(v, 1) + 1);

   sPtrList := TList.Create;
   For i := 0 To VarArrayHighBound(v, 1) Do
      Case TVarData(v[i]).VType Of
         varSmallint, varInteger, varByte, varShortInt, varWord, varLongWord:
            Begin
               ar[i].VType := vtInteger;
               ar[i].VInteger := v[i];
            End;
         varInt64:
            Begin
               ar[i].VType := vtInt64;
               New(ar[i].VInt64);
               ar[i].VInt64^ := v[i];
            End;
         varSingle, varDouble, varCurrency, varDate:
            Begin
               ar[i].VType := vtExtended;
               New(ar[i].VExtended);
               ar[i].VExtended^ := v[i];
            End;
         varBoolean:
            Begin
               ar[i].VType := vtBoolean;
               ar[i].VBoolean := v[i];
            End;
         varString:
            Begin
               ar[i].VType := vtString;
               New(ar[i].VString);
               ar[i].VString^ := v[i];
            End;
         varOleStr:
            Begin
               ar[i].VType := vtAnsiString;
               New(pAStr);
               sPtrList.Add(pAStr);
               pAStr^ := ansistring(v[i]);
               ar[i].VAnsiString := PAnsiString(pAStr^);
            End;
         varVariant:
            Begin
               ar[i].VType := vtVariant;
               New(ar[i].VVariant);
               ar[i].VVariant^ := v[i];
            End;
      End;
End;

Procedure ClearVarRec(Var ar: TmscrVarRecArray; Var sPtrList: TList);
Var
   i: integer;
Begin
   For i := 0 To Length(ar) - 1 Do
      If ar[i].VType In [vtExtended, vtString, vtVariant, vtInt64] Then
         Dispose(ar[i].VExtended);

   For i := 0 To sPtrList.Count - 1 Do
      Dispose(sPtrList[i]);

   sPtrList.Free;
   Finalize(ar);
End;

Function ParserStringToVariant(s: String): variant;
Var
   i: int64;
   k: integer;
   iPos: integer;
Begin
   Result := Null;
   If s <> '' Then
      If s[1] = '''' Then
         Result := Copy(s, 2, Length(s) - 2)
      Else
      Begin
         Val(s, i, k);
         If k = 0 Then

            If i > MaxInt Then
               Result := i
            Else
               Result := integer(i)

         Else
         Begin

            If DecimalSeparator <> '.' Then
            Begin
               iPos := Pos('.', s);
               If iPos > 0 Then
                  s[iPos] := DecimalSeparator;
            End;
            Result := StrToFloat(s);
         End;
      End;
End;

Function ParseMethodSyntax(Const Syntax: String; Script: TmscrScript): TmscrVarForBase;
Var
   Parser: TmscrParser;
   i, j: integer;
   Name, Params, TypeName, s: String;
   isFunc, isMacro, varParam: boolean;
   InitValue: variant;
   IsParamWithDef: boolean;
   v: TmscrVarForBase;

   //................................
   Procedure AddParams;
   Var
      i: integer;
      p: TmscrVarForParamItem;
      sl: TStringList;
   Begin
      sl := TStringList.Create;
      Try
         Delete(Params, Length(Params), 1);
         sl.CommaText := Params;
         For i := 0 To sl.Count - 2 Do
         Begin
            p := TmscrVarForParamItem.Create(sl[i], StrToVarType(TypeName, Script), TypeName, False, varParam);
            Result.Add(p);
         End;
         p := TmscrVarForParamItem.Create(sl[sl.Count - 1], StrToVarType(TypeName, Script), TypeName, IsParamWithDef, varParam);
         p.DefValue := InitValue;
         Result.Add(p);

      Finally
         sl.Free;
      End;
   End;

   //...................................
Begin
   Parser := TmscrParser.Create;
   If Script.UseExtendCharset Then
      For i := 128 To 255 Do
         Parser.IdentifierCharset := Parser.IdentifierCharset + [Chr(i)];
   Parser.Text := Syntax;

   s := Parser.GetIdent;
   isMacro := Pos('macro', AnsiLowercase(s)) = 1;
   If isMacro Then
      s := Copy(s, 6, 255);
   isFunc := CompareText(s, 'function') = 0;
   Name := Parser.GetIdent;

   If isFunc Then
   Begin
      j := Length(Syntax);
      While (Syntax[j] <> ':') And (j <> 0) Do
         Dec(j);

      i := Parser.Position;
      Parser.Position := j + 1;
      TypeName := Parser.GetIdent;
      Parser.Position := i;
   End
   Else
   Begin
      TypeName := '';
   End;

   Result := TmscrVarForBase.Create(Name, StrToVarType(TypeName, Script), TypeName);
   Result.IsMacro := IsMacro;

   Parser.SkipSpaces;
   s := Parser.GetChar;
   If s = '(' Then
   Begin
      Repeat
         varParam := False;
         Params := '';

         Repeat
            s := Parser.GetIdent;
            If CompareText(s, 'var') = 0 Then
               varParam := True
            Else If CompareText(s, 'const') = 0 Then // do nothing
            Else
               Params := Params + s + ',';
            Parser.SkipSpaces;
            s := Parser.GetChar;
            If s = ':' Then
            Begin
               TypeName := Parser.GetIdent;
               Parser.SkipSpaces;
               i := Parser.Position;
               IsParamWithDef := False;
               If Parser.GetChar = '=' Then
               Begin
                  IsParamWithDef := True;
                  s := Parser.GetNumber;

                  If s = '' Then s := Parser.GetString;

                  If s = '' Then
                  Begin
                     i := Parser.Position;
                     s := Parser.GetChar;

                     If s = '-' Then
                        s := '-' + Parser.GetNumber
                     Else
                        Parser.Position := i;
                  End;

                  If s <> '' Then
                  Begin
                     InitValue := ParserStringToVariant(s);
                  End
                  Else
                  Begin
                     s := Parser.GetIdent;
                     v := Script.Find(s);
                     If v <> Nil Then
                        InitValue := v.Value
                     Else
                        InitValue := Null;
                  End;

               End
               Else
               Begin
                  InitValue := Null;
                  Parser.Position := i;
               End;
               AddParams;
               s := ';';
            End
            Else If s = ')' Then
            Begin
               Parser.Position := Parser.Position - 1;
               break;
            End;
         Until s = ';';

         Parser.SkipSpaces;
      Until Parser.GetChar = ')';
   End;

   Parser.Free;
End;

Function MSCR_PosToPoint(Const ErrorPos: String): TPoint;
Begin
   Result.X := 0;
   Result.Y := 0;
   If ErrorPos <> '' Then
   Begin
      Result.Y := StrToInt(Copy(ErrorPos, 1, Pos(':', ErrorPos) - 1));
      Result.X := StrToInt(Copy(ErrorPos, Pos(':', ErrorPos) + 1, 255));
   End;
End;

Procedure GenerateDATAContents(aProj: TmscrScript; Item: TmscrDATAItem; FunctionsOnly: boolean = False);
Var
   i, j: integer;
   v: TmscrVarForBase;
   c: TmscrVarForClass;
   r: TmscrVarForRecord;
   xi: TmscrDATAItem;
   clItem: TmscrEngCustom;
   s: String;
Begin
   Item.FindItem('Functions');
   Item.FindItem('Classes');
   Item.FindItem('Records');
   Item.FindItem('Types');
   Item.FindItem('Variables');
   Item.FindItem('Constants');

   For i := 0 To aProj.Count - 1 Do
   Begin
      v := aProj.Items[i];
      If Not (v Is TmscrVarForMethod) And FunctionsOnly Then continue;

      If v Is TmscrVarForMethod Then
      Begin
         xi := Item.FindItem('Functions');
         xi := xi.FindItem(TmscrVarForMethod(v).Category);
         xi.Text := 'text="' + xi.Name + '"';
         With xi.Add Do
         Begin
            Name := 'item';
            s := TmscrVarForMethod(v).Syntax;
            Text := 'text="' + s + '" description="' + TmscrVarForMethod(v).Description + '"';
         End;
      End
      Else If v Is TmscrVarForClass Then  {//---------------- Classes ----------------------}

      Begin
         c := TmscrVarForClass(v);
         xi := Item.FindItem('Classes');
         xi := xi.Add;
         With xi Do
         Begin
            Name := 'item';
            Text := 'text="' + c.Name + ' = class(' + c.ParentClassStr + ')"';
         End;

         For j := 0 To c.MembersCount - 1 Do
         Begin
            clItem := c.Members[j];
            With xi.Add Do
            Begin
               Name := 'item';
               Text := 'text="';
               If clItem Is TmscrEngProperty Then
                  Text := Text + 'property ' + clItem.Name + ': ' + clItem.GetFullTypeName + '"'
               Else If clItem Is TmscrVarForMethod Then
               Begin
                  s := TmscrVarForMethod(clItem).Syntax;
                  If TmscrVarForMethod(clItem).IndexMethod Then
                     s := 'index property' + Copy(s, Pos(' ', s), 255);
                  Text := Text + s + '"';
               End
               Else
                  Text := Text + 'event ' + clItem.Name + ' := ' + clItem.Declaration + '"'
            End;
         End;

      End
      Else If v Is TmscrVarForRecord Then  {//--------------- Records -----------------------}
      Begin

         r := TmscrVarForRecord(v);
         xi := Item.FindItem('Records');
         xi := xi.Add;

         With xi Do
         Begin
            Name := 'item';
            Text := 'text="' + r.Name + ' = ' + r.TypeInfo.Name + '"';
         End;

         For j := 0 To r.MembersCount - 1 Do
         Begin
            clItem := r.Members[j];
            With xi.Add Do
            Begin
               Name := 'item';
               Text := 'text="';
               If clItem Is TmscrEngRecElement Then
                  Text := Text + clItem.Name + ': ' + clItem.GetFullTypeName + '"';

            End;
         End;

      End
      Else If v Is TmscrVarForVariable Then  {//------------- Variables -------------------------}
      Begin
         If v.Typ = msvtEnum Then
         Begin
            xi := Item.FindItem('Types');
            With xi.FindItem(v.TypeName) Do
            Begin
               If v.Name <> v.TypeName Then
                  If Text = '' Then
                     Text := v.Name
                  Else
                     Text := Text + ',' + v.Name;
            End;
         End
         Else
         Begin

            If v.IsReadOnly Then
               xi := Item.FindItem('Constants')
            Else
               xi := Item.FindItem('Variables');

            With xi.Add Do
            Begin
               Name := 'item';
               Text := 'text="' + v.Name + ': ' + v.GetFullTypeName;

               If v.IsReadOnly Then
                  Text := Text + ' = ' + VarToStr(v.Value);

               Text := Text + '"';
            End;
         End;
      End;
   End;

   xi := Item.FindItem('types');
   For i := 0 To xi.Count - 1 Do
      If xi[i].Name <> 'item' Then
      Begin
         xi[i].Text := 'text="' + xi[i].Name + ': (' + xi[i].Text + ')"';
         xi[i].Name := 'item';
      End;
End;

Procedure GenerateMembers(aProj: TmscrScript; cl: TClass; Item: TmscrDATAItem);
Var
   i, j: integer;
   v: TmscrVarForBase;
   c: TmscrVarForClass;
   xi: TmscrDATAItem;
   clItem: TmscrEngCustom;
   s: String;
Begin
   For i := 0 To aProj.Count - 1 Do
   Begin
      v := aProj.Items[i];
      If v Is TmscrVarForClass Then
      Begin
         c := TmscrVarForClass(v);
         If cl.InheritsFrom(c.ClassRef) Then
         Begin
            xi := Item;
            For j := 0 To c.MembersCount - 1 Do
            Begin
               clItem := c.Members[j];
               With xi.Add Do
               Begin
                  Name := 'item';
                  Text := 'text="';
                  If clItem Is TmscrEngProperty Then
                     Text := Text + 'property ' + clItem.Name + ': ' + clItem.GetFullTypeName + '"'
                  Else If clItem Is TmscrVarForMethod Then
                  Begin
                     s := TmscrVarForMethod(clItem).Syntax;
                     If TmscrVarForMethod(clItem).IndexMethod Then
                        s := 'index property' + Copy(s, Pos(' ', s), 255);
                     Text := Text + s + '"';
                  End
                  Else
                     Text := Text + 'event ' + clItem.Name + '"';
               End;
            End;
         End;
      End;
   End;
End;

Function fsSetToString(PropInfo: PPropInfo; Const Value: variant): String;
Var
   S: TIntegerSet;
   TypeInfo: PTypeInfo;
   I: integer;
Begin
   Result := '';
   TypeInfo := GetTypeData(PropInfo^.PropType)^.CompType;

   integer(S) := 0;
   If VarIsArray(Value) Then
      For I := 0 To VarArrayHighBound(Value, 1) Do
      Begin
         integer(S) := integer(S) Or Value[I];
      End;

   For I := 0 To SizeOf(integer) * 8 - 1 Do
      If I In S Then
      Begin
         If Result <> '' Then
            Result := Result + ',';
         Result := Result + GetEnumName(TypeInfo, I);
      End;
   Result := '[' + Result + ']';
End;

//====================================================

Type

   TmsAction_None = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Great = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Less = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_LessEq = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_GreatEq = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_NonEq = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Eq = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Plus = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_StrCat = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Minus = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Or = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Xor = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Mul = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_DivFloat = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_DivInt = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Mod = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_And = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Shl = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Shr = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_LeftBracket = Class(TmscrActionBase);

   TmsAction_RightBracket = Class(TmscrActionBase);

   TmsAction_Not = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_UnMinus = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_In = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   TmsAction_Is = Class(TmscrActionBase)
   Protected
      Function GetValue: variant; Override;
   End;

   //vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv

Function TmsAction_None.GetValue: variant;
Begin
   Result := FLeft.Value;
End;

Function TmsAction_Great.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result > FRight.Value;
End;

Function TmsAction_Less.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result < FRight.Value;
End;

Function TmsAction_LessEq.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result <= FRight.Value;
End;

Function TmsAction_GreatEq.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result >= FRight.Value;
End;

Function TmsAction_NonEq.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result <> FRight.Value;
End;

Function TmsAction_Eq.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result = FRight.Value;
End;

Function TmsAction_Plus.GetValue: variant;
Begin
   Result := FLeft.Value;
   If TVarData(Result).Vtype = varEmpty Then
      Result := 0;
   Result := Result + FRight.Value;
End;

Function TmsAction_StrCat.GetValue: variant;
Begin
   Result := FLeft.Value;
   If (TVarData(Result).VType = varString) Then
      Result := VarToStr(Result) + VarToStr(FRight.Value)
   Else
      Result := Result + FRight.Value;
End;

Function TmsAction_Minus.GetValue: variant;
Begin
   Result := FLeft.Value;
   If FOptimizeInt Then
   Begin
      Result := fmscrInteger(Result) - fmscrInteger(FRight.Value);
   End
   Else
   Begin
      If TVarData(Result).Vtype = varEmpty Then
         Result := 0;
      Result := Result - FRight.Value;
   End;
End;

Function TmsAction_Or.GetValue: variant;
Begin
   Result := FLeft.Value;

   If FOptimizeBool Then
   Begin
      If boolean(Result) = False Then
         Result := FRight.Value;
   End
   Else
   Begin
      Result := Result Or FRight.Value;
   End;
End;

Function TmsAction_Xor.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result Xor FRight.Value;
End;

Function TmsAction_Mul.GetValue: variant;
Begin
   Result := FLeft.Value;
   If FOptimizeInt Then
   Begin
      Result := fmscrInteger(Result) * fmscrInteger(FRight.Value);
   End
   Else
   Begin
      If TVarData(Result).Vtype = varEmpty Then
         Result := 0;
      Result := Result * FRight.Value;
   End;
End;

Function TmsAction_DivFloat.GetValue: variant;
Begin
   Result := FLeft.Value;
   If TVarData(Result).Vtype = varEmpty Then
      Result := 0;
   Result := Result / FRight.Value;
End;

Function TmsAction_DivInt.GetValue: variant;
Begin
   Result := FLeft.Value;
   If FOptimizeInt Then
   Begin
      Result := fmscrInteger(Result) Div fmscrInteger(FRight.Value);
   End
   Else
   Begin
      If TVarData(Result).Vtype = varEmpty Then
         Result := 0;
      Result := Result Div FRight.Value;
   End;
End;

Function TmsAction_And.GetValue: variant;
Begin
   Result := FLeft.Value;
   If FOptimizeBool Then
   Begin
      If boolean(Result) = True Then
         Result := FRight.Value;
   End
   Else
   Begin
      Result := Result And FRight.Value;
   End;
End;

Function TmsAction_Mod.GetValue: variant;
Begin
   Result := FLeft.Value;
   If FOptimizeInt Then
   Begin
      Result := fmscrInteger(Result) Mod fmscrInteger(FRight.Value);
   End
   Else
   Begin
      If TVarData(Result).Vtype = varEmpty Then
         Result := 0;
      Result := Result Mod FRight.Value;
   End;
End;

Function TmsAction_Shl.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result Shl FRight.Value;
End;

Function TmsAction_Shr.GetValue: variant;
Begin
   Result := FLeft.Value;
   Result := Result Shr FRight.Value;
End;

Function TmsAction_Not.GetValue: variant;
Begin
   Result := Not FLeft.Value;
End;

Function TmsAction_UnMinus.GetValue: variant;
Begin
   Result := -FLeft.Value;
End;

Function TmsAction_In.GetValue: variant;
Var
   i: integer;
   ar, val, selfVal: variant;
   Count: integer;
Begin
   If FRight Is TmscrSetNode Then
   Begin
      Result := TmscrSetNode(FRight).FSetExpression.Check(FLeft.Value);
   End
   Else
   Begin
      Result := False;
      ar := FRight.Value;
      Count := VarArrayHighBound(ar, 1);
      selfVal := FLeft.Value;

      i := 0;
      While i <= Count Do
      Begin
         val := ar[i];
         Result := selfVal = val;
         If (i < Count - 1) And (ar[i + 1] = Null) And Not Result Then
         Begin
            Result := (selfVal >= val) And (selfVal <= ar[i + 2]);
            Inc(i, 2);
         End;

         If Result Then BREAK;
         Inc(i);
      End;
   End;
End;

Function TmsAction_Is.GetValue: variant;
Begin
   Result := TObject(fmscrInteger(FLeft.Value)) Is TmscrVarForClass(TmscrDScannerNode(FRight).FDScanner[0].Ref).ClassRef;
End;

//=================TmscrExpressionNode ============================

Destructor TmscrExpressionNode.Destroy;
Begin
   FLeft.Free;
   FRight.Free;
   Inherited;
End;

Procedure TmscrExpressionNode.RemoveNode(Node: TmscrExpressionNode);
Begin
   If FLeft = Node Then
      FLeft := Nil
   Else If FRight = Node Then
      FRight := Nil;
End;

Procedure TmscrExpressionNode.AddNode(Node: TmscrExpressionNode);
Begin
   If FLeft = Nil Then
      FLeft := Node
   Else If FRight = Nil Then
      FRight := Node;

   If Node <> Nil Then
      Node.FParent := Self;
End;

//=================TmscrActionNode ========================

Constructor TmscrActionNode.Create(Const AValue: variant);
Var
   t: TmscrVarType;
Begin

{$IFDEF CPU64}
   Inherited Create('', msvtInt64, '');
{$ELSE}
   Inherited Create('', msvtInt, '');
{$ENDIF}

   Value := AValue;

   t := msvtInt;
   If TVarData(AValue).VType = varBoolean Then
      t := msvtBool
   Else If TVarData(AValue).VType In [varSingle, varDouble, varCurrency] Then
      t := msvtFloat
   Else If (TVarData(AValue).VType = varInt64) Then
      t := msvtInt64
   Else If (TVarData(AValue).VType = varOleStr) Or (TVarData(AValue).VType = varString) Then
      t := msvtString;

   Typ := t;
End;

Function TmscrActionNode.Priority: integer;
Begin
   Result := 0;
End;

//=================TmscrActionBase ====================

Constructor TmscrActionBase.Create(Op: TmscrOperatorType);
Begin

{$IFDEF CPU64}
   Inherited Create('', msvtInt64, '');
{$ELSE}
   Inherited Create('', msvtInt, '');
{$ENDIF}

   FOp := Op;
End;

Function TmscrActionBase.Priority: integer;
Begin
   Case FOp Of
      msopNone:
         Result := 7;
      msopLeftBracket:
         Result := 6;
      msopRightBracket:
         Result := 5;
      msopGreat, msopLess, msopGreatEq, msopLessEq, msopNonEq, msopEq, msopIn, msopIs:
         Result := 4;
      msopPlus, msopMinus, msopOr, msopXor:
         Result := 3;
      msopMul, msopDivFloat, msopDivInt, msopMod, msopAnd, msopShr, msopShl:
         Result := 2;
      msopNot, msopUnMinus:
         Result := 1;
      Else
         Result := 0;
   End;
End;

//=================TmscrDScannerNode ========================

Constructor TmscrDScannerNode.Create(ADScanner: TmscrDScanner);
Begin
   Inherited Create(0);
   FDScanner := ADScanner;
   Typ := ADScanner.Typ;
   TypeName := ADScanner.TypeName;
   If FDScanner Is TmscrVarForVariableDScanner Then
      FVar := FDScanner.RefItem
   Else
      FVar := FDScanner;
End;

Destructor TmscrDScannerNode.Destroy;
Begin
   FDScanner.Free;
   Inherited;
End;

Function TmscrDScannerNode.GetValue: variant;
Begin
   Result := FVar.Value;
End;

//=================TmscrSetNode ==============================

Constructor TmscrSetNode.Create(ASet: TmscrSetExpression);
Begin
   Inherited Create(0);
   FSetExpression := ASet;
   Typ := msvtVariant;
End;

Destructor TmscrSetNode.Destroy;
Begin
   FSetExpression.Free;
   Inherited;
End;

Function TmscrSetNode.GetValue: variant;
Begin
   Result := FSetExpression.Value;
End;

//================= TmscrExpression =============================

Function TmscrExpression.GetValue: variant;
Begin
   Result := FNode.Value;
End;

Constructor TmscrExpression.Create(Script: TmscrScript);
Begin

{$IFDEF CPU64}
   Inherited Create('', msvtInt64, '');
{$ELSE}
   Inherited Create('', msvtInt, '');
{$ENDIF}

   FNode := TmsAction_None.Create(msopNone);
   FCurNode := FNode;
   FScript := Script;
End;

Destructor TmscrExpression.Destroy;
Begin
   FNode.Free;
   Inherited;
End;

Procedure TmscrExpression.AddOperator(Const Op: String);
Var
   Node: TmscrExpressionNode;
   n, n1: TmscrExpressionNode;

   //......................................
   Function CreateOperaTmsAction_Or(s: String): TmscrActionBase;
   Begin
      s := AnsiUpperCase(s);
      If s = ' ' Then
         Result := TmsAction_None.Create(msopNone)
      Else If s = '>' Then
         Result := TmsAction_Great.Create(msopGreat)
      Else If s = '<' Then
         Result := TmsAction_Less.Create(msopLess)
      Else If s = '<=' Then
         Result := TmsAction_LessEq.Create(msopLessEq)
      Else If s = '>=' Then
         Result := TmsAction_GreatEq.Create(msopGreatEq)
      Else If s = '<>' Then
         Result := TmsAction_NonEq.Create(msopNonEq)
      Else If s = '=' Then
         Result := TmsAction_Eq.Create(msopEq)
      Else If s = '+' Then
         Result := TmsAction_Plus.Create(msopPlus)
      Else If s = 'STRCAT' Then
         Result := TmsAction_StrCat.Create(msopPlus)
      Else If s = '-' Then
         Result := TmsAction_Minus.Create(msopMinus)
      Else If s = 'OR' Then
         Result := TmsAction_Or.Create(msopOr)
      Else If s = 'XOR' Then
         Result := TmsAction_Xor.Create(msopXor)
      Else If s = '*' Then
         Result := TmsAction_Mul.Create(msopMul)
      Else If s = '/' Then
         Result := TmsAction_DivFloat.Create(msopDivFloat)
      Else If s = 'DIV' Then
         Result := TmsAction_DivInt.Create(msopDivInt)
      Else If s = 'MOD' Then
         Result := TmsAction_Mod.Create(msopMod)
      Else If s = 'AND' Then
         Result := TmsAction_And.Create(msopAnd)
      Else If s = 'SHL' Then
         Result := TmsAction_Shl.Create(msopShl)
      Else If s = 'SHR' Then
         Result := TmsAction_Shr.Create(msopShr)
      Else If s = '(' Then
         Result := TmsAction_LeftBracket.Create(msopLeftBracket)
      Else If s = ')' Then
         Result := TmsAction_RightBracket.Create(msopRightBracket)
      Else If s = 'NOT' Then
         Result := TmsAction_Not.Create(msopNot)
      Else If s = 'UNMINUS' Then
         Result := TmsAction_UnMinus.Create(msopUnMinus)
      Else If s = 'IN' Then
         Result := TmsAction_In.Create(msopIn)
      Else If s = 'IS' Then
         Result := TmsAction_Is.Create(msopIs)
      Else
         Result := Nil;
   End;

   //...........................................
Begin
   Node := CreateOperaTmsAction_Or(Op);
   Node.SrcPosition := SrcPosition;

   If (Op = '(') Or (Op = 'unminus') Or (Op = 'not') Then
      AddOperand(Node)
   Else If Op = ')' Then
   Begin
      n := FCurNode;
      While n.Priority <= Node.Priority Do
         n := n.FParent;

      n.FParent.RemoveNode(n);
      n.FParent.AddNode(n.FLeft);

      Node.Free;
      Node := n.FLeft;
      n.FLeft := Nil;
      n.Free;
   End
   Else If FCurNode = FNode Then
      FNode.AddNode(Node)
   Else
   Begin
      n := FCurNode;
      n1 := Nil;
      If FCurNode.Priority <> 6 Then
      Begin
         n := FCurNode.FParent;
         n1 := FCurNode;
      End;

      While n.Priority <= Node.Priority Do
      Begin
         n1 := n;
         n := n.FParent;
      End;

      n.RemoveNode(n1);
      n.AddNode(Node);
      Node.AddNode(n1);
   End;

   FCurNode := Node;
End;

Procedure TmscrExpression.AddOperand(Node: TmscrExpressionNode);
Begin
   FCurNode.AddNode(Node);
   FCurNode := Node;
End;

Procedure TmscrExpression.Register_Const(Const AValue: variant);
Var
   xNode: TmscrActionNode;
Begin
   xNode := TmscrActionNode.Create(AValue);
   xNode.SrcPosition := SrcPosition;
   AddOperand(xNode);
End;

Procedure TmscrExpression.Register_ConstWithType(Const AValue: variant; aTyp: TmscrVarType);
Begin
   Register_Const(AValue);
   If aTyp = msvtClass Then
      FCurNode.Typ := msvtVariant;
End;

Procedure TmscrExpression.AddDScanner(ADScanner: TmscrDScanner);
Var
   Node: TmscrDScannerNode;
Begin
   Node := TmscrDScannerNode.Create(ADScanner);
   Node.SrcPosition := SrcPosition;
   AddOperand(Node);
End;

Procedure TmscrExpression.AddSet(ASet: TmscrSetExpression);
Var
   Node: TmscrSetNode;
Begin
   Node := TmscrSetNode.Create(ASet);
   Node.SrcPosition := SrcPosition;
   AddOperand(Node);
End;

Function TmscrExpression.Finalize: String;
Var
   ErrorPos: String;
   TypeRec: TmscrTypeRec;

   //...................................
   Function GetType(Item: TmscrExpressionNode): TmscrTypeRec;
   Var
      Typ1, Typ2: TmscrTypeRec;
      op: TmscrOperatorType;
      Error: boolean;
   Begin
      If Item = Nil Then
         Result.Typ := msvtVariant
      Else If Item Is TmscrActionNode Then
      Begin
         Result.Typ := Item.Typ;
         Result.TypeName := Item.TypeName;
      End
      Else
      Begin
         Typ1 := GetType(Item.FLeft);
         Typ2 := GetType(Item.FRight);

         If (Typ1.Typ = msvtBool) And (Typ2.Typ = msvtBool) Then
            TmscrActionBase(Item).FOptimizeBool := True;

         op := TmscrActionBase(Item).FOp;

         If (op = msopIs) And (Typ1.Typ = msvtClass) And (Typ2.Typ = msvtClass) Then
         Begin
            Error := False;
         End
         Else
         Begin

            Error := Not TypesCompatible(Typ1, Typ2, FScript);

            If Not Error Then
               Case Typ1.Typ Of
                  msvtBool:
                     Error := Not (op In [msopNonEq, msopEq, msopOr, msopXor, msopAnd, msopNot]);
                  msvtChar, msvtString:
                     Error := Not (op In [msopGreat, msopLess, msopLessEq, msopGreatEq, msopNonEq, msopEq, msopPlus, msopIn]);
                  msvtClass, msvtRecord, msvtArray:
                     Error := Not (op In [msopNonEq, msopEq]);
               End;
         End;

         If Not Error Then
         Begin
            Result := Typ1;

            If [Typ1.Typ] + [Typ2.Typ] = [msvtInt, msvtFloat] Then
               Result.Typ := msvtFloat;

            If (Typ1.Typ = msvtInt) And (Typ2.Typ = msvtInt) And (op = msopDivFloat) Then
               Result.Typ := msvtFloat;
            If [Typ1.Typ] + [Typ2.Typ] = [msvtInt64, msvtFloat] Then
               Result.Typ := msvtFloat;

            If ((Typ1.Typ = msvtInt64) Or (Typ1.Typ = msvtInt)) And ((Typ2.Typ = msvtInt64) Or (Typ2.Typ = msvtInt64)) And (op = msopDivFloat) Then
               Result.Typ := msvtFloat;

            If op In [msopGreat, msopLess, msopLessEq, msopGreatEq, msopNonEq, msopEq, msopIn, msopIs] Then
               Result.Typ := msvtBool;
         End
         Else If ErrorPos = '' Then
            ErrorPos := Item.SrcPosition;

         Item.Typ := Result.Typ;
      End;
   End;

   //.............................................
Begin

   FCurNode := FNode.FLeft;
   FNode.RemoveNode(FCurNode);
   FNode.Free;
   FNode := FCurNode;

   ErrorPos := '';
   TypeRec := GetType(FNode);
   Typ := TypeRec.Typ;
   TypeName := TypeRec.TypeName;
   Result := ErrorPos;

   If Not ((FNode Is TmscrDScannerNode) And Not (TmscrDScannerNode(FNode).FDScanner.IsReadOnly)) Then
      IsReadOnly := True;
End;

Procedure TmscrExpression.SetValue(Const Value: variant);
Begin
   If Not IsReadOnly Then
      TmscrDScannerNode(FNode).FDScanner.Value := Value;
End;

Function TmscrExpression.Optimize(DScanner: TmscrDScanner): String;
Var
   Op: TmscrOperatorType;
Begin
   Result := ' ';

   If Not (DScanner Is TmscrVarForVariableDScanner) Or Not (FNode Is TmscrActionBase) Then
      Exit;

   Op := TmscrActionBase(FNode).FOp;
   If Not (Op In [msopPlus, msopMinus, msopDivFloat, msopMul]) Then
      Exit;

   If (FNode.FLeft Is TmscrDScannerNode) And (TmscrDScannerNode(FNode.FLeft).FDScanner Is TmscrVarForVariableDScanner) And (TmscrDScannerNode(FNode.FLeft).FDScanner.RefItem = DScanner.RefItem) Then
   Begin
      FCurNode := FNode.FRight;
      FNode.RemoveNode(FCurNode);
      FNode.Free;
      FNode := FCurNode;

      If Op = msopPlus Then
         Result := '+'
      Else If Op = msopMinus Then
         Result := '-'
      Else If Op = msopDivFloat Then
         Result := '/'
      Else If Op = msopMul Then
         Result := '*';
   End

   Else If (FNode.FRight Is TmscrDScannerNode) And (TmscrDScannerNode(FNode.FRight).FDScanner Is TmscrVarForVariableDScanner) And (TmscrDScannerNode(FNode.FRight).FDScanner.RefItem = DScanner.RefItem) And
      (Op In [msopPlus, msopMul]) And Not (DScanner.RefItem.Typ In [msvtString, msvtVariant]) Then
   Begin
      FCurNode := FNode.FLeft;
      FNode.RemoveNode(FCurNode);
      FNode.Free;
      FNode := FCurNode;

      If Op = msopPlus Then
         Result := '+'
      Else If Op = msopMul Then
         Result := '*';
   End;
End;

Function TmscrExpression.SingleItem: TmscrVarForBase;
Begin
   Result := Nil;

   If FNode Is TmscrDScannerNode Then
   Begin
      If TmscrDScannerNode(FNode).FDScanner Is TmscrVarForVariableDScanner Then
         Result := TmscrDScannerNode(FNode).FDScanner.RefItem
      Else
         Result := TmscrDScannerNode(FNode).FDScanner;
   End
   Else If FNode Is TmscrActionNode Then
      Result := FNode;
End;

//=============== TmscrEngParser =================================

Constructor TmscrEngParser.Create(aProject: TmscrScript);
Begin
   FNeedDeclareVars := True;
   FScrProject := aProject;
   FGrammar := TmscrDATADocument.Create;
   FFinalScript := TmscrDATADocument.Create;
   FParser := TmscrParser.Create;
   FUsesList := TStringList.Create;
   FWithList := TStringList.Create;
End;

Destructor TmscrEngParser.Destroy;
Begin
   FGrammar.Free;
   FFinalScript.Free;
   FParser.Free;
   FUsesList.Free;
   FWithList.Free;
   Inherited;
End;

Procedure TmscrEngParser.InitGrammar;
Var
   i: integer;
   ss: TStringStream;
Begin
   FParser.Clear;
   //.....................................
   ss := TStringStream.Create(_MAINGRAMMAR);
   Try
      FGrammar.LoadFromStream(ss);
   Finally
      ss.Free;
   End;
   //.....................................

   FRoot := FGrammar.Root;

   FParser.InitKeywords;

   If FScrProject.UseExtendCharset Then
      For i := 128 To 255 Do
         FParser.IdentifierCharset := FParser.IdentifierCharset + [Chr(i)];
End;

Function TmscrEngParser.FindError(Const aErr: integer): String;
Begin
   Result := msEUnError;
   Case aErr Of
      1: Result := msError01;
      2: Result := msError02;
      3: Result := msError03;
      4: Result := msError04;
      5: Result := msError05;
      6: Result := msError06;
      7: Result := msError07;
      8: Result := msError08;
      9: Result := msError09;
      10: Result := msError10;
      11: Result := msError11;
      12: Result := msError12;
      13: Result := msError13;
      14: Result := msError14;
      15: Result := msError15;
      16: Result := msError16;
      17: Result := msError17;
      18: Result := msError18;
      19: Result := msError19;
      20: Result := msError20;
   End;
End;

Function TmscrEngParser.FinalScript_Build(Const Text: String): boolean;
Var
   FList: TStrings;
   FStream: TStream;
   FErrorMsg: String;
   FErrorPos: String;
   FTermError: boolean;
   i: integer;

   //.............................................
   Function _Run(xi: TmscrDATAItem): boolean;
   Var
      i, j, ParsPos, ParsPos1, LoopPos, ListPos: integer;
      s, NodeName, Token, PropText, PropAdd, PropAddText, PropNode: String;
      Completed, TopLevelNode, Flag: boolean;
   Const
      PathD =
   {$IFDEF MSWINDOWS}
      '\'
   {$ELSE}
      '/'
   {$ENDIF}
      ;

      //.............................................
      Procedure __DoInclude(Const Name: String);
      Var
         sl: TStringList;
         p: TmscrEngParser;
         ss: TStringStream;
         s, UnitPath: String;
         idx: integer;
      Begin
         If FUsesList.IndexOf(Name) <> -1 Then
            Exit;
         FUsesList.Add(Name);
         sl := TStringList.Create;
         Try
            If Assigned(FScrProject.OnGetUnit) Then
            Begin
               s := '';
               FScrProject.OnGetUnit(FScrProject, Name, s);
               sl.Text := s;
            End
            Else
            Begin
               UnitPath := '';
               For idx := 0 To FScrProject.IncludePath.Count - 1 Do
               Begin
                  UnitPath := FScrProject.IncludePath[idx];
                  If (UnitPath <> '') And (PathD <> UnitPath[Length(UnitPath)]) Then
                     UnitPath := UnitPath + PathD;
                  If FileExists(UnitPath + Name) Then
                     break;
               End;
               sl.LoadFromFile(UnitPath + Name);
            End;

            p := TmscrEngParser.Create(FScrProject);
            p.FUnitName := Name;
            ss := TStringStream.Create('');
            Try
               s := '';
               If sl.Count > 0 Then
               Begin
                  p.InitGrammar;
                  p.FUsesList.Assign(FUsesList);
                  If p.FinalScript_Build(sl.Text) Then
                  Begin
                     FUsesList.Assign(p.FUsesList);
                     p.FinalScript.SaveToStream(ss);
                     s := ss.DataString;
                     Delete(s, 1, Pos('?>', s) + 1);
                  End
                  Else
                  Begin
                     FErrorMsg := FScrProject.ErrorMsg;
                     FErrorPos := FScrProject.ErrorPos;
                     If FScrProject.ErrorUnit = '' Then
                        FScrProject.ErrorUnit := Name;
                  End;
               End;

               FList.Insert(ListPos, '</uses>');
               FList.Insert(ListPos, s);
               FList.Insert(ListPos, '<uses' + ' unit="' + Name + '">');
               Inc(ListPos, 3);
            Finally
               p.Free;
               ss.Free;
            End;
         Finally
            sl.Free;
         End;
      End;

      //.............................................
      Procedure __CheckPropNode(Flag: boolean);
      Var
         i, ParsPos1: integer;
         s: String;
      Begin

         If CompareText(PropNode, 'uses') = 0 Then
         Begin
            While FList.Count > ListPos Do
            Begin
               s := FList[FList.Count - 1];
               i := Pos('text="', s);
               Delete(s, 1, i + 5);
               i := Pos('" ', s);
               Delete(s, i, 255);
               __DoInclude(Copy(s, 2, Length(s) - 2));
               FList.Delete(FList.Count - 1);
            End;
         End
         Else If PropNode <> '' Then
            If Flag Then
            Begin
               ParsPos1 := FParser.Position;
               FParser.Position := ParsPos;
               FParser.SkipSpaces;

               s := '<' + PropNode + ' pos="' + FParser.GetXYPosition + '"';
               FParser.Position := ParsPos1;

               If PropNode = 'expr' Then
                  s := s + ' pos1="' + FParser.GetXYPosition + '"';
               s := s + '>';

               FList.Insert(ListPos, s);
               FList.Add('</' + PropNode + '>');
            End
            Else
            Begin
               While FList.Count > ListPos Do
                  FList.Delete(FList.Count - 1);
            End;
      End;

      //.............................................
      Procedure __AddError(xi: TmscrDATAItem);
      Var
         propErr: integer;
         ss: String;
      Begin
         ss := xi.Prop['errornum'];
         If ss = '' Then exit;

         propErr := StrToInt(ss);
         If (PropErr > 0) And (FErrorMsg = '') Then
         Begin
            FErrorMsg := FindError(PropErr);
            FParser.Position := ParsPos;
            FParser.SkipSpaces;
            FErrorPos := FParser.GetXYPosition;
            FTermError := xi.Prop['term'] = '1';
         End;
      End;

      //.............................................
   Begin
      Result := True;
      ParsPos := FParser.Position;
      ListPos := FList.Count;
      NodeName := AnsiLowerCase(xi.Name);
      propText := AnsiLowerCase(xi.Prop['text']);
      propNode := LowerCase(xi.Prop['node']);
      TopLevelNode := xi.Parent = FRoot;

      Completed := False;
      Flag := False;
      Token := '';

      If TopLevelNode Then
         Completed := True
      Else If NodeName = 'char' Then
      Begin
         If xi.Prop['skip'] <> '0' Then
            FParser.SkipSpaces;
         Token := FParser.GetChar;
         Flag := True;
      End
      Else If NodeName = 'keyword' Then
      Begin
         Token := FParser.GetWord;
         Flag := True;
      End
      Else If NodeName = 'ident' Then
      Begin
         Token := FParser.GetIdent;
         Flag := True;
      End
      Else If NodeName = 'number' Then
      Begin
         Token := FParser.GetNumber;
         Flag := True;
      End
      Else If NodeName = 'string' Then
      Begin
         Token := FParser.GetString;
         Flag := True;
      End
      Else If NodeName = 'frstring' Then
      Begin
         Token := FParser.GetFRString;
         s := FParser.GetXYPosition;
         FList.Add('<dsgn pos="' + s + '">');
         FList.Add('<node text="Get" pos="' + s + '"/>');
         FList.Add('<expr pos="' + s + '">');
         FList.Add('<string text="''' + msStrToXML(Token) + '''" pos="' + s + '"/>');
         FList.Add('</expr>');
         FList.Add('</dsgn>');
         Flag := True;
      End
      Else If NodeName = 'eol' Then
         Completed := FParser.GetEOL
      Else If NodeName = 'dojob' Then
         Completed := True
      Else If (NodeName = 'switch') Or (NodeName = 'optswitch') Then
      Begin
         Completed := True;

         For i := 0 To xi.Count - 1 Do
         Begin
            Completed := _Run(xi[i]);

            If Completed Then BREAK;
         End;

         If Not Completed Then
            If NodeName <> 'optswitch' Then
            Begin
               Result := False;
               __AddError(xi);
            End;

         EXIT;

      End
      Else If (NodeName = 'loop') Or (NodeName = 'optloop') Then
      Begin
         j := 0;
         Repeat
            Inc(j);
            Flag := False;
            LoopPos := FParser.Position;

            For i := 0 To xi.Count - 1 Do
            Begin
               Result := _Run(xi[i]);
               If Not Result Then
               Begin
                  Flag := True;
                  break;
               End;
            End;

            ParsPos1 := FParser.Position;
            If Result And (PropText <> '') Then
            Begin
               FParser.SkipSpaces;
               If FParser.GetChar <> PropText Then
               Begin
                  FParser.Position := ParsPos1;
                  Flag := True;
               End;
            End;

            If FParser.Position = LoopPos Then Flag := True;

         Until Flag;

         If j > 1 Then
         Begin
            If (xi.Prop['skip'] = '1') Or FTermError Then FErrorMsg := '';
            FParser.Position := ParsPos1;
            Result := True;
         End;

         If NodeName = 'optloop' Then
         Begin
            If Not Result Then FParser.Position := ParsPos;
            Result := True;
         End;
         Exit;

      End
      Else If NodeName = 'optional' Then
      Begin
         For i := 0 To xi.Count - 1 Do
            If Not _Run(xi[i]) Then
            Begin
               FParser.Position := ParsPos;
               break;
            End;
         Exit;
      End
      Else
      Begin
         j := FRoot.Find(NodeName);
         If j = -1 Then Raise Exception.Create(msEInvalidLanguage);
         Completed := _Run(FRoot[j]);
      End;

      If Flag Then
      Begin
         If FParser.CaseSensitive Then
            Completed := (Token <> '') And ((PropText = '') Or (Token = PropText))
         Else
            Completed := (Token <> '') And ((PropText = '') Or (AnsiCompareText(Token, PropText) = 0));
      End;

      If Not Completed Then
      Begin
         Result := False;
         __AddError(xi);
      End
      Else
      Begin
         If Not TopLevelNode Then __CheckPropNode(True);

         propAdd := xi.Prop['add'];
         propAddText := xi.Prop['addtext'];

         If PropAdd <> '' Then
         Begin
            If PropAddText = '' Then
               s := Token
            Else
               s := PropAddText;

            FList.Add('<' + PropAdd + ' text="' + msStrToXML(s) + '" pos="' + FParser.GetXYPosition + '"/>');
         End;

         For i := 0 To xi.Count - 1 Do
         Begin
            Result := _Run(xi[i]);
            If Not Result Then
               break;
         End;
      End;

      If Not Result Then FParser.Position := ParsPos;
      If TopLevelNode Then __CheckPropNode(Result);
   End;

   //...............................................
Begin
   FList := TStringList.Create;
   FErrorMsg := '';
   FErrorPos := '';
   Result := False;

   Try
      FParser.Text := Text;

      i := 1;
      If FParser.GetChar = '#' Then
      Begin
         If CompareText(FParser.GetIdent, 'language') = 0 Then
         Begin
            i := FParser.Position;

{$IFDEF MSWINDOWS}
            While (i <= Length(Text)) And (Text[i] <> #13) Do
{$ELSE}
            While (i <= Length(Text)) And (Text[i] <> #10) Do
{$ENDIF}
               Inc(i);
            InitGrammar;
            Inc(i, 2);
         End;
      End;
      FParser.Position := i;

      If _Run(FRoot.FindItem('program')) And (FErrorMsg = '') Then
      Begin
         FErrorMsg := '';
         FErrorPos := '';
         FStream := TMemoryStream.Create;
         Try
            FList.Insert(0, '<?xml version="1.0"?>');
            FList.Insert(1, '<program>');
            FList.Add('</program>');
            FList.SaveToStream(FStream);
            FStream.Position := 0;
            FFinalScript.LoadFromStream(FStream);
            FFinalScript.Root.Add.Assign(FRoot.FindItem('types'));
            Result := True;
         Finally
            FStream.Free;
         End;

      End;

      FScrProject.ErrorPos := FErrorPos;
      FScrProject.ErrorMsg := FErrorMsg;
   Finally
      FList.Free;
   End;
End;

Function TmscrEngParser.PropPos(xi: TmscrDATAItem): String;
Begin
   Result := xi.Prop['pos'];
End;

Procedure TmscrEngParser.ErrorPos(xi: TmscrDATAItem);
Begin
   FErrorPos := PropPos(xi);
End;

Procedure TmscrEngParser.FinalScript_Parse;
Begin
   FWithList.Clear;
   FScrProject.ErrorUnit := '';
   FUnitName := '';
   FUsesList.Clear;
   Try
      Compile_Project(FFinalScript.Root, FScrProject);
      FScrProject.ErrorPos := '';
   Except
      On e: Exception Do
      Begin
         FScrProject.ErrorMsg := e.Message;
         FScrProject.ErrorPos := FErrorPos;
         FScrProject.ErrorUnit := FUnitName;
      End;
   End;
End;

Procedure TmscrEngParser.CheckIdent(aProj: TmscrScript; Const Name: String);
Begin
   If aProj.FindLocal(Name) <> Nil Then
      Raise Exception.Create(msEIdRedeclared + '''' + Name + '''');
End;

Function TmscrEngParser.FindClass(Const TypeName: String): TmscrVarForClass;
Begin
   Result := FScrProject.FindClass(TypeName);
   If Result = Nil Then
      Raise Exception.Create(msEUnknownType + '''' + TypeName + '''');
End;

Function TmscrEngParser.FindRecord(Const TypeName: String): TmscrVarForRecord;
Begin
   Result := FScrProject.FindRecord(TypeName);
   If Result = Nil Then
      Raise Exception.Create(msEUnknownType + '''' + TypeName + '''');
End;

Procedure TmscrEngParser.CheckTypeCompatibility(Var1, Var2: TmscrVarForBase);
Begin
   If Not AssignCompatible(Var1, Var2, FScrProject) Then
      Raise Exception.Create(msEIncompatibleTypes + ': ''' + Var1.GetFullTypeName + ''', ''' + Var2.GetFullTypeName + '''');
End;

Function TmscrEngParser.FindType(s: String): TmscrVarType;
Var
   xi: TmscrDATAItem;
Begin
   xi := FProjRoot.FindItem('types');

   If xi.Find(s) <> -1 Then
   Begin
      s := xi[xi.Find(s)].Prop['type'];
   End
   Else
   Begin
      xi := FGrammar.Root.FindItem('types');
      If xi.Find(s) <> -1 Then
         s := xi[xi.Find(s)].Prop['type'];
   End;

   Result := StrToVarType(s, FScrProject);

   If Result = msvtClass Then FindClass(s);
   If Result = msvtRecord Then FindRecord(s);
End;

Function TmscrEngParser.FindVar(aProj: TmscrScript; Const Name: String): TmscrVarForBase;
Begin
   Result := aProj.Find(Name);
   If Result = Nil Then
      If Not FNeedDeclareVars Then
      Begin
         Result := TmscrVarForVariable.Create(Name, msvtVariant, '');
         FScrProject.Add(Name, Result);
      End
      Else
      Begin
         Raise Exception.Create(msEUndeclaredId + '''' + Name + '''');
      End;
End;

// ct9999 zzzz

Function TmscrEngParser.CreateVar(xi: TmscrDATAItem; aProj: TmscrScript; Const Name: String; Statement: TmscrStatement = Nil; CreateParam: boolean = False; IsVarParam: boolean = False): TmscrVarForBase;
Var
   i, j: integer;
   Typ: TmscrVarType;
   TypeName: String;
   RefItem: TmscrVarForBase;
   InitValue: variant;
   InitItem: TmscrDATAItem;
   CodeAssign: TmscrCodeStmAssign_Base;
   IsPascal: boolean;
   SrcPosition: String;
   vRecA, vRecB: TmscrVarForRecord;
   vRecElA, vRecElB: TmscrEngRecElement;

   //..............................................
   Procedure _DoArray(xi: TmscrDATAItem);
   Var
      i, n: integer;
      v: Array Of SizeInt;
      Expr: TmscrExpression;
   Begin
      n := xi.Count;
      SetLength(v, n * 2);

      For i := 0 To n - 1 Do
      Begin
         Expr := Compile_Expression(xi[i][0], aProj);
         v[i * 2] := Expr.Value;
         Expr.Free;

         If xi[i].Count = 2 Then
         Begin
            Expr := Compile_Expression(xi[i][1], aProj);
            v[i * 2 + 1] := Expr.Value;
            Expr.Free;
         End
         Else
         Begin
            v[i * 2 + 1] := v[i * 2] - 1;
            v[i * 2] := 0;
         End;
      End;

      If n = 0 Then
      Begin
         SetLength(v, 2);
         v[0] := 0;
         v[1] := 0;
         n := 1;
      End;

      InitValue := VarArrayCreate(v, varVariant);
      RefItem := TmscrEngArray.Create('', n, Typ, TypeName);
      aProj.Add('', RefItem);
      v := Nil;
      Typ := msvtArray;
   End;

   //..............................................
   Procedure _DoInit(xi: TmscrDATAItem);
   Var
      Expr: TmscrExpression;
      Temp: TmscrVarForVariable;
   Begin
      Temp := TmscrVarForVariable.Create('', Typ, TypeName);
      Try
         Expr := Compile_Expression(xi[0], aProj);
         InitValue := Expr.Value;
         Try
            CheckTypeCompatibility(Temp, Expr);
         Finally
            Expr.Free;
         End;
      Finally
         Temp.Free;
      End;
   End;

   //..............................................
Begin
   RefItem := Nil;
   InitItem := Nil;
   TypeName := 'Variant';
   IsPascal := False;
   SrcPosition := FErrorPos;

   For i := 0 To xi.Count - 1 Do
      If CompareText(xi[i].Name, 'type') = 0 Then
      Begin
         IsPascal := i <> 0;
         TypeName := xi[i].Prop['text'];
         ErrorPos(xi[i]);
         break;
      End;

   Typ := FindType(TypeName);
   Case Typ Of
      msvtInt, msvtInt64, msvtFloat, msvtClass:
         InitValue := 0;
      msvtRecord:
         InitValue := Nil;
      msvtBool:
         InitValue := False;
      msvtChar, msvtString:
         InitValue := '';
      Else
         InitValue := Null;
   End;

   For i := 0 To xi.Count - 1 Do
      If CompareText(xi[i].Prop['text'], Name) = 0 Then
      Begin

         j := i + 1;
         While (j < xi.Count) And (IsPascal Or (CompareText(xi[j].Name, 'ident') <> 0)) Do
         Begin
            If CompareText(xi[j].Name, 'array') = 0 Then
               _DoArray(xi[j])
            Else If CompareText(xi[j].Name, 'init') = 0 Then
            Begin
               If Statement = Nil Then
                  _DoInit(xi[j]);
               InitItem := xi[j];
            End;
            Inc(j);
         End;
         break;
      End;

   If CreateParam Then
      Result := TmscrVarForParamItem.Create(Name, Typ, TypeName, InitItem <> Nil, IsVarParam)
   Else If Typ In [msvtRecord] Then
   Begin
      vRecA := FindRecord(TypeName);

      If vRecA <> Nil Then
      Begin
         Result := TmscrVarForRecord.Create(vRecA.TypeInfo, TypeName);
         vRecB := TmscrVarForRecord(Result);

         If vRecB.MembersCount < vRecA.MembersCount Then
            For i := 0 To vRecA.GetMembersCount - 1 Do
            Begin
               vRecElA := TmscrEngRecElement(vRecA.Members[i]);
               vRecElB := TmscrEngRecElement.Create(vRecElA.Name, vRecElA.Typ, vRecElA.TypeName);
               vRecElB.ParentRef := Result;
               vRecElB.ParentValue := Result.Value;
               vRecElB.FOnGetValue := vRecElA.FOnGetValue;
               vRecElB.FOnSetValue := vRecElA.FOnSetValue;
               vRecElB.IsReadOnly := vRecElA.IsReadOnly;

               vRecB.FMembers.Add(vRecElB);
            End;

      End;

   End
   Else If Typ In [msvtChar, msvtString] Then
      Result := TmscrVarForString.Create(Name, Typ, TypeName)
   Else
      Result := TmscrVarForVariable.Create(Name, Typ, TypeName);

   Try
      If Typ <> msvtRecord Then
         Result.Value := InitValue;
      Result.RefItem := RefItem;
      Result.SrcPosition := SrcPosition;
      Result.SourceUnit := FUnitName;
      Result.OnGetVarValue := FScrProject.OnGetVarValue;

      If (InitItem <> Nil) And (Statement <> Nil) Then
      Begin
         CodeAssign := TmscrCodeStmAssign_Base.Create(aProj, FUnitName, PropPos(xi));
         Statement.Add(CodeAssign);
         CodeAssign.DScanner := TmscrVarForVariableDScanner.Create(aProj);
         CodeAssign.DScanner.RefItem := Result;
         CodeAssign.Expression := Compile_Expression(InitItem[0], aProj);
         CheckTypeCompatibility(Result, CodeAssign.Expression);
         CodeAssign.Optimize;
      End;

   Except
      On e: Exception Do
      Begin
         Result.Free;
         Raise;
      End;
   End;
End;

{$HINTS OFF}

Function TmscrEngParser.DoDScanner(xi: TmscrDATAItem; aProj: TmscrScript; EmitOp: TmscrEmitOp = mseoNone): TmscrDScanner;
Var
   i, j: integer;
   NodeName, NodeText, TypeName: String;
   Expr: TmscrExpression;
   Item, PriorItem: TmscrDScannerItem;
   ClassVar: TmscrVarForClass;
   RecordVar: TmscrVarForRecord;
   StringVar: TmscrVarForString;
   Typ: TmscrVarType;
   LateBinding, PriorIsIndex: boolean;
   NewDScanner: TmscrDScanner;
   PriorValue: variant;
   Component: TComponent;

   //.....................................
   Function FindInWithList(Const Name: String; ResultDS: TmscrDScanner; Item: TmscrDScannerItem): boolean;
   Var
      i: integer;
      Withsrcstm: TmscrCodeStm_With;
      WithItem: TmscrDScannerItem;
      ClassVar: TmscrVarForClass;
      xi1: TmscrDATAItem;
   Begin
      Result := False;
      LateBinding := False;
      For i := FWithList.Count - 1 Downto 0 Do
      Begin

         If aProj.FindLocal(FWithList[i]) = Nil Then
            continue;
         Withsrcstm := TmscrCodeStm_With(FWithList.Objects[i]);

         If Withsrcstm.Variable.Typ = msvtVariant Then
         Begin
            If aProj.Find(Name) <> Nil Then Exit;

            Item.Ref := Withsrcstm.Variable;
            ResultDS.Finalize;
            ResultDS.LateBindingDATASource := TmscrDATAItem.Create;
            ResultDS.LateBindingDATASource.Assign(xi);
            xi1 := TmscrDATAItem.Create;
            xi1.Name := 'node';
            xi1.Text := 'text="' + FWithList[i] + '"';
            ResultDS.LateBindingDATASource.InsertItem(0, xi1);
            LateBinding := True;
            Result := True;
            break;
         End
         Else
         Begin
            ClassVar := FindClass(Withsrcstm.Variable.TypeName);
            Item.Ref := ClassVar.Find(NodeText);
         End;

         If Item.Ref <> Nil Then
         Begin
            WithItem := TmscrDScannerItem.Create;
            WithItem.Ref := Withsrcstm.Variable;
            WithItem.SrcPosition := Item.SrcPosition;

            ResultDS.Remove(Item);
            ResultDS.Add(WithItem);
            ResultDS.Add(Item);
            Result := True;
            break;
         End;
      End;
   End;

   //.....................................
{$IFDEF MAGSCR_USEOLE}
   Procedure CreateOLEHelpers(Index: integer);
   Var
      i: integer;
      OLEHelper: TmscrEngOLE;
   Begin
      For i := Index To xi.Count - 1 Do
      Begin
         ErrorPos(xi[i]);
         NodeName := LowerCase(xi[i].Name);
         NodeText := xi[i].Prop['text'];

         If (NodeName = 'node') And (NodeText <> '[') Then
         Begin
            Item := TmscrDScannerItem.Create;
            Result.Add(Item);
            Item.SrcPosition := FErrorPos;
            OLEHelper := TmscrEngOLE.Create(NodeText);
            aProj.Add('', OLEHelper);
            Item.Ref := OLEHelper;
         End
         Else If NodeName = 'expr' Then
         Begin
            Expr := Compile_Expression(xi[i], aProj);
            PriorItem := Result.Items[Result.Count - 1];
            PriorItem.Add(Expr);
            PriorItem.Ref.Add(TmscrVarForParamItem.Create('', msvtVariant, '', False, False));
         End;
      End;
   End;

{$ENDIF}
   //.....................................
Begin
   Result := TmscrDScanner.Create(aProj);
   Try

      For i := 0 To xi.Count - 1 Do
      Begin
         ErrorPos(xi[i]);
         NodeName := LowerCase(xi[i].Name);
         NodeText := xi[i].Prop['text'];

         If NodeName = 'node' Then
         Begin
            Item := TmscrDScannerItem.Create;
            Result.Add(Item);
            Item.SrcPosition := FErrorPos;

            If Result.Count = 1 Then
            Begin
               If Not FindInWithList(NodeText, Result, Item) Then
                  Item.Ref := FindVar(aProj, NodeText);

               If LateBinding Then Exit;

               If EmitOp = mseoCreate Then
               Begin
                  If Not (Item.Ref Is TmscrVarForClass) Then
                     Raise Exception.Create(msEClassRequired);
                  ClassVar := TmscrVarForClass(Item.Ref);
                  Item := TmscrDScannerItem.Create;
                  Result.Add(Item);
                  Item.Ref := ClassVar.Find('Create');
               End;
            End
            Else
            Begin
               PriorItem := Result.Items[Result.Count - 2];
               PriorIsIndex := (PriorItem.Ref Is TmscrVarForMethod) And TmscrVarForMethod(PriorItem.Ref).IndexMethod And Not PriorItem.Flag;
               Typ := PriorItem.Ref.Typ;

               //---------- msvtVariant --------------
               If (Typ = msvtVariant) And Not PriorIsIndex Then
               Begin
                  PriorValue := PriorItem.Ref.Value;
                  If VarIsNull(PriorValue) Then
                  Begin
                     Result.Remove(Item);
                     Item.Free;
                     Result.Finalize;
                     Result.LateBindingDATASource := TmscrDATAItem.Create;
                     Result.LateBindingDATASource.Assign(xi);
                     Exit;
                  End
                  Else
                  Begin
                     If (TVarData(PriorValue).VType = varString) Then
                        Typ := msvtString
{$IFDEF MAGSCR_USEOLE}
                     Else If TVarData(PriorValue).VType = varDispatch Then
                     Begin
                        Result.Remove(Item);
                        Item.Free;
                        CreateOLEHelpers(i);
                        Result.Finalize;
                        Exit;
                     End
{$ENDIF}
                     Else If (TVarData(PriorValue).VType And varArray) = varArray Then
                     Begin
                        If NodeText = '[' Then
                           Item.Ref := FindVar(aProj, '__HelperForArray')
                        Else
                           Raise Exception.Create(msEIndexRequired);
                        continue;
                     End
                     Else If (TVarData(PriorValue).VType And varRecord) = varRecord Then
                     Begin
                        Typ := msvtRecord;
                        PriorItem.Ref := FindRecord(PriorItem.Ref.Value);
                        continue;
                     End
                     Else
                     Begin
                        Typ := msvtClass;
                        PriorItem.Ref.TypeName := TObject(fmscrInteger(PriorItem.Ref.Value)).ClassName;
                     End;
                  End;
               End;

               If PriorIsIndex Then
               Begin
                  PriorItem.Flag := True;
                  Result.Remove(Item);
                  Item.Free;
                  FErrorPos := PriorItem.SrcPosition;
                  If NodeText <> '[' Then
                     Raise Exception.Create(msEIndexRequired);
               End
               Else If Typ = msvtString Then
               Begin
                  If NodeText = '[' Then
                     Item.Ref := FindVar(aProj, '__HelperForString')
                  Else
                     Raise Exception.Create(msEStringError);
               End
               Else If Typ = msvtClass Then
               Begin
                  TypeName := PriorItem.Ref.TypeName;
                  ClassVar := FindClass(TypeName);

                  If NodeText = '[' Then
                  Begin
                     Item.Flag := True;
                     Item.Ref := ClassVar.DefProperty;
                     If Item.Ref = Nil Then
                        Raise Exception.CreateFmt(msEClassError, [TypeName]);
                  End
                  Else
                  Begin
                     Item.Ref := ClassVar.Find(NodeText);

                     If Item.Ref = Nil Then
                     Begin
                        PriorValue := PriorItem.Ref.Value;
                        If ((VarIsNull(PriorValue) Or (PriorValue = 0)) And Not aProj.IsRunning) And aProj.UseClassLateBinding Then
                        Begin
                           Result.Remove(Item);
                           Item.Free;
                           While Result.Count > 1 Do
                           Begin
                              Item := Result.Items[Result.Count - 1];
                              Result.Remove(Item);
                              Item.Free;
                           End;
                           Item := Result.Items[0];
                           Result.Finalize;
                           Result.Typ := msvtVariant;
                           Result.LateBindingDATASource := TmscrDATAItem.Create;
                           Result.LateBindingDATASource.Assign(xi);
                           Exit;
                        End
                        Else
                        Begin

                           If TObject(fmscrInteger(PriorValue)) Is TComponent Then
                           Begin
                              Component := TComponent(fmscrInteger(PriorValue)).FindComponent(NodeText);
                              If Component <> Nil Then
                              Begin
                                 Item.Ref := TmscrVarForBase.Create('', msvtClass, Component.ClassName);
                                 Item.Ref.Value := fmscrInteger(Component);
                              End;
                           End;
                           If Item.Ref = Nil Then
                              Raise Exception.Create(msEUndeclaredId + '''' + NodeText + '''');
                        End;
                     End;
                  End;
               End
               Else If Typ = msvtRecord Then
               Begin
                  TypeName := PriorItem.Ref.TypeName;
                  RecordVar := FindRecord(TypeName);

                  If NodeText = '[' Then
                  Begin
                     Item.Flag := True;
                     Item.Ref := RecordVar.DefProperty;
                     If Item.Ref = Nil Then
                        Raise Exception.CreateFmt(msEClassError, [TypeName]);
                  End
                  Else
                  Begin
                     Item.Ref := RecordVar.Find(NodeText);

                     If Item.Ref = Nil Then
                     Begin
                        PriorValue := PriorItem.Ref.Value;
                        If ((VarIsNull(PriorValue) Or (PriorValue = 0)) And Not aProj.IsRunning) And aProj.UseClassLateBinding Then
                        Begin
                           Result.Remove(Item);
                           Item.Free;
                           While Result.Count > 1 Do
                           Begin
                              Item := Result.Items[Result.Count - 1];
                              Result.Remove(Item);
                              Item.Free;
                           End;
                           Item := Result.Items[0];
                           Result.Finalize;
                           Result.Typ := msvtVariant;
                           Result.LateBindingDATASource := TmscrDATAItem.Create;
                           Result.LateBindingDATASource.Assign(xi);
                           Exit;
                        End
                        Else
                        Begin

                           If TObject(fmscrInteger(PriorValue)) Is TComponent Then
                           Begin
                              Component := TComponent(fmscrInteger(PriorValue)).FindComponent(NodeText);
                              If Component <> Nil Then
                              Begin
                                 Item.Ref := TmscrVarForBase.Create('', msvtRecord, Component.ClassName);
                                 Item.Ref.Value := fmscrInteger(Component);
                              End;
                           End;
                           If Item.Ref = Nil Then
                              Raise Exception.Create(msEUndeclaredId + '''' + NodeText + '''');
                        End;
                     End;
                  End;
               End
               Else If Typ = msvtArray Then
                  Item.Ref := PriorItem.Ref.RefItem
               Else
                  //---------- Nothing from Above -------
                  Raise Exception.Create(msEArrayRequired);
            End;
         End
         Else If NodeName = 'expr' Then
         Begin
            Expr := Compile_Expression(xi[i], aProj);
            Result.Items[Result.Count - 1].Add(Expr);
         End
         Else If NodeName = 'addr' Then
         Begin
            If xi.Count <> 2 Then
               Raise Exception.Create(msEVarRequired);

            Item := TmscrDScannerItem.Create;
            Result.Add(Item);
            ErrorPos(xi[1]);
            Item.SrcPosition := FErrorPos;

            StringVar := TmscrVarForString.Create('', msvtString, '');
            StringVar.Value := xi[1].Prop['text'];
            aProj.Add('', StringVar);
            Item.Ref := StringVar;

            break;
         End;
      End;

      If EmitOp = mseoFree Then
      Begin
         PriorItem := Result.Items[Result.Count - 1];
         If (PriorItem.Ref.Typ <> msvtClass) And (PriorItem.Ref.Typ <> msvtVariant) Then
            Raise Exception.Create(msEClassRequired);
         Item := TmscrDScannerItem.Create;
         Result.Add(Item);
         ClassVar := FindClass('TObject');
         Item.Ref := ClassVar.Find('Free');
      End;

      Result.Finalize;
      If Result.Kind <> msskOther Then
      Begin
         NewDScanner := Nil;
         If Result.Kind = msskVariable Then
            NewDScanner := TmscrVarForVariableDScanner.Create(aProj)
         Else If Result.Kind = msskStringArray Then
            NewDScanner := TmscrStringDScanner.Create(aProj)
         Else If Result.Kind = msskArray Then
            NewDScanner := TmscrArrayDScanner.Create(aProj);

         NewDScanner.Borrow(Result);
         Result.Free;
         Result := NewDScanner;
      End;

      For i := 0 To Result.Count - 1 Do
      Begin
         Item := Result[i];
         FErrorPos := Item.SrcPosition;
         If Item.Ref Is TmscrDScanner Then
            continue;

         If Item.Count < Item.Ref.GetNumberOfRequiredParams Then
            Raise Exception.Create(msENotEnoughParams)
         Else If Item.Count > Item.Ref.Count Then
            Raise Exception.Create(msETooManyParams)
         Else If Item.Count <> Item.Ref.Count Then
            For j := Item.Count To Item.Ref.Count - 1 Do
            Begin
               Expr := TmscrExpression.Create(FScrProject);
               Item.Add(Expr);
               Expr.Register_ConstWithType(Item.Ref[j].DefValue, Item.Ref[j].Typ);
               Expr.Finalize;
            End;

         For j := 0 To Item.Count - 1 Do
         Begin
            FErrorPos := Item[j].SrcPosition;
            CheckTypeCompatibility(Item.Ref[j], Item[j]);
         End;
      End;

   Except
      On e: Exception Do
      Begin
         Result.Free;
         Raise;
      End;
   End;
End;

{$HINTS ON}

Function TmscrEngParser.Compile_Set(xi: TmscrDATAItem; aProj: TmscrScript): TmscrSetExpression;
Var
   i: integer;
   Name: String;
Begin
   Result := TmscrSetExpression.Create('', msvtVariant, '');
   Try
      For i := 0 To xi.Count - 1 Do
      Begin
         Name := LowerCase(xi[i].Name);
         If Name = 'expr' Then
            Result.Add(Compile_Expression(xi[i], aProj))
         Else If Name = 'range' Then
            Result.Add(Nil);
      End;

   Except
      On e: Exception Do
      Begin
         Result.Free;
         Raise;
      End;
   End;
End;

Function TmscrEngParser.Compile_Expression(xi: TmscrDATAItem; aProj: TmscrScript): TmscrExpression;
Var
   ErPos: String;
   SrcPosition1, SrcPosition2: TPoint;

   //.........................................
   Procedure DoExpressionItems(xi: TmscrDATAItem; Expression: TmscrExpression);
   Var
      i: integer;
      NodeName: String;
      OpName: String;
   Begin
      i := 0;
      While i < xi.Count Do
      Begin
         ErrorPos(xi[i]);
         Expression.SrcPosition := FErrorPos;
         NodeName := Lowercase(xi[i].Name);
         OpName := xi[i].Prop['text'];

         If (NodeName = 'op') Then
         Begin
            OpName := LowerCase(OpName);
            If (OpName = ')') Or (i < xi.Count - 1) Then
               Expression.AddOperator(OpName);
         End
         Else If (NodeName = 'number') Or (NodeName = 'string') Then
            Expression.Register_Const(ParserStringToVariant(OpName))
         Else If NodeName = 'dsgn' Then
            Expression.AddDScanner(DoDScanner(xi[i], aProj))
         Else If NodeName = 'set' Then
            Expression.AddSet(Compile_Set(xi[i], aProj))
         Else If NodeName = 'new' Then
            Expression.AddDScanner(DoDScanner(xi[i][0], aProj, mseoCreate))
         Else If NodeName = 'expr' Then
            DoExpressionItems(xi[i], Expression);

         Inc(i);
      End;
   End;

   //.........................................
   Function GetSource(pt1, pt2: TPoint): String;
   Var
      i1, i2: integer;
   Begin
      i1 := FParser.GetPlainPosition(pt1);
      i2 := FParser.GetPlainPosition(pt2);
      If (i1 = -1) Or (i2 = -1) Then
         Result := ''
      Else
         Result := Copy(FParser.Text, i1, i2 - i1);
   End;

   //.........................................
Begin
   Result := TmscrExpression.Create(FScrProject);
   Try
      DoExpressionItems(xi, Result);
      SrcPosition1 := MSCR_PosToPoint(PropPos(xi));
      SrcPosition2 := MSCR_PosToPoint(xi.Prop['pos1']);
      Result.Source := GetSource(SrcPosition1, SrcPosition2);

      ErPos := Result.Finalize;
      If ErPos <> '' Then
      Begin
         FErrorPos := ErPos;
         Raise Exception.Create(msEIncompatibleTypes);
      End;

   Except
      On e: Exception Do
      Begin
         Result.Free;
         Raise;
      End;
   End;
End;

Procedure TmscrEngParser.Compile_Uses(xi: TmscrDATAItem; aProj: TmscrScript);
Var
   i: integer;
   SaveUnitName: String;
   s: String;
   sl: TStringList;
   ms: TMemoryStream;
   xd: TmscrDATADocument;
Begin
   SaveUnitName := FUnitName;
   FUnitName := xi.Prop['unit'];
   xd := Nil;

   If FUsesList.IndexOf(FUnitName) <> -1 Then
   Begin
      FUnitName := SaveUnitName;
      Exit;
   End;
   FUsesList.Add(FUnitName);
   If Assigned(FScrProject.OnGetFinalUnit) Then
   Begin
      s := '';
      FScrProject.OnGetFinalUnit(FScrProject, FUnitName, s);
      If s <> '' Then
      Begin
         sl := TStringList.Create;
         sl.Text := s;

         ms := TMemoryStream.Create;
         sl.SaveToStream(ms);
         sl.Free;
         ms.Position := 0;

         xd := TmscrDATADocument.Create;
         xd.LoadFromStream(ms);
         ms.Free;
      End;
   End;

   If xd <> Nil Then
   Begin
      Try
         Compile_Project(xd.Root, aProj);
      Finally
         xd.Free;
      End;
   End
   Else
   Begin
      For i := 0 To xi.Count - 1 Do
         Compile_Project(xi[i], aProj);
   End;

   FUnitName := SaveUnitName;
End;

Procedure TmscrEngParser.Compile_Var(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i: integer;
   Name: String;
Begin
   For i := 0 To xi.Count - 1 Do
   Begin
      ErrorPos(xi[i]);
      If CompareText(xi[i].Name, 'ident') = 0 Then
      Begin
         Name := xi[i].Prop['text'];
         CheckIdent(aProj, Name);
         aProj.Add(Name, CreateVar(xi, aProj, Name, Statement));
      End;
   End;
End;

Procedure TmscrEngParser.Compile_Const(xi: TmscrDATAItem; aProj: TmscrScript);
Var
   Name: String;
   Expr: TmscrExpression;
   v: TmscrVarForVariable;
Begin
   Name := xi[0].Prop['text'];
   ErrorPos(xi[0]);
   CheckIdent(aProj, Name);

   Expr := Compile_Expression(xi[1], aProj);
   v := TmscrVarForVariable.Create(Name, Expr.Typ, Expr.TypeName);
   v.Value := Expr.Value;
   v.IsReadOnly := True;
   Expr.Free;

   aProj.Add(Name, v);
End;

Procedure TmscrEngParser.Compile_Parameters(xi: TmscrDATAItem; v: TmscrVarForProc);
Var
   i: integer;
   s: String;
   varParams: boolean;

   //.........................................
   Procedure DoParam(xi: TmscrDATAItem);
   Var
      i: integer;
      Name: String;
      Param: TmscrVarForParamItem;
      varParam: boolean;
   Begin
      varParam := False;

      For i := 0 To xi.Count - 1 Do
      Begin
         ErrorPos(xi[i]);
         If CompareText(xi[i].Name, 'varparam') = 0 Then
            varParam := True
         Else If CompareText(xi[i].Name, 'ident') = 0 Then
         Begin
            Name := xi[i].Prop['text'];
            CheckIdent(v.Project, Name);
            Param := TmscrVarForParamItem(CreateVar(xi, v.Project, Name, Nil, True, varParams Or VarParam));
            Param.DefValue := Param.Value;
            v.Add(Param);
            v.Project.Add(Name, Param);
            varParam := False;
         End;
      End;
   End;

   //.........................................
Begin
   If CompareText(xi.Name, 'parameters') <> 0 Then
      Exit;
   varParams := False;
   For i := 0 To xi.Count - 1 Do
   Begin
      s := LowerCase(xi[i].Name);
      If s = 'varparams' Then
         varParams := True
      Else If s = 'var' Then
      Begin
         DoParam(xi[i]);
         varParams := False;
      End;
   End;
End;

Procedure TmscrEngParser.Compile_ProcA(xi: TmscrDATAItem; aProj: TmscrScript);
Var
   i: integer;
   s, Name: String;
   proc: TmscrVarForProc;
Begin
   ErrorPos(xi[0]);
   Name := xi[0].Prop['text'];
   CheckIdent(aProj, Name);

{$IFDEF CPU64}
   proc := TmscrVarForProc.Create(Name, msvtInt64, '', aProj, False);
{$ELSE}
   proc := TmscrVarForProc.Create(Name, msvtInt, '', aProj, False);
{$ENDIF}

   proc.SrcPosition := PropPos(xi);
   proc.SourceUnit := FUnitName;
   aProj.Add(Name, Proc);

   For i := 0 To xi.Count - 1 Do
   Begin
      s := LowerCase(xi[i].Name);
      If s = 'parameters' Then
         Compile_Parameters(xi[i], Proc);
   End;
End;

Procedure TmscrEngParser.Compile_FuncA(xi: TmscrDATAItem; aProj: TmscrScript);
Var
   i: integer;
   s, Name, TypeName: String;
   Typ: TmscrVarType;
   Func: TmscrVarForProc;
Begin
   Name := '';
   TypeName := '';
   Typ := msvtVariant;

   For i := 0 To xi.Count - 1 Do
   Begin
      ErrorPos(xi[i]);
      s := LowerCase(xi[i].Name);
      If s = 'type' Then
      Begin
         TypeName := xi[i].Prop['text'];
         Typ := FindType(TypeName);
      End
      Else If s = 'name' Then
      Begin
         Name := xi[i].Prop['text'];
         CheckIdent(aProj, Name);
      End;
   End;

   Func := TmscrVarForProc.Create(Name, Typ, TypeName, aProj, CompareText(TypeName, 'void') <> 0);
   Func.SrcPosition := PropPos(xi);
   Func.SourceUnit := FUnitName;
   aProj.Add(Name, Func);

   For i := 0 To xi.Count - 1 Do
   Begin
      s := LowerCase(xi[i].Name);
      If s = 'parameters' Then
         Compile_Parameters(xi[i], Func);
   End;
End;

Procedure TmscrEngParser.Compile_ProcB(xi: TmscrDATAItem; aProj: TmscrScript);
Var
   Name: String;
   proc: TmscrVarForProc;
Begin
   Name := xi[0].Prop['text'];
   proc := TmscrVarForProc(FindVar(aProj, Name));
   Compile_Project(xi, Proc.Project);
End;

Procedure TmscrEngParser.Compile_FuncB(xi: TmscrDATAItem; aProj: TmscrScript);
Var
   i: integer;
   s, Name: String;
   Func: TmscrVarForProc;
Begin
   Name := '';

   For i := 0 To xi.Count - 1 Do
   Begin
      s := LowerCase(xi[i].Name);
      If s = 'name' Then
         Name := xi[i].Prop['text'];
   End;

   Func := TmscrVarForProc(FindVar(aProj, Name));
   Compile_Project(xi, Func.Project);
End;

Procedure TmscrEngParser.Compile_Assign(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i: integer;
   srcstm: TmscrCodeStmAssign_Base;
   DScanner: TmscrDScanner;
   Expression: TmscrExpression;
   Modificator: String;
Begin
   DScanner := Nil;
   Expression := Nil;

   Try
      Modificator := ' ';
      DScanner := DoDScanner(xi[0], aProj);

      i := 1;
      If CompareText(xi[1].Name, 'modificator') = 0 Then
      Begin
         Modificator := xi[1].Prop['text'];
         Inc(i);
      End;
      Expression := Compile_Expression(xi[i], aProj);

      If DScanner.IsReadOnly Then
         Raise Exception.Create(msELeftCantAssigned);

      CheckTypeCompatibility(DScanner, Expression);
      If Modificator = ' ' Then
         Modificator := Expression.Optimize(DScanner);
   Except
      On e: Exception Do
      Begin
         If DScanner <> Nil Then
            DScanner.Free;
         If Expression <> Nil Then
            Expression.Free;
         Raise;
      End;
   End;

   Case Modificator[1] Of
      '+': srcstm := TmscrCodeStmAssign_Plus.Create(aProj, FUnitName, PropPos(xi));
      '-': srcstm := TmscrCodeStmAssign_Minus.Create(aProj, FUnitName, PropPos(xi));
      '*': srcstm := TmscrCodeStmAssign_Mul.Create(aProj, FUnitName, PropPos(xi));
      '/': srcstm := TmscrCodeStmAssign_Div.Create(aProj, FUnitName, PropPos(xi));
      Else
         srcstm := TmscrCodeStmAssign_Base.Create(aProj, FUnitName, PropPos(xi));
   End;

   Statement.Add(srcstm);
   srcstm.DScanner := DScanner;
   srcstm.Expression := Expression;
   srcstm.Optimize;
   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Call(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   srcstm: TmscrCodeStm_Call;
Begin
   srcstm := TmscrCodeStm_Call.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(srcstm);
   srcstm.DScanner := DoDScanner(xi[0], aProj);
   If xi.Count > 1 Then
   Begin
      srcstm.Modificator := xi[1].Prop['text'];
      If srcstm.DScanner.IsReadOnly Then
         Raise Exception.Create(msELeftCantAssigned);
   End;

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_If(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i: integer;
   s: String;
   srcstm: TmscrCodeStm_If;
Begin
   srcstm := TmscrCodeStm_If.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(srcstm);
   srcstm.Condition := Compile_Expression(xi[0], aProj);

   For i := 1 To xi.Count - 1 Do
   Begin
      s := Lowercase(xi[i].Name);
      If s = 'thensrcstm' Then
         Compile_mscodecompound(xi[1], aProj, srcstm)
      Else If s = 'elsesrcstm' Then
         Compile_mscodecompound(xi[2], aProj, srcstm.Elsesrcstm);
   End;

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_For(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i: integer;
   srcstm: TmscrCodeStm_For;
Begin
   srcstm := TmscrCodeStm_For.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(srcstm);
   ErrorPos(xi[0]);
   srcstm.Variable := FindVar(aProj, xi[0].Prop['text']);
   If Not ((srcstm.Variable Is TmscrVarForVariable) And (srcstm.Variable.Typ In [msvtInt, msvtInt64, msvtVariant, msvtFloat])) Then
      Raise Exception.Create(msEForError);

   srcstm.ValueBegin := Compile_Expression(xi[1], aProj);
   CheckTypeCompatibility(srcstm.Variable, srcstm.ValueBegin);

   i := 2;
   If CompareText(xi[2].Name, 'downto') = 0 Then
   Begin
      srcstm.Down := True;
      Inc(i);
   End;

   srcstm.ValueEnd := Compile_Expression(xi[i], aProj);
   CheckTypeCompatibility(srcstm.Variable, srcstm.ValueEnd);
   If i + 1 < xi.Count Then
      Compile_SrcStm(xi[i + 1], aProj, srcstm);

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_While(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   xsrcstm: TmscrCodeStm_While;
Begin
   xsrcstm := TmscrCodeStm_While.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(xsrcstm);
   xsrcstm.Condition := Compile_Expression(xi[0], aProj);
   If xi.Count > 1 Then
      Compile_SrcStm(xi[1], aProj, xsrcstm);

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Repeat(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i, j: integer;
   xsrcstm: TmscrCodeStm_Repeat;
Begin
   xsrcstm := TmscrCodeStm_Repeat.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(xsrcstm);

   j := xi.Count - 1;
   If CompareText(xi[j].Name, 'inverse') = 0 Then
   Begin
      xsrcstm.InverseCondition := True;
      Dec(j);
   End;
   xsrcstm.Condition := Compile_Expression(xi[j], aProj);
   Dec(j);

   For i := 0 To j Do
      Compile_SrcStm(xi[i], aProj, xsrcstm);

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Case(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i: integer;
   srcstm: TmscrCodeStm_Case;

   //.........................................
   Procedure DoCaseSelector(xi: TmscrDATAItem);
   Var
      Selector: TmscrCaseSelector;
   Begin
      If (CompareText(xi.Name, 'caseselector') <> 0) Or (xi.Count <> 2) Then
         Exit;
      Selector := TmscrCaseSelector.Create(aProj, FUnitName, PropPos(xi));
      srcstm.Add(Selector);

      Selector.SetExpression := Compile_Set(xi[0], aProj);
      Compile_SrcStm(xi[1], aProj, Selector);
   End;

   //.........................................
Begin
   srcstm := TmscrCodeStm_Case.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(srcstm);
   srcstm.Condition := Compile_Expression(xi[0], aProj);

   For i := 1 To xi.Count - 1 Do
      DoCaseSelector(xi[i]);
   If CompareText(xi[xi.Count - 1].Name, 'caseselector') <> 0 Then
      Compile_SrcStm(xi[xi.Count - 1], aProj, srcstm.Elsesrcstm);

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Try(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i: integer;
   srcstm: TmscrCodesrcstmry;
Begin
   srcstm := TmscrCodesrcstmry.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(srcstm);

   For i := 0 To xi.Count - 1 Do
      If CompareText(xi[i].Name, 'ExceptSrcStm') = 0 Then
      Begin
         srcstm.IsExcept := True;
         Compile_mscodecompound(xi[i], aProj, srcstm.ExceptSrcStm);
      End
      Else If CompareText(xi[i].Name, 'finallysrcstm') = 0 Then
         Compile_mscodecompound(xi[i], aProj, srcstm.ExceptSrcStm)
      Else
         Compile_SrcStm(xi[i], aProj, srcstm);

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Continue(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   srcstm: TmscrCodeStm_Continue;
Begin
   srcstm := TmscrCodeStm_Continue.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(srcstm);
   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Exit(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   xsrcstm: TmscrCodeStm_Exit;
Begin
   xsrcstm := TmscrCodeStm_Exit.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(xsrcstm);
   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Break(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   xsrcstm: TmscrCodeStm_Break;
Begin
   xsrcstm := TmscrCodeStm_Break.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(xsrcstm);
   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Return(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   xi1: TmscrDATAItem;
Begin
   If xi.Count = 1 Then
   Begin
      xi1 := TmscrDATAItem.Create;
      xi1.Name := 'dsgn';
      xi.InsertItem(0, xi1);
      With xi1.Add Do
      Begin
         Name := 'node';
         Text := 'text="Result" pos="' + xi[1].Prop['pos'] + '"';
      End;

      Compile_Assign(xi, aProj, Statement);
   End;

   Compile_Exit(xi, aProj, Statement);
End;

Procedure TmscrEngParser.Compile_With(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   d: TmscrDScanner;
   i, n: integer;
   s: String;
   v: TmscrVarForVariable;
   srcstm: TmscrCodeStm_With;

   //............................
   Function CreateUniqueVariable: String;
   Var
      i: integer;
   Begin
      i := 0;
      While (aProj.FindLocal(IntToStr(i)) <> Nil) Or (FWithList.IndexOf(IntToStr(i)) <> -1) Do
         Inc(i);
      Result := '_WithList_' + IntToStr(i);
   End;

   //............................
Begin
   n := xi.Count - 1;

   For i := 0 To n - 1 Do
   Begin
      d := DoDScanner(xi[i], aProj);
      If Not ((d.Typ = msvtClass) Or (d.Typ = msvtVariant)) Then
      Begin
         d.Free;
         Raise Exception.Create(msEClassRequired);
      End;

      s := CreateUniqueVariable;
      v := TmscrVarForVariable.Create(s, d.Typ, d.TypeName);
      aProj.Add(s, v);

      srcstm := TmscrCodeStm_With.Create(aProj, FUnitName, PropPos(xi));
      srcstm.Variable := v;
      srcstm.DScanner := d;
      Statement.Add(srcstm);
      FWithList.AddObject(s, srcstm);
   End;

   Compile_SrcStm(xi[xi.Count - 1], aProj, Statement);

   For i := 0 To n - 1 Do
      FWithList.Delete(FWithList.Count - 1);

   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_Delete(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   xsrcstm: TmscrCodeStm_Call;
Begin
   xsrcstm := TmscrCodeStm_Call.Create(aProj, FUnitName, PropPos(xi));
   Statement.Add(xsrcstm);
   xsrcstm.DScanner := DoDScanner(xi[0], aProj, mseoFree);
   FScrProject.AddCodeLine(FUnitName, PropPos(xi));
End;

Procedure TmscrEngParser.Compile_mscodecompound(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   i: integer;
Begin
   For i := 0 To xi.Count - 1 Do
      Compile_SrcStm(xi[i], aProj, Statement);
End;

Procedure TmscrEngParser.Compile_SrcStm(xi: TmscrDATAItem; aProj: TmscrScript; Statement: TmscrStatement);
Var
   s: String;
Begin
   s := LowerCase(xi.Name);

   If s = 'codeassign' Then
   Begin
      Compile_Assign(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscodecall' Then
   Begin
      Compile_Call(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscodeif' Then
   Begin
      Compile_If(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscodecases' Then
   Begin
      Compile_Case(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscodefor' Then
   Begin
      Compile_For(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscodewhile' Then
   Begin
      Compile_While(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscoderepeat' Then
   Begin
      Compile_Repeat(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscodetry' Then
   Begin
      Compile_Try(xi, aProj, Statement);
      Exit;
   End;
   If s = 'break' Then
   Begin
      Compile_Break(xi, aProj, Statement);
      Exit;
   End;
   If s = 'continue' Then
   Begin
      Compile_Continue(xi, aProj, Statement);
      Exit;
   End;
   If s = 'exit' Then
   Begin
      Compile_Exit(xi, aProj, Statement);
      Exit;
   End;
   If s = 'return' Then
   Begin
      Compile_Return(xi, aProj, Statement);
      Exit;
   End;
   If s = 'with' Then
   Begin
      Compile_With(xi, aProj, Statement);
      Exit;
   End;
   If s = 'delete' Then
   Begin
      Compile_Delete(xi, aProj, Statement);
      Exit;
   End;
   If s = 'mscodecompound' Then
   Begin
      Compile_mscodecompound(xi, aProj, Statement);
      Exit;
   End;
   If s = 'uses' Then
   Begin
      Compile_Uses(xi, aProj);
      Exit;
   End;
   If s = 'var' Then
   Begin
      Compile_Var(xi, aProj, Statement);
      Exit;
   End;
   If s = 'const' Then
   Begin
      Compile_Const(xi, aProj);
      Exit;
   End;
   If s = 'procedure' Then
   Begin
      Compile_ProcB(xi, aProj);
      Exit;
   End;
   If s = 'function' Then
   Begin
      Compile_FuncB(xi, aProj);
      Exit;
   End;
End;

Procedure TmscrEngParser.Compile_Project(xi: TmscrDATAItem; aProj: TmscrScript);
Var
   TempRoot: TmscrDATAItem;

   //...........................................
   Procedure DoFirstPass(xi: TmscrDATAItem);
   Var
      i: integer;
      s: String;
   Begin
      For i := 0 To xi.Count - 1 Do
      Begin
         s := LowerCase(xi[i].Name);
         If s = 'mscodecompound' Then
            DoFirstPass(xi[i])
         Else If s = 'procedure' Then
            Compile_ProcA(xi[i], aProj)
         Else If s = 'function' Then
            Compile_FuncA(xi[i], aProj);
      End;
   End;

   //...........................................
Begin
   TempRoot := FProjRoot;
   FProjRoot := xi;
   DoFirstPass(xi);
   Compile_mscodecompound(xi, aProj, aProj.Statement);
   FProjRoot := TempRoot;
End;

//====================================================

{$IFDEF MAGSCR_USEOLE}

Constructor TmscrEngOLE.Create(Const AName: String);
Begin
   Inherited Create(AName, msvtVariant, '');
End;

Function TmscrEngOLE.DispatchInvoke(Const ParamArray: variant; ParamCount: integer; Flags: word): variant;
Const
   DispIDArgs: longint = DISPID_PROPERTYPUT;
Var
   DispId: TDispId;
   Params: TDISPPARAMS;
   pName: WideString;
   ExcepMess: WideString;
   Args: Array[0..63] Of variant;
   i: integer;
   PResult: PVariant;
   Status: integer;
   ExcepInfo: TExcepInfo;
Begin
   ExcepMess := '';
   pName := WideString(Name);
   IDispatch(ParentValue).GetIDsOfNames(GUID_NULL, @pName, 1, GetThreadLocale, @DispId);

   For i := 0 To ParamCount - 1 Do
      Args[i] := ParamArray[ParamCount - i - 1];

   Params.rgvarg := @Args;
   Params.rgdispidNamedArgs := Nil;
   Params.cArgs := ParamCount;
   Params.cNamedArgs := 0;

   If (Flags = DISPATCH_PROPERTYPUT) Or (Flags = DISPATCH_PROPERTYPUTREF) Then
   Begin
      Params.rgdispidNamedArgs := @DispIDArgs;
      Params.cNamedArgs := 1;
   End;

   If NeedResult And (Flags <> DISPATCH_PROPERTYPUT) And (Flags <> DISPATCH_PROPERTYPUTREF) Then
      PResult := @Result
   Else
      PResult := Nil;

   If PResult <> Nil Then Variants.VarClear(PResult^);

   If (Flags = DISPATCH_METHOD) And (PResult <> Nil) Then
      Flags := DISPATCH_METHOD Or DISPATCH_PROPERTYGET;

   Status := IDispatch(ParentValue).Invoke(DispId, GUID_NULL, 0, Flags, Params, PResult, @ExcepInfo, Nil);

   If Status <> 0 Then
   Begin
      If ExcepInfo.Source <> '' Then
         ExcepMess := #13 + #10 + 'Source          ::  ' + ExcepInfo.Source;
      If ExcepInfo.Description <> '' Then
         ExcepMess := ExcepMess + #13#10 + 'Description ::  ' + ExcepInfo.Description;
      If ExcepInfo.HelpFile <> '' Then
         ExcepMess := ExcepMess + #13#10 + 'Help File      ::  ' + ExcepInfo.HelpFile;

      Raise Exception.Create('OLE error ' + IntToHex(Status, 8) + ': ' + Name + ': ' + SysErrorMessage(Status) + ExcepMess);
   End;
End;

Procedure TmscrEngOLE.SetValue(Const Value: variant);
Var
   i: integer;
   v: variant;
   Flag: word;
Begin
   v := VarArrayCreate([0, Count], varVariant);

   For i := 0 To Count - 1 Do
      v[i] := Params[i].Value;

   v[Count] := Value;
   Flag := DISPATCH_PROPERTYPUT;

   If VarType(Value) = varDispatch Then
      Flag := DISPATCH_PROPERTYPUTREF;

   DispatchInvoke(v, Count + 1, Flag);
   ParentValue := Null;
End;

Function TmscrEngOLE.GetValue: variant;
Var
   i: integer;
   v: variant;
Begin
   v := VarArrayCreate([0, Count - 1], varVariant);
   For i := 0 To Count - 1 Do
      v[i] := Params[i].Value;

   Result := DispatchInvoke(v, Count, DISPATCH_METHOD);
   ParentValue := Null;
End;
{$ENDIF}

//====================================================

Function FormatV(Const Fmt: String; Args: variant): String;
Var
   ar: TmscrVarRecArray;
   sPtrList: TList;
Begin
   VariantToVarRec(Args, ar, sPtrList);
   Try
      Result := Format(Fmt, ar);
   Finally
      ClearVarRec(ar, sPtrList);
   End;
End;

Function VArrayCreate(Args: variant; Typ: integer): variant;
Var
   i, n: integer;
   ar: Array Of SizeInt;
Begin
   n := VarArrayHighBound(Args, 1) + 1;
   SetLength(ar, n);
   For i := 0 To n - 1 Do
      ar[i] := Args[i];

   Result := VarArrayCreate(ar, Typ);
   ar := Nil;
End;

Function NameCase(Const s: String): String;
Var
   i: integer;
Begin

   Result := UTF8LowerString(s);
   For i := 1 To UTF8Length(s) Do
      If i = 1 Then
      Begin
         UTF8Insert(UTF8UpperString(UTF8Copy(s, i, 1)), Result, i + 1);
         UTF8Delete(Result, i, 1);
      End
      Else If i < UTF8Length(s) Then
         If UTF8Copy(s, i, 1)[1] = ' ' Then
         Begin
            UTF8Insert(UTF8UpperString(UTF8Copy(s, i + 1, 1)), Result, i + 2);
            UTF8Delete(Result, i + 1, 1);
         End;
End;

Function ValidInt(cInt: String): boolean;
Begin
   Result := True;
   Try
      StrToInt(cInt);
   Except
      Result := False;
   End;
End;

Function ValidFloat(cFlt: String): boolean;
Begin
   Result := True;
   Try
      StrToFloat(cFlt);
   Except
      Result := False;
   End;
End;

Function ValidDate(cDate: String): boolean;
Begin
   Result := True;
   Try
      StrToDate(cDate);
   Except
      Result := False;
   End;
End;

Function DaysInMonth(nYear, nMonth: integer): integer;
Const
   Days: Array[1..12] Of integer = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
Begin
   Result := Days[nMonth];
   If (nMonth = 2) And IsLeapYear(nYear) Then
      Inc(Result);
End;

Procedure Write(Msg: variant);
Var
   i: integer;
Begin
   If VarMSCR_ConcoleOut = Nil Then Exit;
   i := VarMSCR_ConcoleOut.Lines.Count - 1;

   VarMSCR_ConcoleOut.Lines[i] := VarMSCR_ConcoleOut.Lines[i] + Msg;
End;

Procedure WriteLN(Msg: variant);
Begin
   If VarMSCR_ConcoleOut = Nil Then Exit;
   VarMSCR_ConcoleOut.Lines.Add(Msg);
End;

Procedure Print(Msg: variant);
Begin
   Write(Msg);
End;

Procedure PrintLN(Msg: variant);
Begin
   WriteLN(Msg);
End;

//============== TmscrSysFunctions ===============================

Constructor TmscrSysFunctions.Create(AScript: TmscrScript);
Begin
   Inherited Create(AScript);
   FCatStr := cnCatStr;
   FCatDate := cnCatDate;
   FCatConv := cnCatConv;
   FCatFormat := cnCatFormat;
   FCatMath := cnCatMath;
   FCatOther := cnCatOther;

   With AScript Do
   Begin
      Register_Type('Byte', msvtInt);
      Register_Type('Word', msvtInt);
      Register_Type('Integer', msvtInt);
      Register_Type('Int64', msvtInt64);
      Register_Type('Longint', msvtInt);
      Register_Type('Cardinal', msvtInt);
      Register_Type('TColor', msvtInt);
      Register_Type('TAlphaColor', msvtInt);
      Register_Type('Boolean', msvtBool);
      Register_Type('Real', msvtFloat);
      Register_Type('Single', msvtFloat);
      Register_Type('Double', msvtFloat);
      Register_Type('Float', msvtFloat);
      Register_Type('Extended', msvtFloat);
      Register_Type('Currency', msvtFloat);
      Register_Type('TDate', msvtFloat);
      Register_Type('TTime', msvtFloat);
      Register_Type('TDateTime', msvtFloat);
      Register_Type('Char', msvtChar);
      Register_Type('String', msvtString);
      Register_Type('Variant', msvtVariant);
      Register_Type('Pointer', msvtVariant);
      Register_Type('Array', msvtArray);
      Register_Type('Constructor', msvtConstructor);

      Register_Const('True', 'Boolean', True);
      Register_Const('False', 'Boolean', False);
      Register_Const('nil', 'Variant', 0);
      Register_Const('Null', 'Variant', Null);

      //--- From fpcsrc/rtl/inc/varianth.inc ----
      Register_Const('varEmpty', 'Integer', 0);
      Register_Const('varNull', 'Integer', 1);
      Register_Const('varSmallint', 'Integer', 2);
      Register_Const('varInteger', 'Integer', 3);
      Register_Const('varSingle', 'Integer', 4);
      Register_Const('varDouble', 'Integer', 5);
      Register_Const('varDate', 'Integer', 7);
      Register_Const('varCurrency', 'Integer', 6);
      Register_Const('varOleStr', 'Integer', 8);
      Register_Const('varDispatch', 'Integer', 9);
      Register_Const('varError', 'Integer', 10);
      Register_Const('varBoolean', 'Integer', 11);
      Register_Const('varVariant', 'Integer', 12);
      Register_Const('varUnknown', 'Integer', 13);
      Register_Const('varDecimal', 'Integer', 14);
      Register_Const('varShortInt', 'Integer', 16);
      Register_Const('varByte', 'Integer', 17);
      Register_Const('varWord', 'Integer', 18);
      Register_Const('varLongWord', 'Integer', 19);
      Register_Const('varInt64', 'Integer', 20);
      Register_Const('varQword', 'Integer', 21);
      Register_Const('varRecord', 'Integer', 36);
      Register_Const('varStrArg', 'Integer', $48);
      Register_Const('varUStrArg', 'Integer', $49);
      Register_Const('varString', 'Integer', $100);
      Register_Const('varAny', 'Integer', $101);
      Register_Const('varUString', 'Integer', $102);
      Register_Const('varTypeMask', 'Integer', $FFF);
      Register_Const('varArray', 'Integer', $2000);
      Register_Const('varByRef', 'Integer', $4000);

      Add('__HelperForString', TmscrEngString.Create);
      Add('__HelperForArray', TmscrEngArray.Create('__HelperForArray', -1, msvtVariant, ''));
      Register_Variable('ExceptionClassName', 'String', '');
      Register_Variable('ExceptionMessage', 'String', '');

      //.... Conversion ....
      Register_Method('function IntToStr(i: Integer): String', CallMethod1, FCatConv);
      Register_Method('function FloatToStr(e: Extended): String', CallMethod1, FCatConv);
      Register_Method('function DateToStr(e: Extended): String', CallMethod1, FCatConv);
      Register_Method('function TimeToStr(e: Extended): String', CallMethod1, FCatConv);
      Register_Method('function DateTimeToStr(e: Extended): String', CallMethod1, FCatConv);
      Register_Method('function BoolToStr(B: Boolean): string;', CallMethod1, FCatConv);
      Register_Method('function VarToStr(v: Variant): String', CallMethod7, FCatConv);
      Register_Method('function StrToInt(s: String): Integer', CallMethod2, FCatConv);
      Register_Method('function StrToInt64(s: String): Int64', CallMethod2, FCatConv);
      Register_Method('function StrToFloat(s: String): Extended', CallMethod2, FCatConv);
      Register_Method('function StrToDate(s: String): Extended', CallMethod2, FCatConv);
      Register_Method('function StrToTime(s: String): Extended', CallMethod2, FCatConv);
      Register_Method('function StrToDateTime(s: String): Extended', CallMethod2, FCatConv);
      Register_Method('function StrToBool(const S: string): Boolean;', CallMethod2, FCatConv);

      //.... Format ....
      Register_Method('function Format(Fmt: String; Args: array): String', CallMethod3, FCatFormat);
      Register_Method('function FormatFloat(Fmt: String; Value: Extended): String', CallMethod3, FCatFormat);
      Register_Method('function FormatDateTime(Fmt: String; DateTime: TDateTime): String', CallMethod3, FCatFormat);
      Register_Method('function FormatMaskText(EditMask: string; Value: string): string', CallMethod3, FCatFormat);

      //.... Date/Time ....
      Register_Method('function EncodeDate(Year, Month, Day: Word): TDateTime', CallMethod4, FCatDate);
      Register_Method('procedure DecodeDate(Date: TDateTime; var Year, Month, Day: Word)', CallMethod4, FCatDate);
      Register_Method('function EncodeTime(Hour, Min, Sec, MSec: Word): TDateTime', CallMethod4, FCatDate);
      Register_Method('procedure DecodeTime(Time: TDateTime; var Hour, Min, Sec, MSec: Word)', CallMethod4, FCatDate);
      Register_Method('function Date: TDateTime', CallMethod4, FCatDate);
      Register_Method('function Time: TDateTime', CallMethod4, FCatDate);
      Register_Method('function Now: TDateTime', CallMethod4, FCatDate);
      Register_Method('function DayOfWeek(aDate: TDateTime): Integer', CallMethod4, FCatDate);
      Register_Method('function IsLeapYear(Year: Word): Boolean', CallMethod4, FCatDate);
      Register_Method('function DaysInMonth(nYear, nMonth: Integer): Integer', CallMethod4, FCatDate);

      //.... String ....
      Register_Method('function Length(s: Variant): Integer', CallMethod5, FCatStr);
      Register_Method('function Copy(s: String; from, count: Integer): String', CallMethod5, FCatStr);
      Register_Method('function Pos(substr, s: String): Integer', CallMethod5, FCatStr);
      Register_Method('procedure Delete(var s: String; from, count: Integer)', CallMethod5, FCatStr);
      Register_Method('procedure DeleteStr(var s: String; from, count: Integer)', CallMethod5, FCatStr);
      Register_Method('procedure Insert(s: String; var s2: String; pos: Integer)', CallMethod5, FCatStr);
      Register_Method('function Uppercase(s: String): String', CallMethod5, FCatStr);
      Register_Method('function Lowercase(s: String): String', CallMethod5, FCatStr);
      Register_Method('function Trim(s: String): String', CallMethod5, FCatStr);
      Register_Method('function NameCase(s: String): String', CallMethod5, FCatStr);
      Register_Method('function CompareText(s, s1: String): Integer', CallMethod5, FCatStr);
      Register_Method('function Chr(i: Integer): Char', CallMethod5, FCatStr);
      Register_Method('function Ord(ch: Char): Integer', CallMethod5, FCatStr);
      Register_Method('procedure SetLength(var S: Variant; L: Integer)', CallMethod5, FCatStr);

      //.... Math ....
      Register_Method('function Round(e: Extended): Integer', CallMethod6, FCatMath);
      Register_Method('function Trunc(e: Extended): Integer', CallMethod6, FCatMath);
      Register_Method('function Int(e: Extended): Integer', CallMethod6, FCatMath);
      Register_Method('function Frac(X: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Sqrt(e: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Abs(e: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Sin(e: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Cos(e: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function ArcTan(X: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Tan(X: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Exp(X: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Ln(X: Extended): Extended', CallMethod6, FCatMath);
      Register_Method('function Pi: Extended', CallMethod6, FCatMath);

      Register_Method('procedure Inc(var i: Integer; incr: Integer = 1)', CallMethod7, FCatOther);
      Register_Method('procedure Dec(var i: Integer; decr: Integer = 1)', CallMethod7, FCatOther);
      Register_Method('procedure RaiseException(Param: String)', CallMethod7, FCatOther);
      Register_Method('procedure ShowMessage(Msg: Variant)', CallMethod7, FCatOther);

      //.... Other ....
      Register_Method('procedure Print(Msg: Variant)', CallMethod7, FCatOther);
      Register_Method('procedure PrintLN(Msg: Variant)', CallMethod7, FCatOther);
      Register_Method('procedure Write(Msg: Variant)', CallMethod7, FCatOther);
      Register_Method('procedure WriteLN(Msg: Variant)', CallMethod7, FCatOther);

      Register_Method('procedure Randomize', CallMethod7, FCatOther);
      Register_Method('function Random: Extended', CallMethod7, FCatOther);
      Register_Method('function RandomInt(l:Int64):Int64;', CallMethod7, FCatOther);
      Register_Method('function ValidInt(cInt: String): Boolean', CallMethod7, FCatOther);
      Register_Method('function ValidFloat(cFlt: String): Boolean', CallMethod7, FCatOther);
      Register_Method('function ValidDate(cDate: String): Boolean', CallMethod7, FCatOther);
      Register_Method('function ExtractFilePath(const FileName: string): string;', CallMethod7, FCatOther);
{$IFDEF MAGSCR_USEOLE}
      Register_Method('function CreateOleObject(ClassName: String): Variant', CallMethod7, FCatOther);
      Register_Method('function GetActiveOleObject(const ClassName: string) : Variant', CallMethod7, FCatOther);
{$ENDIF}
      Register_Method('function VarArrayCreate(Bounds: Array; Typ: Integer): Variant', CallMethod7, FCatOther);
      Register_Method('function VarType(V: Variant): Integer', CallMethod7, FCatOther);

   End;
End;

Function TmscrSysFunctions.CallMethod1(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
Var
   i: int64;
Begin
   If aMethodName = 'INTTOSTR' Then
   Begin
      i := aCallVar.Params[0];
      Result := IntToStr(i);
      exit;
   End;

   If aMethodName = 'FLOATTOSTR' Then
   Begin
      Result := FloatToStr(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'DATETOSTR' Then
   Begin
      Result := DateToStr(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'TIMETOSTR' Then
   Begin
      Result := TimeToStr(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'DATETIMETOSTR' Then
   Begin
      Result := DateTimeToStr(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'BOOLTOSTR' Then
   Begin
      Result := BoolToStr(aCallVar.Params[0]);
      exit;
   End;
End;

Function TmscrSysFunctions.CallMethod2(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
Begin
   If aMethodName = 'STRTOINT' Then
   Begin
      Result := StrToInt(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'STRTOINT64' Then
   Begin
      Result := StrToInt64(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'STRTOFLOAT' Then
   Begin
      Result := StrToFloat(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'STRTODATE' Then
   Begin
      Result := StrToDate(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'STRTOTIME' Then
   Begin
      Result := StrToTime(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'STRTODATETIME' Then
   Begin
      Result := StrToDateTime(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'STRTOBOOL' Then
   Begin
      Result := BoolToStr(aCallVar.Params[0]);
      exit;
   End;
End;

Function TmscrSysFunctions.CallMethod3(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
Begin
   If aMethodName = 'FORMAT' Then
   Begin
      Result := FormatV(aCallVar.Params[0], aCallVar.Params[1]);
      exit;
   End;
   If aMethodName = 'FORMATFLOAT' Then
   Begin
      Result := FormatFloat(aCallVar.Params[0], aCallVar.Params[1] * 1.00000);
      exit;
   End;
   If aMethodName = 'FORMATDATETIME' Then
   Begin
      Result := FormatDateTime(aCallVar.Params[0], aCallVar.Params[1]);
      exit;
   End;
End;

Function TmscrSysFunctions.CallMethod4(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
Var
   w1, w2, w3, w4: word;
Begin
   If aMethodName = 'ENCODEDATE' Then
   Begin
      Result := EncodeDate(aCallVar.Params[0], aCallVar.Params[1], aCallVar.Params[2]);
      exit;
   End;
   If aMethodName = 'ENCODETIME' Then
   Begin
      Result := EncodeTime(aCallVar.Params[0], aCallVar.Params[1], aCallVar.Params[2], aCallVar.Params[3]);
      exit;
   End;

   If aMethodName = 'DECODEDATE' Then
   Begin
      DecodeDate(aCallVar.Params[0], w1, w2, w3);
      aCallVar.Params[1] := w1;
      aCallVar.Params[2] := w2;
      aCallVar.Params[3] := w3;
      exit;
   End;

   If aMethodName = 'DECODETIME' Then
   Begin
      DecodeTime(aCallVar.Params[0], w1, w2, w3, w4);
      aCallVar.Params[1] := w1;
      aCallVar.Params[2] := w2;
      aCallVar.Params[3] := w3;
      aCallVar.Params[4] := w4;
      exit;
   End;

   If aMethodName = 'DATE' Then
   Begin
      Result := SysUtils.Date;
      exit;
   End;
   If aMethodName = 'TIME' Then
   Begin
      Result := SysUtils.Time;
      exit;
   End;
   If aMethodName = 'NOW' Then
   Begin
      Result := SysUtils.Now;
      exit;
   End;
   If aMethodName = 'DAYOFWEEK' Then
   Begin
      Result := SysUtils.DayOfWeek(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'ISLEAPYEAR' Then
   Begin
      Result := SysUtils.IsLeapYear(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'DAYSINMONTH' Then
   Begin
      Result := DaysInMonth(aCallVar.Params[0], aCallVar.Params[1]);
      exit;
   End;
End;

Function TmscrSysFunctions.CallMethod5(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
Var
   s: String;
   v: variant;
Begin
   If aMethodName = 'LENGTH' Then
   Begin
      v := aCallVar.Params[0];
      If VarIsArray(v) Then
         Result := VarArrayHighBound(v, 1) - VarArrayLowBound(v, 1) + 1
      Else
         Result := UTF8Length(v);
      exit;
   End;
   If aMethodName = 'COPY' Then
   Begin
      Result := UTF8Copy(aCallVar.Params[0], fmscrInteger(aCallVar.Params[1]), fmscrInteger(aCallVar.Params[2]));
      exit;
   End;
   If aMethodName = 'POS' Then
   Begin
      Result := UTF8Pos(ansistring(aCallVar.Params[0]), ansistring(aCallVar.Params[1]));
      exit;
   End;
   If (aMethodName = 'DELETE') Or (aMethodName = 'DELETESTR') Then
   Begin
      s := aCallVar.Params[0];
      UTF8Delete(s, fmscrInteger(aCallVar.Params[1]), fmscrInteger(aCallVar.Params[2]));
      aCallVar.Params[0] := s;
      exit;
   End;
   If aMethodName = 'INSERT' Then
   Begin
      s := aCallVar.Params[1];
      UTF8Insert(ansistring(aCallVar.Params[0]), s, integer(aCallVar.Params[2]));
      aCallVar.Params[1] := s;
      exit;
   End;
   If aMethodName = 'UPPERCASE' Then
   Begin
      Result := UTF8UpperString(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'LOWERCASE' Then
   Begin
      Result := UTF8LowerString(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'TRIM' Then
   Begin
      Result := UTF8Trim(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'NAMECASE' Then
   Begin
      Result := NameCase(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'COMPARETEXT' Then
   Begin
      Result := UTF8CompareText(aCallVar.Params[0], aCallVar.Params[1]);
      exit;
   End;
   If aMethodName = 'CHR' Then
   Begin
      Result := Chr(integer(aCallVar.Params[0]));
      exit;
   End;
   If aMethodName = 'ORD' Then
   Begin
      Result := Ord(String(aCallVar.Params[0])[1]);
      exit;
   End;
   If aMethodName = 'SETLENGTH' Then
   Begin
      If (TVarData(aCallVar.Params[0]).VType = varString) Or (TVarData(aCallVar.Params[0]).VType = varOleStr) Then
      Begin
         s := aCallVar.Params[0];
         SetLength(s, integer(aCallVar.Params[1]));
         aCallVar.Params[0] := s;
      End
      Else
      Begin
         v := aCallVar.Params[0];
         VarArrayRedim(v, integer(aCallVar.Params[1]) - 1);
         aCallVar.Params[0] := v;
      End;
      exit;
   End;
End;

Function TmscrSysFunctions.CallMethod6(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
Begin
   If aMethodName = 'ROUND' Then
   Begin
      Result := fmscrInteger(Round(extended(aCallVar.Params[0]) * 1.00000));
      exit;
   End;
   If aMethodName = 'TRUNC' Then
   Begin
      Result := fmscrInteger(Trunc(extended(aCallVar.Params[0]) * 1.00000));
      exit;
   End;
   If aMethodName = 'INT' Then
   Begin
      Result := Int(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'FRAC' Then
   Begin
      Result := Frac(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'SQRT' Then
   Begin
      Result := Sqrt(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'ABS' Then
   Begin
      Result := Abs(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'SIN' Then
   Begin
      Result := Sin(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'COS' Then
   Begin
      Result := Cos(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'ARCTAN' Then
   Begin
      Result := ArcTan(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'TAN' Then
   Begin
      Result := Sin(aCallVar.Params[0] * 1.00000) / Cos(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'EXP' Then
   Begin
      Result := Exp(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'LN' Then
   Begin
      Result := Ln(aCallVar.Params[0] * 1.00000);
      exit;
   End;
   If aMethodName = 'PI' Then
   Begin
      Result := Pi;
      exit;
   End;
End;

Function TmscrSysFunctions.CallMethod7(aInstance: TObject; aClassType: TClass; Const aMethodName: String; aCallVar: TmscrVarForMethod): variant;
Begin
   If aMethodName = 'INC' Then
   Begin
      aCallVar.Params[0] := aCallVar.Params[0] + aCallVar.Params[1];
      exit;
   End;
   If aMethodName = 'DEC' Then
   Begin
      aCallVar.Params[0] := aCallVar.Params[0] - aCallVar.Params[1];
      exit;
   End;
   If aMethodName = 'RAISEEXCEPTION' Then
   Begin
      Raise Exception.Create(aCallVar.Params[0]);
      exit;
   End;
{$IFNDEF MAGSCR_NOFORMS}
   If aMethodName = 'SHOWMESSAGE' Then
   Begin
      ShowMessage(aCallVar.Params[0]);
      exit;
   End;
{$ENDIF}
   If aMethodName = 'PRINT' Then
   Begin
      Print(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'PRINTLN' Then
   Begin
      PrintLN(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'WRITE' Then
   Begin
      Write(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'WRITELN' Then
   Begin
      WriteLN(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'RANDOMIZE' Then
   Begin
      Randomize;
      exit;
   End;
   If aMethodName = 'RANDOM' Then
   Begin
      Result := Random;
      exit;
   End;
   If aMethodName = 'RANDOMINT' Then
   Begin
      Result := Random(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'VALIDINT' Then
   Begin
      Result := ValidInt(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'VALIDFLOAT' Then
   Begin
      Result := ValidFloat(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'VALIDDATE' Then
   Begin
      If VarType(aCallVar.Params[0]) = varDate Then
         Result := ValidDate(DateToStr(aCallVar.Params[0]))
      Else
         Result := ValidDate(aCallVar.Params[0]);
      exit;
   End;
{$IFDEF MAGSCR_USEOLE}
   If aMethodName = 'CREATEOLEOBJECT' Then
   Begin
      Result := CreateOleObject(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'GETACTIVEOLEOBJECT' Then
   Begin
      Result := GetActiveOleObject(aCallVar.Params[0]);
      exit;
   End;
{$ENDIF}
   If aMethodName = 'VARARRAYCREATE' Then
   Begin
      Result := VArrayCreate(aCallVar.Params[0], aCallVar.Params[1]);
      exit;
   End;
   If aMethodName = 'VARTOSTR' Then
   Begin
      Result := VarToStr(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'VARTYPE' Then
   Begin
      Result := VarType(aCallVar.Params[0]);
      exit;
   End;
   If aMethodName = 'EXTRACTFILEPATH' Then
   Begin
      Result := ExtractFilePath(aCallVar.Params[0]);
      exit;
   End;
End;

//===================================================================
//===================================================================
//===================================================================

Initialization
   FGlobalScripterIsNIL := False;
   FRTTILibrariesIsNIL := False;
   MSCR_RTTILibraries;

Finalization
   VarMSCR_ConcoleOut := Nil;
   If FGlobalScripter <> Nil Then FGlobalScripter.Free;
   FGlobalScripter := Nil;
   FGlobalScripterIsNIL := True;
   FRTTILibraries.Free;
   FRTTILibraries := Nil;
   FRTTILibrariesIsNIL := True;

End.
